#===========================================================================================
# HSKP2UNR.rb                                                             2010/05/26 - RUBY
#===========================================================================================
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place - Suite 330, Boston, MA 02111-1307, USA, or go to
# http://www.gnu.org/copyleft/lesser.txt.
#
#-------------------------------------------------------------------------------------------
#
# Author: Raphael Couturier (HardPCM)
# Mail: hardpcm666@hotmail.com, hardpcm666@gmail.com
#
# File: HSKP2UNR.rb
# Version: 0.9.8
# Started: 2009/06/28
# Released: 2010/05/26
# Langage: English
# Font32: Courier New
# SpaceTab: 8
#
# Description: Main Source
#
#-------------------------------------------------------------------------------------------
#
# Special thanks to:
#
#   - Dire "solosnake" Stockdale 
#      www.solosnake.com
#
#   - Darknet
#
#   - Optimus_P-Fat
#
#   - Active_Trash 
#
#===========================================================================================


require 'sketchup.rb'

#===========================================================================================
#
#===========================================================================================

class WaitWalkerT3D_UTX

	#===================================================================================
	#
	#===================================================================================

	def initialize

		@m_nIndex = 0
		@m_sTable = ["|","/","-","\\","|","/","-","\\"]
	end

	#===================================================================================
	# Public
	#===================================================================================

	def Tick(nLevel = 1)

		# -- show step --
		a_sText = "Processing at level No = #{"%0.f" %(nLevel)}       \"O.{#{@m_sTable[@m_nIndex]}}.O\" \n"
		code = Sketchup.set_status_text a_sText

		# -- next step --
		@m_nIndex += 1
		if @m_nIndex >= 8

			@m_nIndex = 0;
		end
	end

	#-----------------------------------------------------------------------------------

	def Say(sMessage)

		code = Sketchup.set_status_text sMessage
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class TranslationT3D_UTX

	#===================================================================================
	# Creator
	#===================================================================================

	def initialize()

		
	end

	#===================================================================================
	# Private
	#===================================================================================



	#===================================================================================
	# Public
	#===================================================================================

	def PreTranslate(pContainer)
		return pContainer
	end

	#-----------------------------------------------------------------------------------

	def PostTranslate(pContainer)
		return pContainer
	end

	#-----------------------------------------------------------------------------------

	def TranslateContainer(pContainer)

		print "--TranslateContainer\r\n"
		print pContainer.Name
		print "\r\n"
		print pContainer.Type
		print "\r\n"

		return pContainer
	end

	#-----------------------------------------------------------------------------------
		
	def TranslateValue(pContainer, pValue)

		print "--TranslateValue\r\n"
		print pValue.Name
		print "\r\n"
		print pValue.Value
		print "\r\n"

		return pValue
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class PolygonTranslationToSkpT3D_UTX

	#===================================================================================
	# Creator
	#===================================================================================

	def initialize()

		@m_pCurrentModel = nil
		@m_pProgress = WaitWalkerT3D_UTX.new

		# --- put into a group ---------------------------------------------------------
		self.NextModel("Polygon")
	end

	#===================================================================================
	# Private
	#===================================================================================

	def NextModel(sName = nil)

		a_pModel = Sketchup.active_model
		a_pEntities = a_pModel.active_entities
		@m_pCurrentModel = a_pEntities.add_group
		@m_pCurrentModel.description = sName
		@m_pCurrentModel.name = sName

	end

	#-----------------------------------------------------------------------------------

	def DiscreteSubstractFromValues(pA, pB)

		a_vResult = []
		
		a_nCount = pA.length
		a_nIndex = 0
		while a_nIndex < a_nCount

			a_vResult.push(pA[a_nIndex] - pB[a_nIndex])

			a_nIndex = a_nIndex + 1
		end

		return a_vResult
	end

	#-----------------------------------------------------------------------------------

	def DotProductFromValues(pA, pB)

		a_nResult = 0
		
		a_nCount = pA.length
		a_nIndex = 0
		while a_nIndex < a_nCount

			a_nResult = a_nResult + (pA[a_nIndex] * pB[a_nIndex])
			a_nIndex = a_nIndex + 1
		end

		return a_nResult
	end

	#-----------------------------------------------------------------------------------

	def ExtractValuesFromContents(sContent)

		if sContent == nil
			return []
		end

		a_pSplit = sContent.split(",")
		
		a_nCount = a_pSplit.length
		a_nIndex = 0
		while a_nIndex < a_nCount

			a_pSplit[a_nIndex] = a_pSplit[a_nIndex].to_f
			a_nIndex = a_nIndex + 1
		end

		return a_pSplit
	end

	#===================================================================================
	# Public
	#===================================================================================

	def PreTranslate(pContainer)

		return pContainer

	end

	#-----------------------------------------------------------------------------------

	def PostTranslate(pContainer)

		return pContainer

	end

	#-----------------------------------------------------------------------------------

	def TranslateContainer(pContainer)

		# --- indicate progression

		@m_pProgress.Tick(666)

		# --- control the stupidity ----------------------------------------------------

		if pContainer == nil
			return nil
		end
		if pContainer == nil
			return nil
		end

		# --- we want a polygon only ---------------------------------------------------

		if pContainer.Type != "Polygon"
			return pContainer
		end

		# --- initialise ---------------------------------------------------------------

		a_vOrigin = nil
		a_vNormal = nil
		a_vTextureU = nil
		a_vTextureV = nil
		a_pVertex = []
		a_pTextureUV = []

		# --- extract the vector value -------------------------------------------------

		a_nCount = pContainer.GetCount
		a_nIndex = 0
		while a_nIndex < a_nCount

			a_pValue = pContainer.GetValueByIndex(a_nIndex)

			if a_pValue != nil

				a_pName = a_pValue.Name

				if a_pName == "Origin"
					a_vOrigin = self.ExtractValuesFromContents(a_pValue.Value)
				end
				if a_pName == "Normal"
					a_vNormal = self.ExtractValuesFromContents(a_pValue.Value)
				end
				if a_pName == "TextureU"
					a_vTextureU = self.ExtractValuesFromContents(a_pValue.Value)
				end
				if a_pName == "TextureV"
					a_vTextureV = self.ExtractValuesFromContents(a_pValue.Value)
				end
				if a_pName == "Vertex"
					a_pVertex.push(self.ExtractValuesFromContents(a_pValue.Value))
				end
			end

			a_nIndex = a_nIndex + 1
		end

		# --- create the UV coordinates -----------------------------------------------
		
		a_nCount = a_pVertex.size
		a_nIndex = 0
		while a_nIndex < a_nCount

			a_vTemp = self.DiscreteSubstractFromValues(a_pVertex[a_nIndex], a_vOrigin)
			a_vUV = [self.DotProductFromValues(a_vTemp, a_vTextureU), self.DotProductFromValues(a_vTemp, a_vTextureV)]
			a_pTextureUV.push(a_vUV)

			a_nIndex = a_nIndex + 1
		end

		# --- draw the polygon ---------------------------------------------------------

		begin
			a_pNewFace = @m_pCurrentModel.entities.add_face a_pVertex
		rescue => e
#			print e
#			print "\r\n"
#			print a_pPoints.to_s
#			print "\r\n"
		end 

		# --- everything is ok ---------------------------------------------------------
		return pContainer
	end

	#-----------------------------------------------------------------------------------
		
	def TranslateValue(pContainer, pValue)

		return pValue
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class TriangleTranslationToSkpT3D_UTX

	#===================================================================================
	# Creator
	#===================================================================================

	def initialize()

		@m_pCurrentModel = nil
		@m_pProgress = WaitWalkerT3D_UTX.new

		# --- put into a group ---------------------------------------------------------
		self.NextModel("Polygon")
	end

	#===================================================================================
	# Private
	#===================================================================================

	def NextModel(sName = nil)

		a_pModel = Sketchup.active_model
		a_pEntities = a_pModel.active_entities
		@m_pCurrentModel = a_pEntities.add_group
		@m_pCurrentModel.description = sName
		@m_pCurrentModel.name = sName

	end

	#-----------------------------------------------------------------------------------

	def DiscreteSubstractFromValues(pA, pB)

		a_vResult = []
		
		a_nCount = pA.length
		a_nIndex = 0
		while a_nIndex < a_nCount

			a_vResult.push(pA[a_nIndex] - pB[a_nIndex])

			a_nIndex = a_nIndex + 1
		end

		return a_vResult
	end

	#-----------------------------------------------------------------------------------

	def DotProductFromValues(pA, pB)

		a_nResult = 0
		
		a_nCount = pA.length
		a_nIndex = 0
		while a_nIndex < a_nCount

			a_nResult = a_nResult + (pA[a_nIndex] * pB[a_nIndex])
			a_nIndex = a_nIndex + 1
		end

		return a_nResult
	end

	#-----------------------------------------------------------------------------------

	def ExtractValuesFromContents(sContent)

		if sContent == nil
			return []
		end

		a_pSplit = sContent.split(",")
		
		a_nCount = a_pSplit.length
		a_nIndex = 0
		while a_nIndex < a_nCount

			a_pSplit[a_nIndex] = a_pSplit[a_nIndex].to_f
			a_nIndex = a_nIndex + 1
		end

		return a_pSplit
	end

	#===================================================================================
	# Public
	#===================================================================================

	def PreTranslate(pContainer)

		return pContainer

	end

	#-----------------------------------------------------------------------------------

	def PostTranslate(pContainer)

		return pContainer

	end

	#-----------------------------------------------------------------------------------

	def TranslateContainer(pContainer)

		# --- indicate progression

		@m_pProgress.Tick(666)

		# --- control the stupidity ----------------------------------------------------

		if pContainer == nil
			return nil
		end
		if pContainer == nil
			return nil
		end

		# --- we want a polygon only ---------------------------------------------------

		if pContainer.Type != "Triangle"
			return pContainer
		end

		# --- initialise ---------------------------------------------------------------

		a_vOrigin = nil
		a_vNormal = nil
		a_vTextureU = nil
		a_vTextureV = nil
		a_pVertex = []
		a_pTextureUV = []

		# --- extract the vector value -------------------------------------------------

		a_nCount = pContainer.GetCount
		a_nIndex = 0
		while a_nIndex < a_nCount

			a_pValue = pContainer.GetValueByIndex(a_nIndex)

			if a_pValue != nil

				a_pName = a_pValue.Name

				if a_pName == "Origin"
					a_vOrigin = self.ExtractValuesFromContents(a_pValue.Value)
				end
				if a_pName == "Normal"
					a_vNormal = self.ExtractValuesFromContents(a_pValue.Value)
				end
				if a_pName == "TextureU"
					a_vTextureU = self.ExtractValuesFromContents(a_pValue.Value)
				end
				if a_pName == "TextureV"
					a_vTextureV = self.ExtractValuesFromContents(a_pValue.Value)
				end
				if a_pName == "Vertex"
					a_pVertex.push(self.ExtractValuesFromContents(a_pValue.Value))
				end
			end

			a_nIndex = a_nIndex + 1
		end

		# --- create the UV coordinates -----------------------------------------------
		
		a_nCount = a_pVertex.size
		a_nIndex = 0
		while a_nIndex < a_nCount

			a_vTemp = self.DiscreteSubstractFromValues(a_pVertex[a_nIndex], a_vOrigin)
			a_vUV = [self.DotProductFromValues(a_vTemp, a_vTextureU), self.DotProductFromValues(a_vTemp, a_vTextureV)]
			a_pTextureUV.push(a_vUV)

			a_nIndex = a_nIndex + 1
		end

		# --- draw the polygon ---------------------------------------------------------

		begin
			a_pNewFace = @m_pCurrentModel.entities.add_face a_pVertex
		rescue => e
#			print e
#			print "\r\n"
#			print a_pPoints.to_s
#			print "\r\n"
		end 

		# --- everything is ok ---------------------------------------------------------
		return pContainer
	end

	#-----------------------------------------------------------------------------------
		
	def TranslateValue(pContainer, pValue)

		return pValue
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class ParserT3D_UTX

	#===================================================================================
	# Creator
	#===================================================================================

	def initialize()

		@m_pFirstObject = nil
		@m_pProgress = WaitWalkerT3D_UTX.new
	end

	#===================================================================================
	# Private
	#===================================================================================

	def ReadTXT(sPath = nil)

		if sPath == nil
			return nil
		end

		a_pFile = File.new(sPath,"rb")
		a_pStats = a_pFile.stat

		a_sText = a_pFile.sysread a_pStats.size

		return a_sText
	end

	#-----------------------------------------------------------------------------------

	def isValidSpace(nChar, bNewLine)

		if nChar == " "
			return true
		end

		if nChar == "="
			return true
		end

		if (nChar == "\r") && (bNewLine == true)
			return true
		end

		if (nChar == "\n") && (bNewLine == true)
			return true
		end

		if nChar == "\t"
			return true
		end

		return false
	end

	#-----------------------------------------------------------------------------------

	def SpaceLength(sContent, nIndex, bNewLine)

		a_nCount = 0

		while nIndex < sContent.length

			if !self.isValidSpace(sContent[nIndex].chr, bNewLine) == true
				return a_nCount
			end

			nIndex = nIndex + 1
			a_nCount = a_nCount + 1
		end

		return a_nCount
	end

	#-----------------------------------------------------------------------------------

	def WordLength(sContent, nIndex, bNewLine)

		a_nCount = 0

		while nIndex < sContent.length

			if self.isValidSpace(sContent[nIndex].chr, bNewLine) == true
				return a_nCount
			end

			nIndex = nIndex + 1
			a_nCount = a_nCount + 1
		end
			
		return a_nCount
	end

	#-----------------------------------------------------------------------------------

	def LineRemainingLength(sContent, nIndex)

		a_nCount = 0;

		while nIndex < sContent.length

			if sContent[nIndex].chr == "\r"
				return a_nCount;
			end

			if sContent[nIndex].chr == "\n"
				return a_nCount;
			end
				
			nIndex = nIndex + 1;
			a_nCount = a_nCount + 1;
		end

		return a_nCount
	end

	#-----------------------------------------------------------------------------------

	def ExtractSubstring(sContent, nIndex, nLength)

		s_sExtracted = String.new
		a_nCount = sContent.length

		while (nLength > 0) && (nIndex < a_nCount)
			s_sExtracted = s_sExtracted + (sContent[nIndex].chr)
			nIndex = nIndex + 1
			nLength = nLength - 1
		end

		return s_sExtracted
	end

	#-----------------------------------------------------------------------------------

        def FilterContainerName(sName)
            return sName
        end

	#-----------------------------------------------------------------------------------

    def FindOrCreateContainer(pParent, sContainerName)

#		print sContainerName
#		print "\r\n"

		a_pContainer = nil

		if (pParent == nil)

			a_pContainer = ContainerT3D_UTX.new(FilterContainerName(sContainerName))
			@m_pFirstObject = a_pContainer

		else

			if pParent.Name == self.FilterContainerName(sContainerName)
				return pParent
			end

                	a_TmpContainer = pParent.GetValueByName(sContainerName)

	                if a_TmpContainer == nil
				a_pContainer = ContainerT3D_UTX.new(self.FilterContainerName(sContainerName))
				pParent.Push(a_pContainer)
	                else
				a_pContainer = a_TmpContainer
			end
		end

		return a_pContainer
	end

	#===================================================================================
	# Public
	#===================================================================================

	def ParseObject(pParent, sContent, nIndex)

		# --- indicate progression

		@m_pProgress.Tick(nIndex)

		# --- get the begin keyword
		a_nSpaceLength = self.SpaceLength(sContent, nIndex, false)
		nIndex += a_nSpaceLength
		a_nWordLength = self.WordLength(sContent, nIndex, true)

		if a_nWordLength < 1
			return nIndex
		end

		a_sWordExtracted = self.ExtractSubstring(sContent, nIndex, a_nWordLength)
		nIndex += a_nWordLength

		if a_sWordExtracted != "Begin"
			return nIndex
		end

		# --- get the object type
		a_nSpaceLength = self.SpaceLength(sContent, nIndex, false)
		nIndex += a_nSpaceLength
		a_nWordLength = self.WordLength(sContent, nIndex, true)
		if a_nWordLength < 1
			return nIndex
		end

		a_sWordExtracted = self.ExtractSubstring(sContent, nIndex, a_nWordLength)
		nIndex += a_nWordLength

		# --- find or create the object container
      	a_pObject = self.FindOrCreateContainer(pParent, a_sWordExtracted)

        # --- extract values on first line
		a_nSpaceLength = SpaceLength(sContent, nIndex, false)

		if a_nSpaceLength > 0

			nIndex += a_nSpaceLength
			a_nWordLength = self.WordLength(sContent, nIndex, true)

			if a_nWordLength < 1
				return nIndex
			end

			a_sWordExtracted = self.ExtractSubstring(sContent, nIndex, a_nWordLength)
			nIndex += a_nWordLength



			while a_sWordExtracted != "End"

				# --- get the value content
				if sContent[nIndex].chr == "="

					nIndex = nIndex + 1
					a_nWordLength = WordLength(sContent, nIndex, true)

					if a_nWordLength < 1

						return nIndex
					end
					a_sContentExtracted = self.ExtractSubstring(sContent, nIndex, a_nWordLength)
					nIndex += a_nWordLength
					a_pObject.Push(ValueT3D_UTX.new(a_sWordExtracted, a_sContentExtracted, false))

				else

					# --- error
					return nIndex
				end
				
				# --- get the next value
				a_nSpaceLength = self.SpaceLength(sContent, nIndex, false)
				if a_nSpaceLength < 1

					break;
				end
				nIndex += a_nSpaceLength
				a_nWordLength = self.WordLength(sContent, nIndex, true)
				if a_nWordLength < 1
					return nIndex
				end
				a_sWordExtracted = self.ExtractSubstring(sContent, nIndex, a_nWordLength)
				nIndex += a_nWordLength
			end
		end

		# --- extract values
		a_nSpaceLength = SpaceLength(sContent, nIndex, true)

		if a_nSpaceLength > 0

			# ---

#			if a_sWordExtracted != "Begin"

				nIndex += a_nSpaceLength
				a_nWordLength = self.WordLength(sContent, nIndex, true)
	
				if a_nWordLength < 1
					return nIndex
				end
	
				a_sWordExtracted = self.ExtractSubstring(sContent, nIndex, a_nWordLength)
				nIndex += a_nWordLength
#			end

			# ---

			while a_sWordExtracted != "End"

				if a_sWordExtracted == "Begin"

					nIndex -= a_nWordLength
					nIndex = self.ParseObject(a_pObject, sContent, nIndex)

				else

					a_nSpaceLength = self.SpaceLength(sContent, nIndex, true)
					nIndex += a_nSpaceLength;
					a_nWordLength = self.LineRemainingLength(sContent, nIndex)
					if a_nWordLength < 1
						return nIndex
					end
					String a_sContentExtracted = self.ExtractSubstring(sContent, nIndex, a_nWordLength)
					nIndex += a_nWordLength

					a_pObject.Push(ValueT3D_UTX.new(a_sWordExtracted, a_sContentExtracted, true))
				end

				# --- get the next value
				a_nSpaceLength = self.SpaceLength(sContent, nIndex, true)
				nIndex += a_nSpaceLength
				a_nWordLength = self.WordLength(sContent, nIndex, true)

				if a_nWordLength < 1
					return nIndex
				end

				a_sWordExtracted = self.ExtractSubstring(sContent, nIndex, a_nWordLength);
				nIndex += a_nWordLength;
			end
			if a_sWordExtracted == "End"

				a_nSpaceLength = self.SpaceLength(sContent, nIndex, true)
				nIndex += a_nSpaceLength
				a_nWordLength = self.WordLength(sContent, nIndex, true)
				if a_nWordLength < 1
					return nIndex
				end
				a_sWordExtracted = self.ExtractSubstring(sContent, nIndex, a_nWordLength)
				nIndex += a_nWordLength
			end
		end
			
		return nIndex
	end

	#-----------------------------------------------------------------------------------

	def ImportT3D(sPath = nil)

		a_sContent = self.ReadTXT(sPath)
           	self.ParseObject(@m_pFirstObject, a_sContent, 0)
	end

	#-----------------------------------------------------------------------------------

        def TranslateTheKeys(pTranslation = nil)

		if (@m_pFirstObject != nil) && (pTranslation != nil)
			@m_pFirstObject = pTranslation.TranslateContainer(@m_pFirstObject)

			if @m_pFirstObject != nil
				@m_pFirstObject.TranslateTheKeys(pTranslation)
			end
		end
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class ValueT3D_UTX

	#===================================================================================
	# Creator
	#===================================================================================

	def initialize(sName = nil, sValue = nil, bNewLine = true, sOperator = "=")

		@m_sName = sName
		@m_sValue = sValue
		@m_bNewLine = bNewLine
		@m_sOperator = sOperator
	end

	#===================================================================================
	# Private
	#===================================================================================

	def generateTab(nTabLevel)

		a_sText = ""
		while nTabLevel > 0
			
			a_sText += "   "
			nTabLevel -= 1
		end

		return a_sText
	end

	#===================================================================================
	# Public
	#===================================================================================

	def Name
		return @m_sName
	end

	#-----------------------------------------------------------------------------------

	def Value
		return @m_sValue
	end

	#-----------------------------------------------------------------------------------

	def isNewLine

		return @m_bNewLine
	end

	#-----------------------------------------------------------------------------------

	def toText(nTabLevel = 0)

		a_sText = ""
		if @m_bNewLine == true

			a_sText += generateTab(nTabLevel)
		end
		a_sText += @m_sName + @m_sOperator + @m_sValue
		if @m_bNewLine == true

			a_sText = a_sText + "\r\n"
		end
		return a_sText
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class ContainerT3D_UTX

	#===================================================================================
	# Creator
	#===================================================================================

	def initialize(sType = "Actor")

		@m_sName = "ContainerT3D_UTX"
		@m_sType = sType
		@m_pValues = []
	end

	#===================================================================================
	# Private
	#===================================================================================

	def generateTab(nTabLevel)

		a_sText = ""
		while nTabLevel > 0
			
			a_sText += "   "
			nTabLevel -= 1
		end

		return a_sText
	end

	#-----------------------------------------------------------------------------------

	def Name

		return @m_sName
	end

	#-----------------------------------------------------------------------------------

	def Type

		return @m_sType
	end

	#===================================================================================
	# Public
	#===================================================================================

	def GetValueByIndex(nIndex)

		return @m_pValues[nIndex]
	end

	#-----------------------------------------------------------------------------------

	def GetValueByName(sName)

		a_nCount = @m_pValues.size
		a_nIndex = 0
		while a_nIndex < a_nCount

			a_pValue = @m_pValues[a_nIndex]
			if a_pValue == nil
				next
			end

			if a_pValue.Name == sName
				return a_pValue
			end

			a_nIndex += 1
		end

		return nil
	end

	#-----------------------------------------------------------------------------------

	def GetCount

		return @m_pValues.size
	end

	#-----------------------------------------------------------------------------------

	def isNewLine

		return true
	end

	#-----------------------------------------------------------------------------------

	def Push(pValue = nil)

		@m_pValues.push(pValue)
	end

	#-----------------------------------------------------------------------------------

	def toText(nTabLevel = 0)

		# --- add the header
		a_sText = generateTab(nTabLevel) + "Begin " + @m_sType
		
		# --- export the value on first line
		a_nIndexB = 0
		for a_nIndex in a_nIndexB..@m_pValues.length-1

			a_pValue = @m_pValues[a_nIndex]
			if a_pValue == nil

				next
			end
			if a_pValue.isNewLine == true

				a_nIndexB = a_nIndex
				break
			else
				a_sText += " " + a_pValue.toText(nTabLevel + 1)
			end
			a_nIndex += 1
		end

		# --- just the line and dump the remaining values
		a_sText += "\r\n"

		# --- export the remaining values
		for a_nIndex in a_nIndexB..@m_pValues.length-1

			a_pValue = @m_pValues[a_nIndex]
			if a_pValue == nil

				next
			end
			a_sText += a_pValue.toText(nTabLevel + 1)
			a_nIndex += 1
		end

		# --- add the footer
		a_sText += generateTab(nTabLevel) + "End " + @m_sType + "\r\n"

		# --- return that
		return a_sText
	end

	#-----------------------------------------------------------------------------------

        def TranslateTheKeys(pTranslation = nil)

		a_nIndex = 0
		while a_nIndex < @m_pValues.size

			# --- check if the value is a container in fact
			pValue = @m_pValues[a_nIndex]

			if pValue == nil
				continue
			end

			if pValue.class == ContainerT3D_UTX

				# --- this a container, so process it as that
				pTemp = pTranslation.TranslateContainer(pValue)

				if pTemp != nil
					pTemp.TranslateTheKeys(pTranslation)
				end
				@m_pValues[a_nIndex] = pTemp

			else

				# --- this a single value, so process it as that
				@m_pValues[a_nIndex] = pTranslation.TranslateValue(self, pValue)
			end

			a_nIndex = a_nIndex + 1
		end
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class VertexT3D_UTX

	#===================================================================================
	# Creator
	#===================================================================================

	def initialize(nX = 0.0, nY = 0.0, nZ = 0.0)

		@m_nX = nX
		@m_nY = nY
		@m_nZ = nZ
	end

	#===================================================================================
	# Private
	#===================================================================================

	def NumberToStringT3D(nNumber)

		return "#{"%+013.6f" %(nNumber)}"
	end

	#===================================================================================
	# Public
	#===================================================================================

	def x

		return @m_nX
	end

	#-----------------------------------------------------------------------------------

	def y

		return @m_nY
	end

	#-----------------------------------------------------------------------------------

	def z

		return @m_nZ
	end

	#-----------------------------------------------------------------------------------

	def x=(nValue)

		@m_nX = nValue
	end

	#-----------------------------------------------------------------------------------

	def y=(nValue)

		@m_nY = nValue
	end

	#-----------------------------------------------------------------------------------

	def z=(nValue)

		@m_nZ = nValue
	end

	#-----------------------------------------------------------------------------------

	def GetClone

		return VertexT3D_UTX.new(@m_nX, @m_nY, @m_nZ)
	end

	#-----------------------------------------------------------------------------------

	def toValue(sName)

		a_sContent = NumberToStringT3D(@m_nX) + "," + NumberToStringT3D(@m_nY) + "," + NumberToStringT3D(@m_nZ)

		return ValueT3D_UTX.new(sName, a_sContent, true, "   ")
	end

	#-----------------------------------------------------------------------------------

	def Normalise

		a_nLenght = (@m_nX  * @m_nX) + (@m_nY  * @m_nY) + (@m_nZ  * @m_nZ)
		a_nLenghtSquare = 1.0;
		if a_nLenghtSquare > 0.0

			a_nLenghtSquare /= Math.sqrt(a_nLenght);
		end

		@m_nX *= a_nLenghtSquare;
		@m_nY *= a_nLenghtSquare;
		@m_nZ *= a_nLenghtSquare;

	end

	#-----------------------------------------------------------------------------------

	def ScalarScale(nFactor)

		@m_nX *= nFactor
		@m_nY *= nFactor
		@m_nZ *= nFactor
	end

	#-----------------------------------------------------------------------------------

	def Scale(pVector)

		@m_nX *= pVector.x
		@m_nY *= pVector.y
		@m_nZ *= pVector.z
	end

	#-----------------------------------------------------------------------------------

	def Translate(nX = 0.0, nY = 0.0, nZ = 0.0)

		@m_nX += nX
		@m_nY += nY
		@m_nZ += nZ
	end

	#-----------------------------------------------------------------------------------

	def Rotate(pVector)
	
		a_nFactor = (2.0 * 3.1415926535897932384626433832795) / 65536.0

		a_nPitch = pVector.x * a_nFactor;
		a_nRoll = pVector.y * a_nFactor;
		a_nYaw = pVector.z * a_nFactor;

		# --- Roll, Rotate X Axis --------------------------------------------------
		a_nX = @m_nX;
		a_nY = @m_nY;
		a_nZ = @m_nZ;
		@m_nY = (a_nY * Math.cos(nPitch)) - (a_nZ * Math.sin(nPitch));
		@m_nZ = (a_nY * Math.sin(nPitch)) + (a_nZ * Math.cos(nPitch));

		# --- Pitch, Rotate Y Axis -------------------------------------------------
		a_nX = @m_nX;
		a_nY = @m_nY;
		a_nZ = @m_nZ;
		@m_nZ = (a_nZ * Math.cos(nRoll)) - (a_nX * Math.sin(nRoll));
		@m_nX = (a_nZ * Math.sin(nRoll)) + (a_nX * Math.cos(nRoll));

		# --- Yaw, Rotate Z Axis ---------------------------------------------------
		a_nX = @m_nX;
		a_nY = @m_nY;
		a_nZ = @m_nZ;
		@m_nX = (a_nX * Math.cos(nYaw)) - (a_nY * Math.sin(nYaw));
		@m_nY = (a_nX * Math.sin(nYaw)) + (a_nY * Math.cos(nYaw));

	end

	#===================================================================================
	# Static
	#===================================================================================

	def VertexT3D_UTX::DotProduct(pLeft, pRight)

		return ((pLeft.x * pRight.x) + (pLeft.y * pRight.y) + (pLeft.z * pRight.z))
	end

	#-----------------------------------------------------------------------------------

	def VertexT3D_UTX::FromSketchup(vVector)

		return VertexT3D_UTX.new(vVector.x, vVector.y, vVector.z)
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class PolygonT3D_UTX

	#===================================================================================
	# Creator
	#===================================================================================

	def initialize(vVertex = nil, vUvsList = nil)

		@m_vVertex = []
		if vVertex != nil

			@m_vVertex = vVertex
		end

		@m_nVertexCount = @m_vVertex.length

		@m_vUvsList = []
		@m_vUvsOrigin = VertexT3D_UTX.new(0.0, 0.0, 0.0)
		if vUvsList != nil

			@m_vUvsList = vUvsList
			@m_vUvsOrigin = @m_vUvsList[0]
		end

		@m_nU = 1.0
		@m_nV = 1.0
		@m_vOrigin = VertexT3D_UTX.new(0.0, 0.0, 0.0)
		@m_vNormal = VertexT3D_UTX.new(0.0, 0.0, 0.0)
		@m_vTextureU = VertexT3D_UTX.new(0.0, 0.0, 0.0)
		@m_vTextureV = VertexT3D_UTX.new(0.0, 0.0, 0.0)

		@m_bChanged = true
		@m_bImpossible = false
		
	end

	#===================================================================================
	# Public
	#===================================================================================

	def SetTextureScaling(nU = 1.0, nV = 1.0)


		@m_nU = nU
		@m_nV = nV
	end

	#-----------------------------------------------------------------------------------

	def Update

		# --- control stupidity ----------------------------------------------------

		if @m_bChanged == false

			return
		end

		@m_nVertexCount = @m_vVertex.length
		if @m_nVertexCount < 1

			return
		end


		# --- translate texture coordinate -----------------------------------------

		a_vUvs = [];
		for a_nIndex in 0..@m_nVertexCount-1

			a_vTemp = VertexT3D_UTX.new(@m_vUvsList[a_nIndex].x, @m_vUvsList[a_nIndex].y, @m_vUvsList[a_nIndex].z)
			a_vTemp.Translate(-@m_vUvsOrigin.x, -@m_vUvsOrigin.y, -@m_vUvsOrigin.z)
			a_vUvs.push(a_vTemp)
		end

		# --- compute the normal (unreal don"T check it?) --------------------------

		a_nX = @m_vVertex[1].x - @m_vVertex[0].x
		a_nY = @m_vVertex[1].y - @m_vVertex[0].y
		a_nZ = @m_vVertex[1].z - @m_vVertex[0].z
		a_vDeltaLineA = VertexT3D_UTX.new(a_nX, a_nY, a_nZ)

		a_nX = @m_vVertex[2].x - @m_vVertex[0].x
		a_nY = @m_vVertex[2].y - @m_vVertex[0].y
		a_nZ = @m_vVertex[2].z - @m_vVertex[0].z
		a_vDeltaLineB = VertexT3D_UTX.new(a_nX, a_nY, a_nZ)

		a_nX = (a_vDeltaLineA.y * a_vDeltaLineB.z) - (a_vDeltaLineA.z * a_vDeltaLineB.y)
		a_nY = (a_vDeltaLineA.z * a_vDeltaLineB.x) - (a_vDeltaLineA.x * a_vDeltaLineB.z)
		a_nZ = (a_vDeltaLineA.x * a_vDeltaLineB.y) - (a_vDeltaLineA.y * a_vDeltaLineB.x)
		@m_vNormal = VertexT3D_UTX.new(a_nX, a_nY, a_nZ)
		@m_vNormal.Normalise()

		# --- compute the texture vertex --------------------------------------------

		a_nX = a_vUvs[1].x - a_vUvs[0].x
		a_nY = a_vUvs[1].y - a_vUvs[0].y
		a_nZ = 0
		a_vDeltaVectorA = VertexT3D_UTX.new(a_nX, a_nY, a_nZ)

		a_nX = a_vUvs[2].x - a_vUvs[0].x
		a_nY = a_vUvs[2].y - a_vUvs[0].y
		a_nZ = 0
		a_vDeltaVectorB = VertexT3D_UTX.new(a_nX, a_nY, a_nZ)

		a_nDelta11 = (a_vDeltaLineA.x * a_vDeltaLineA.x) + (a_vDeltaLineA.y * a_vDeltaLineA.y) + (a_vDeltaLineA.z * a_vDeltaLineA.z)
		a_nDelta12 = (a_vDeltaLineA.x * a_vDeltaLineB.x) + (a_vDeltaLineA.y * a_vDeltaLineB.y) + (a_vDeltaLineA.z * a_vDeltaLineB.z)
		a_nDelta22 = (a_vDeltaLineB.x * a_vDeltaLineB.x) + (a_vDeltaLineB.y * a_vDeltaLineB.y) + (a_vDeltaLineB.z * a_vDeltaLineB.z)
		a_nDivisorXX = (a_nDelta11 * a_nDelta22) - (a_nDelta12 * a_nDelta12);

		if a_nDivisorXX < 0.00000000001 && a_nDivisorXX >= 0.0

			a_nDivisorXX = 1.0
		end
		if a_nDivisorXX < 0.0 && a_nDivisorXX > -0.00000000001

			a_nDivisorXX = -1.0
		end

		a_nFactor = 1.0 / a_nDivisorXX

		a_nX = ((a_vDeltaVectorA.x * a_nDelta22) - (a_vDeltaVectorB.x * a_nDelta12)) * a_nFactor
		a_nY = ((a_vDeltaVectorA.y * a_nDelta22) - (a_vDeltaVectorB.y * a_nDelta12)) * a_nFactor
		a_nZ = ((a_vDeltaVectorA.z * a_nDelta22) - (a_vDeltaVectorB.z * a_nDelta12)) * a_nFactor
		a_vGradientA = VertexT3D_UTX.new(a_nX, a_nY, a_nZ)

		a_nX = ((a_vDeltaVectorB.x * a_nDelta11) - (a_vDeltaVectorA.x * a_nDelta12)) * a_nFactor
		a_nY = ((a_vDeltaVectorB.y * a_nDelta11) - (a_vDeltaVectorA.y * a_nDelta12)) * a_nFactor
		a_nZ = ((a_vDeltaVectorB.z * a_nDelta11) - (a_vDeltaVectorA.z * a_nDelta12)) * a_nFactor
		a_vGradientB = VertexT3D_UTX.new(a_nX, a_nY, a_nZ)

		a_nX = (a_vDeltaLineA.x * a_vGradientA.x) + (a_vDeltaLineB.x * a_vGradientB.x)
		a_nY = (a_vDeltaLineA.y * a_vGradientA.x) + (a_vDeltaLineB.y * a_vGradientB.x)
		a_nZ = (a_vDeltaLineA.z * a_vGradientA.x) + (a_vDeltaLineB.z * a_vGradientB.x)
		@m_vTextureU = VertexT3D_UTX.new(a_nX, a_nY, a_nZ)

		a_nX = (a_vDeltaLineA.x * a_vGradientA.y) + (a_vDeltaLineB.x * a_vGradientB.y)
		a_nY = (a_vDeltaLineA.y * a_vGradientA.y) + (a_vDeltaLineB.y * a_vGradientB.y)
		a_nZ = (a_vDeltaLineA.z * a_vGradientA.y) + (a_vDeltaLineB.z * a_vGradientB.y)
		@m_vTextureV = VertexT3D_UTX.new(a_nX, a_nY, a_nZ)

		# --- compute the origin -----------------------------------------------------

		a_nDeltaU1 = (a_vDeltaLineA.x * @m_vTextureU.x) + (a_vDeltaLineA.y * @m_vTextureU.y) + (a_vDeltaLineA.z * @m_vTextureU.z)
		a_nDeltaU2 = (a_vDeltaLineB.x * @m_vTextureU.x) + (a_vDeltaLineB.y * @m_vTextureU.y) + (a_vDeltaLineB.z * @m_vTextureU.z)

		a_nDeltaV1 = (a_vDeltaLineA.x * @m_vTextureV.x) + (a_vDeltaLineA.y * @m_vTextureV.y) + (a_vDeltaLineA.z * @m_vTextureV.z)
		a_nDeltaV2 = (a_vDeltaLineB.x * @m_vTextureV.x) + (a_vDeltaLineB.y * @m_vTextureV.y) + (a_vDeltaLineB.z * @m_vTextureV.z)

		a_nDivisorXX = (a_nDeltaU1 * a_nDeltaV2) - (a_nDeltaU2 * a_nDeltaV1)
		@m_bImpossible = false
		if (a_nDivisorXX < 0.00000000001 && a_nDivisorXX > -0.00000000001)

				a_nDivisorXX = 1.0
				@m_bImpossible = true
		end
		a_nFactor = 1.0 / a_nDivisorXX

		a_nB1 = ((@m_vUvsOrigin.x * a_nDeltaV2) - (@m_vUvsOrigin.y * a_nDeltaU2)) * a_nFactor
		a_nB2 = ((@m_vUvsOrigin.y * a_nDeltaU1) - (@m_vUvsOrigin.x * a_nDeltaV1)) * a_nFactor

		a_nX = @m_vVertex[0].x - ((a_vDeltaLineA.x * a_nB1) + (a_vDeltaLineB.x * a_nB2))
		a_nY = @m_vVertex[0].y - ((a_vDeltaLineA.y * a_nB1) + (a_vDeltaLineB.y * a_nB2))
		a_nZ = @m_vVertex[0].z - ((a_vDeltaLineA.z * a_nB1) + (a_vDeltaLineB.z * a_nB2))
		@m_vOrigin = VertexT3D_UTX.new(a_nX, a_nY, a_nZ)

		# ----- scale the texture vertex ------

		@m_vTextureU.ScalarScale(@m_nU)
		@m_vTextureV.ScalarScale(@m_nV)

		# --- adjust the orientation -------------------------------------------------

		@m_vVertex[0].y *= -1.0
		@m_vVertex[1].y *= -1.0
		@m_vVertex[2].y *= -1.0

		@m_vOrigin.y *= -1.0

		@m_vNormal.y *= -1.0

		@m_vTextureU.y *= -1.0
		@m_vTextureV.y *= -1.0
	end

	#-----------------------------------------------------------------------------------

	def toValue(sTexture = "DefaultTexture", bBlankFace = true)

		self.Update

		a_pContainer = ContainerT3D_UTX.new("Polygon")

		a_pContainer.Push(ValueT3D_UTX.new("Texture", sTexture, false))
#		a_pContainer.Push(ValueT3D_UTX.new("Flags", "3584", false))			# todo: shit in ut1999

		if @m_bImpossible == true

			a_pContainer.Push(@m_vVertex[0].toValue("Origin"))
		else

			a_pContainer.Push(@m_vOrigin.toValue("Origin"))
		end

		a_pContainer.Push(@m_vNormal.toValue("Normal"))

		if @m_bImpossible == true

			a_pContainer.Push(@m_vNormal.toValue("TextureU"));
			a_pContainer.Push(@m_vNormal.toValue("TextureV"));
		else
			a_pContainer.Push(@m_vTextureU.toValue("TextureU"));
			a_pContainer.Push(@m_vTextureV.toValue("TextureV"));
		end

		if bBlankFace == true

			a_pContainer.Push(@m_vVertex[1].toValue("Vertex"))
			a_pContainer.Push(@m_vVertex[0].toValue("Vertex"))
			a_pContainer.Push(@m_vVertex[2].toValue("Vertex"))
		else

			a_pContainer.Push(@m_vVertex[0].toValue("Vertex"))
			a_pContainer.Push(@m_vVertex[1].toValue("Vertex"))
			a_pContainer.Push(@m_vVertex[2].toValue("Vertex"))
		end

		return a_pContainer
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class DataProcessorT3D_UT2K4

	#===================================================================================
	#
	#===================================================================================

	def initialize(sPath = "C:\ut2k4.t3d", bWorld = true, bBlankFace = true, bAdditiveBrush = true, bAdditiveSystem = false)

		@m_sPath = sPath
		@m_bWorld = bWorld
		@m_bBlankFace = bBlankFace
		@m_bAdditiveBrush = bAdditiveBrush
		@m_bAdditiveSystem = bAdditiveSystem

		@m_pFirstContainer = nil
		@m_pCurrentContainerA = nil
		@m_pCurrentContainerB = nil
		@m_pCurrentContainerC = nil

		@m_nCurrentIndex = 5

		@m_bBlankFace = true

		@m_pFile = nil
	end

	#===================================================================================
	# Public
	#===================================================================================

	def BeginWorld


		# --- create the file ---

		@m_pFile = File.new(@m_sPath,"wb")

		# --- write header to file ---

		@m_pFile.write("Begin Map\r\n")

		# ---

		a_pTempA = ContainerT3D_UTX.new("Actor")

		a_pTempA.Push(ValueT3D_UTX.new("Class","LevelInfo",false))
		a_pTempA.Push(ValueT3D_UTX.new("Name","LevelInfo4",false))

		a_pTempA.Push(ValueT3D_UTX.new("TimeSeconds","2.711100",true))
		a_pTempA.Push(ValueT3D_UTX.new("Level","LevelInfo'myLevel.LevelInfo4'",true))
		a_pTempA.Push(ValueT3D_UTX.new("Region","(Zone=LevelInfo'myLevel.LevelInfo4',iLeaf=-1)",true))
		a_pTempA.Push(ValueT3D_UTX.new("Tag","\"LevelInfo\"",true))
		a_pTempA.Push(ValueT3D_UTX.new("PhysicsVolume","DefaultPhysicsVolume'myLevel.DefaultPhysicsVolume4'",true))

		@m_pFile.write(a_pTempA.toText(1))

		# ---

		a_pTempA = ContainerT3D_UTX.new("Actor")

		a_pTempA.Push(ValueT3D_UTX.new("Class","Brush",false))
		a_pTempA.Push(ValueT3D_UTX.new("Name","Brush4",false))

		a_pTempA.Push(ValueT3D_UTX.new("MainScale","(SheerAxis=SHEER_ZX)",true))
		a_pTempA.Push(ValueT3D_UTX.new("PostScale","(SheerAxis=SHEER_ZX)",true))
		a_pTempA.Push(ValueT3D_UTX.new("Level","LevelInfo'myLevel.LevelInfo4'",true))
		a_pTempA.Push(ValueT3D_UTX.new("Region","(Zone=LevelInfo'myLevel.LevelInfo4',iLeaf=-1)",true))
		a_pTempA.Push(ValueT3D_UTX.new("Tag","\"Brush\"",true))
		a_pTempA.Push(ValueT3D_UTX.new("Group","\"Cube\"",true))
		a_pTempA.Push(ValueT3D_UTX.new("PhysicsVolume","DefaultPhysicsVolume'myLevel.DefaultPhysicsVolume4'",true))

			a_pTempB = ContainerT3D_UTX.new("Brush")
			a_pTempB.Push(ValueT3D_UTX.new("Name","Brush",false))

				a_pTempC = ContainerT3D_UTX.new("PolyList")

				a_pTempD = ContainerT3D_UTX.new("Polygon")

					a_pTempD.Push(ValueT3D_UTX.new("Origin","-00128.000000,-00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Normal","-00001.000000,+00000.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00000.000000,+00001.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00000.000000,-00001.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,-00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,-00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,+00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,+00128.000000,-00128.000000",true,"  "))

				a_pTempC.Push(a_pTempD)

				a_pTempD = ContainerT3D_UTX.new("Polygon")

					a_pTempD.Push(ValueT3D_UTX.new("Origin","-00128.000000,+00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Normal","+00000.000000,+00001.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00001.000000,+00000.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00000.000000,-00001.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,+00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,+00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,+00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,+00128.000000,-00128.000000",true,"  "))

				a_pTempC.Push(a_pTempD)

				a_pTempD = ContainerT3D_UTX.new("Polygon")

					a_pTempD.Push(ValueT3D_UTX.new("Origin","+00128.000000,+00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Normal","+00001.000000,+00000.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00000.000000,-00001.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00000.000000,-00001.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,+00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,+00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,-00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,-00128.000000,-00128.000000",true,"  "))

				a_pTempC.Push(a_pTempD)

				a_pTempD = ContainerT3D_UTX.new("Polygon")

					a_pTempD.Push(ValueT3D_UTX.new("Origin","+00128.000000,-00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Normal","+00000.000000,-00001.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureU","-00001.000000,+00000.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00000.000000,-00001.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,-00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,-00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,-00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,-00128.000000,-00128.000000",true,"  "))

				a_pTempC.Push(a_pTempD)

				a_pTempD = ContainerT3D_UTX.new("Polygon")

					a_pTempD.Push(ValueT3D_UTX.new("Origin","-00128.000000,+00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Normal","+00000.000000,+00000.000000,+00001.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00001.000000,+00000.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00001.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,+00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,-00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,-00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,+00128.000000,+00128.000000",true,"  "))

				a_pTempC.Push(a_pTempD)

				a_pTempD = ContainerT3D_UTX.new("Polygon")

					a_pTempD.Push(ValueT3D_UTX.new("Origin","-00128.000000,-00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Normal","+00000.000000,+00000.000000,-00001.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00001.000000,+00000.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,-00001.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,-00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,+00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,+00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,-00128.000000,-00128.000000",true,"  "))

				a_pTempC.Push(a_pTempD)

				a_pTempB.Push(a_pTempC)
	
			a_pTempA.Push(a_pTempB)

		a_pTempA.Push(ValueT3D_UTX.new("Brush","Model'myLevel.Brush'",true))

		@m_pFile.write(a_pTempA.toText(1))

		# ---

		a_pTempA = ContainerT3D_UTX.new("Actor")

		a_pTempA.Push(ValueT3D_UTX.new("Class","DefaultPhysicsVolume",false))
		a_pTempA.Push(ValueT3D_UTX.new("Name","DefaultPhysicsVolume4",false))

		a_pTempA.Push(ValueT3D_UTX.new("Priority","-1000000",true))
		a_pTempA.Push(ValueT3D_UTX.new("bNoDelete","Vrai",true))
		a_pTempA.Push(ValueT3D_UTX.new("Level","LevelInfo'myLevel.LevelInfo4'",true))
		a_pTempA.Push(ValueT3D_UTX.new("Region","(Zone=LevelInfo'myLevel.LevelInfo4',iLeaf=-1)",true))
		a_pTempA.Push(ValueT3D_UTX.new("Tag","\"DefaultPhysicsVolume\"",true))
		a_pTempA.Push(ValueT3D_UTX.new("PhysicsVolume","DefaultPhysicsVolume'myLevel.DefaultPhysicsVolume4'",true))

		@m_pFile.write(a_pTempA.toText(1))

		# ---

		if @m_bWorld == true

			if @m_bAdditiveSystem == true
	
				
	
			end

		end

		# ---
		
	end

	#-----------------------------------------------------------------------------------

	def BeginBrush(sConfig)

		@m_pCurrentContainerA = ContainerT3D_UTX.new("Actor")

		@m_pCurrentContainerA.Push(ValueT3D_UTX.new("Class","Brush",false))
		@m_pCurrentContainerA.Push(ValueT3D_UTX.new("Name","Brush#{"%0.f" %(@m_nCurrentIndex)}",false))

		if @m_bAdditiveBrush

			@m_pCurrentContainerA.Push(ValueT3D_UTX.new("CsgOper","CSG_Add",true))
		else

			@m_pCurrentContainerA.Push(ValueT3D_UTX.new("CsgOper","CSG_Subtract",true))
		end

		@m_pCurrentContainerA.Push(ValueT3D_UTX.new("MainScale","(SheerAxis=SHEER_ZX)",true))
		@m_pCurrentContainerA.Push(ValueT3D_UTX.new("PostScale","(SheerAxis=SHEER_ZX)",true))
		@m_pCurrentContainerA.Push(ValueT3D_UTX.new("Level","LevelInfo'myLevel.LevelInfo4'",true))
		@m_pCurrentContainerA.Push(ValueT3D_UTX.new("Region","(Zone=LevelInfo'myLevel.LevelInfo4',iLeaf=-1)",true))
		@m_pCurrentContainerA.Push(ValueT3D_UTX.new("Tag","\"Brush\"",true))
		@m_pCurrentContainerA.Push(ValueT3D_UTX.new("PhysicsVolume","DefaultPhysicsVolume'myLevel.DefaultPhysicsVolume4'",true))

		@m_pCurrentContainerB = ContainerT3D_UTX.new("Brush")
		@m_pCurrentContainerB.Push(ValueT3D_UTX.new("Name","Model#{"%0.f" %(@m_nCurrentIndex)}",false))

		@m_pCurrentContainerC = ContainerT3D_UTX.new("PolyList")
	end

	#-----------------------------------------------------------------------------------

	def ProcessFace(pFace, pMaterial, pTransformations)

		# --- extract the mesh -----------------------------------------------------
		a_pMesh = pFace.mesh 7

		# --- transform the mesh ---------------------------------------------------
		a_pMesh.transform! pTransformations

		# --- extract material info ------------------------------------------------
		a_nMaterialU = 1.0
		a_nMaterialV = 1.0
		a_sTexture = "Engine.DefaultTexture"
		if pMaterial != nil
			a_sTexture = pMaterial.name
			if a_sTexture == nil
				a_sTexture = "Engine.DefaultTexture"
			end
			a_pTexture = pMaterial.texture
			if a_pTexture != nil
				if a_pTexture.width != 0.0
					a_nMaterialU = a_pTexture.width
				else
					a_nMaterialU = 1.0
				end
				if a_pTexture.height != 0.0
					a_nMaterialV = a_pTexture.height
				else
					a_nMaterialV = 1.0
				end
			end
		end
		a_nMaterialU *= 0.125
		a_nMaterialV *= 0.125

		# --- loop for each triangle in the face -----------------------------------
		a_nIndex = 1
		while (a_nIndex <= a_pMesh.count_polygons)

			# --- extract the triangle indexes ---------------------------------
			a_pPolygon = a_pMesh.polygon_at a_nIndex

			# --- extract the triangle vertex ----------------------------------
			a_vA = VertexT3D_UTX.FromSketchup(a_pMesh.point_at a_pPolygon[0])
			a_vB = VertexT3D_UTX.FromSketchup(a_pMesh.point_at a_pPolygon[1])
			a_vC = VertexT3D_UTX.FromSketchup(a_pMesh.point_at a_pPolygon[2])
			a_vVertexList = [a_vA, a_vB, a_vC]

			# --- extract the texture coordinate -------------------------------
			a_vA = VertexT3D_UTX.FromSketchup(a_pMesh.uv_at a_pPolygon[0],a_pPolygon[0])
			a_vB = VertexT3D_UTX.FromSketchup(a_pMesh.uv_at a_pPolygon[1],a_pPolygon[1])
			a_vC = VertexT3D_UTX.FromSketchup(a_pMesh.uv_at a_pPolygon[2],a_pPolygon[2])
			a_vA.y = 1.0 - a_vA.y
			a_vB.y = 1.0 - a_vB.y
			a_vC.y = 1.0 - a_vC.y
			a_vUvsList = [a_vA, a_vB, a_vC]

			# --- convert to value ---------------------------------------------
			a_pPolygon = PolygonT3D_UTX.new(a_vVertexList, a_vUvsList)
			a_pPolygon.SetTextureScaling(a_nMaterialU, a_nMaterialV)
			a_pContainer = a_pPolygon.toValue(a_sTexture, @m_bBlankFace)
			@m_pCurrentContainerC.Push(a_pContainer)

			# --- next triangle ------------------------------------------------
			a_nIndex += 1
		end
	end

	#-----------------------------------------------------------------------------------

	def ProcessText(pText, pTransformations)

		

	end

	#-----------------------------------------------------------------------------------

	def FinishBrush

		@m_pCurrentContainerB.Push(@m_pCurrentContainerC)
		@m_pCurrentContainerA.Push(@m_pCurrentContainerB)

		@m_pCurrentContainerA.Push(ValueT3D_UTX.new("Brush","Model'myLevel.Model#{"%0.f" %(@m_nCurrentIndex)}'",true))

		# --- write brush class ---
		@m_pFile.write(@m_pCurrentContainerA.toText(1))
		@m_pFile.flush

		# --- next brush ---
		@m_nCurrentIndex += 1

	end

	#-----------------------------------------------------------------------------------

	def FinishWorld

		# --- write footer to file ---
		@m_pFile.write("   Begin Surface\r\n")
		@m_pFile.write("   End Surface\r\n")
		@m_pFile.write("End Map\r\n")

		# --- flush and close the file ---
		@m_pFile.flush
		@m_pFile.close
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class DataProcessorT3D_UT3

	#===================================================================================
	#
	#===================================================================================

	def initialize(sPath = "C:\ut3.t3d", bWorld = true, bBlankFace = true, bAdditiveBrush = true, bAdditiveSystem = false)

		@m_sPath = sPath
		@m_bWorld = bWorld
		@m_bBlankFace = bBlankFace
		@m_bAdditiveBrush = bAdditiveBrush
		@m_bAdditiveSystem = bAdditiveSystem

		@m_pFirstContainer = nil
		@m_pCurrentContainerA = nil
		@m_pCurrentContainerB = nil
		@m_pCurrentContainerC = nil
		@m_pCurrentContainerD = nil

		@m_nCurrentIndex = 5

		@m_bBlankFace = true

		@m_pFile = nil

	end

	#===================================================================================
	# Public
	#===================================================================================

	def BeginWorld

		# --- create the file ---

		@m_pFile = File.new(@m_sPath,"wb")

		# --- write header to file ---

		@m_pFile.write("Begin Map\r\n")
		@m_pFile.write("   Begin Level NAME=PersistentLevel\r\n")

		# --- write world info ---

		a_pTempA = ContainerT3D_UTX.new("Actor")

		a_pTempA.Push(ValueT3D_UTX.new("Class","WorldInfo",false))
		a_pTempA.Push(ValueT3D_UTX.new("Name","WorldInfo_1",false))
		a_pTempA.Push(ValueT3D_UTX.new("Archetype","WorldInfo'Engine.Default__WorldInfo'",false))

		a_pTempA.Push(ValueT3D_UTX.new("TimeSeconds","13.666049",true))
		a_pTempA.Push(ValueT3D_UTX.new("RealTimeSeconds","22.255566",true))
		a_pTempA.Push(ValueT3D_UTX.new("AudioTimeSeconds","22.255566",true))
		a_pTempA.Push(ValueT3D_UTX.new("Tag","\"WorldInfo\"",true))
		a_pTempA.Push(ValueT3D_UTX.new("Name","\"WorldInfo_1\"",true))
		a_pTempA.Push(ValueT3D_UTX.new("ObjectArchetype","WorldInfo'Engine.Default__WorldInfo'",true))

		@m_pFile.write(a_pTempA.toText(2))

		# --- write user brush (avoid stupid crashing) ---

		a_pTempA = ContainerT3D_UTX.new("Actor")

		a_pTempA.Push(ValueT3D_UTX.new("Class","Brush",false))
		a_pTempA.Push(ValueT3D_UTX.new("Name","Brush_1",false))
		a_pTempA.Push(ValueT3D_UTX.new("Archetype","Brush'Engine.Default__Brush'",false))

		a_pTempE = ContainerT3D_UTX.new("Object")
		
			a_pTempE.Push(ValueT3D_UTX.new("Class","BrushComponent",false))
			a_pTempE.Push(ValueT3D_UTX.new("Name","BrushComponent0",false))
			a_pTempE.Push(ValueT3D_UTX.new("ObjName","BrushComponent_1",false))
			a_pTempE.Push(ValueT3D_UTX.new("Archetype","BrushComponent'Engine.Default__Brush:BrushComponent0'",false))
	
			a_pTempE.Push(ValueT3D_UTX.new("Brush","Model'Brush'",true))
			a_pTempE.Push(ValueT3D_UTX.new("LightingChannels","(bInitialized=True,Dynamic=True)",true))
			a_pTempE.Push(ValueT3D_UTX.new("Name","\"BrushComponent_1\"",true))
			a_pTempE.Push(ValueT3D_UTX.new("ObjectArchetype","BrushComponent'Engine.Default__Brush:BrushComponent0'",true))

		a_pTempA.Push(a_pTempE)

		a_pTempB = ContainerT3D_UTX.new("Brush")

			a_pTempB.Push(ValueT3D_UTX.new("Name","Brush",false))

				a_pTempC = ContainerT3D_UTX.new("PolyList")

				a_pTempD = ContainerT3D_UTX.new("Polygon")

					a_pTempD.Push(ValueT3D_UTX.new("Flags","3584",false))

					a_pTempD.Push(ValueT3D_UTX.new("Origin","-00128.000000,-00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Normal","-00001.000000,+00000.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00000.000000,+00001.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00000.000000,-00001.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,-00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,-00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,+00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,+00128.000000,-00128.000000",true,"  "))

				a_pTempC.Push(a_pTempD)

				a_pTempD = ContainerT3D_UTX.new("Polygon")

					a_pTempD.Push(ValueT3D_UTX.new("Flags","3584",false))

					a_pTempD.Push(ValueT3D_UTX.new("Origin","-00128.000000,+00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Normal","+00000.000000,+00001.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00001.000000,+00000.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00000.000000,-00001.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,+00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,+00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,+00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,+00128.000000,-00128.000000",true,"  "))

				a_pTempC.Push(a_pTempD)

				a_pTempD = ContainerT3D_UTX.new("Polygon")

					a_pTempD.Push(ValueT3D_UTX.new("Flags","3584",false))

					a_pTempD.Push(ValueT3D_UTX.new("Origin","+00128.000000,+00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Normal","+00001.000000,+00000.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00000.000000,-00001.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00000.000000,-00001.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,+00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,+00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,-00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,-00128.000000,-00128.000000",true,"  "))

				a_pTempC.Push(a_pTempD)

				a_pTempD = ContainerT3D_UTX.new("Polygon")

					a_pTempD.Push(ValueT3D_UTX.new("Flags","3584",false))

					a_pTempD.Push(ValueT3D_UTX.new("Origin","+00128.000000,-00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Normal","+00000.000000,-00001.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureU","-00001.000000,+00000.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00000.000000,-00001.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,-00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,-00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,-00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,-00128.000000,-00128.000000",true,"  "))

				a_pTempC.Push(a_pTempD)

				a_pTempD = ContainerT3D_UTX.new("Polygon")

					a_pTempD.Push(ValueT3D_UTX.new("Flags","3584",false))

					a_pTempD.Push(ValueT3D_UTX.new("Origin","-00128.000000,+00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Normal","+00000.000000,+00000.000000,+00001.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00001.000000,+00000.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00001.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,+00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,-00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,-00128.000000,+00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,+00128.000000,+00128.000000",true,"  "))

				a_pTempC.Push(a_pTempD)

				a_pTempD = ContainerT3D_UTX.new("Polygon")

					a_pTempD.Push(ValueT3D_UTX.new("Flags","3584",false))

					a_pTempD.Push(ValueT3D_UTX.new("Origin","-00128.000000,-00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Normal","+00000.000000,+00000.000000,-00001.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00001.000000,+00000.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,-00001.000000,+00000.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,-00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","-00128.000000,+00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,+00128.000000,-00128.000000",true,"  "))
					a_pTempD.Push(ValueT3D_UTX.new("Vertex","+00128.000000,-00128.000000,-00128.000000",true,"  "))

				a_pTempC.Push(a_pTempD)

				a_pTempB.Push(a_pTempC)
	
			a_pTempA.Push(a_pTempB)

		a_pTempA.Push(ValueT3D_UTX.new("Brush","Model'Brush'",true))
		a_pTempA.Push(ValueT3D_UTX.new("BrushComponent","BrushComponent'BrushComponent_1'",true))
		a_pTempA.Push(ValueT3D_UTX.new("Components(0)","BrushComponent'BrushComponent_1'",true))
		a_pTempA.Push(ValueT3D_UTX.new("bHidden","False",true))
		a_pTempA.Push(ValueT3D_UTX.new("Tag","\"Brush\"",true))
		a_pTempA.Push(ValueT3D_UTX.new("Group","\"Cube\"",true))
		a_pTempA.Push(ValueT3D_UTX.new("Location","(X=383.999969,Y=0.000000,Z=0.000000)",true))
		a_pTempA.Push(ValueT3D_UTX.new("CollisionComponent","BrushComponent'BrushComponent_1'",true))
		a_pTempA.Push(ValueT3D_UTX.new("Name","\"Brush_1\"",true))
		a_pTempA.Push(ValueT3D_UTX.new("ObjectArchetype","Brush'Engine.Default__Brush'",true))

		@m_pFile.write(a_pTempA.toText(2))

		# --- write the famous big substractive workaround ----

		if @m_bWorld == true

			# ---
	
			if @m_bAdditiveSystem == false

				a_pTempA = ContainerT3D_UTX.new("Actor")
	
				a_pTempA.Push(ValueT3D_UTX.new("Class","Brush",false))
				a_pTempA.Push(ValueT3D_UTX.new("Name","Brush_2",false))
				a_pTempA.Push(ValueT3D_UTX.new("Archetype","Brush'Engine.Default__Brush'",false))
	
				a_pTempE = ContainerT3D_UTX.new("Object")
				
					a_pTempE.Push(ValueT3D_UTX.new("Class","BrushComponent",false))
					a_pTempE.Push(ValueT3D_UTX.new("Name","BrushComponent0",false))
					a_pTempE.Push(ValueT3D_UTX.new("ObjName","BrushComponent_2",false))
					a_pTempE.Push(ValueT3D_UTX.new("Archetype","BrushComponent'Engine.Default__Brush:BrushComponent0'",false))
		
					a_pTempE.Push(ValueT3D_UTX.new("Brush","Model'Model_2'",true))
					a_pTempE.Push(ValueT3D_UTX.new("LightingChannels","(bInitialized=True,Dynamic=True)",true))
					a_pTempE.Push(ValueT3D_UTX.new("Name","\"BrushComponent_2\"",true))
					a_pTempE.Push(ValueT3D_UTX.new("ObjectArchetype","BrushComponent'Engine.Default__Brush:BrushComponent0'",true))
	
				a_pTempA.Push(a_pTempE)

				a_pTempA.Push(ValueT3D_UTX.new("CsgOper","CSG_Add",true))
	
				a_pTempB = ContainerT3D_UTX.new("Brush")
	
					a_pTempB.Push(ValueT3D_UTX.new("Name","Model_2",false))
	
						a_pTempC = ContainerT3D_UTX.new("PolyList")
	
							a_pTempD = ContainerT3D_UTX.new("Polygon")
	
								a_pTempD.Push(ValueT3D_UTX.new("Texture","EngineMaterials.RemoveSurfaceMaterial",false))
								a_pTempD.Push(ValueT3D_UTX.new("Flags","3584",false))
	
								a_pTempD.Push(ValueT3D_UTX.new("Origin","-262144.000000,-262144.000000,-262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Normal","-00001.000000,+00000.000000,+00000.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00000.000000,+00001.000000,+00000.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00000.000000,-00001.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","-262144.000000,-262144.000000,-262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","-262144.000000,-262144.000000,+262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","-262144.000000,+262144.000000,+262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","-262144.000000,+262144.000000,-262144.000000",true,"  "))
	
							a_pTempC.Push(a_pTempD)
	
							a_pTempD = ContainerT3D_UTX.new("Polygon")
	
								a_pTempD.Push(ValueT3D_UTX.new("Texture","EngineMaterials.RemoveSurfaceMaterial",false))
								a_pTempD.Push(ValueT3D_UTX.new("Flags","3584",false))
	
								a_pTempD.Push(ValueT3D_UTX.new("Origin","-262144.000000,+262144.000000,-262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Normal","+00000.000000,+00001.000000,+00000.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00001.000000,-00000.000000,+00000.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00000.000000,-00001.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","-262144.000000,+262144.000000,-262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","-262144.000000,+262144.000000,+262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","+262144.000000,+262144.000000,+262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","+262144.000000,+262144.000000,-262144.000000",true,"  "))
	
							a_pTempC.Push(a_pTempD)
	
							a_pTempD = ContainerT3D_UTX.new("Polygon")
	
								a_pTempD.Push(ValueT3D_UTX.new("Texture","EngineMaterials.RemoveSurfaceMaterial",false))
								a_pTempD.Push(ValueT3D_UTX.new("Flags","3584",false))
	
								a_pTempD.Push(ValueT3D_UTX.new("Origin","+262144.000000,+262144.000000,-262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Normal","+00001.000000,+00000.000000,+00000.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00000.000000,-00001.000000,+00000.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00000.000000,-00001.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","+262144.000000,+262144.000000,-262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","+262144.000000,+262144.000000,+262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","+262144.000000,-262144.000000,+262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","+262144.000000,-262144.000000,-262144.000000",true,"  "))
	
							a_pTempC.Push(a_pTempD)
	
							a_pTempD = ContainerT3D_UTX.new("Polygon")
	
								a_pTempD.Push(ValueT3D_UTX.new("Texture","EngineMaterials.RemoveSurfaceMaterial",false))
								a_pTempD.Push(ValueT3D_UTX.new("Flags","3584",false))
	
								a_pTempD.Push(ValueT3D_UTX.new("Origin","+262144.000000,-262144.000000,-262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Normal","+00000.000000,-00001.000000,+00000.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("TextureU","-00001.000000,-00000.000000,-00000.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00000.000000,-00001.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","+262144.000000,-262144.000000,-262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","+262144.000000,-262144.000000,+262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","-262144.000000,-262144.000000,+262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","-262144.000000,-262144.000000,-262144.000000",true,"  "))
	
							a_pTempC.Push(a_pTempD)
	
							a_pTempD = ContainerT3D_UTX.new("Polygon")
	
								a_pTempD.Push(ValueT3D_UTX.new("Texture","EngineMaterials.RemoveSurfaceMaterial",false))
								a_pTempD.Push(ValueT3D_UTX.new("Flags","3584",false))
	
								a_pTempD.Push(ValueT3D_UTX.new("Origin","-262144.000000,+262144.000000,+262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Normal","+00000.000000,+00000.000000,+00001.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00001.000000,+00000.000000,+00000.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,+00001.000000,+00000.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","-262144.000000,+262144.000000,+262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","-262144.000000,-262144.000000,+262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","+262144.000000,-262144.000000,+262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","+262144.000000,+262144.000000,+262144.000000",true,"  "))
	
							a_pTempC.Push(a_pTempD)
	
							a_pTempD = ContainerT3D_UTX.new("Polygon")
	
								a_pTempD.Push(ValueT3D_UTX.new("Texture","EngineMaterials.RemoveSurfaceMaterial",false))
								a_pTempD.Push(ValueT3D_UTX.new("Flags","3584",false))
	
								a_pTempD.Push(ValueT3D_UTX.new("Origin","-262144.000000,-262144.000000,-262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Normal","+00000.000000,+00000.000000,-00001.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("TextureU","+00001.000000,+00000.000000,+00000.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("TextureV","+00000.000000,-00001.000000,+00000.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","-262144.000000,-262144.000000,-262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","-262144.000000,+262144.000000,-262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","+262144.000000,+262144.000000,-262144.000000",true,"  "))
								a_pTempD.Push(ValueT3D_UTX.new("Vertex","+262144.000000,-262144.000000,-262144.000000",true,"  "))
	
							a_pTempC.Push(a_pTempD)
	
						a_pTempB.Push(a_pTempC)
		
					a_pTempA.Push(a_pTempB)
	
				a_pTempA.Push(ValueT3D_UTX.new("Brush","Model'Model_2'",true))
				a_pTempA.Push(ValueT3D_UTX.new("BrushComponent","BrushComponent'BrushComponent_2'",true))
				a_pTempA.Push(ValueT3D_UTX.new("Components(0)","BrushComponent'BrushComponent_2'",true))
				a_pTempA.Push(ValueT3D_UTX.new("Tag","\"Brush\"",true))
				a_pTempA.Push(ValueT3D_UTX.new("CollisionComponent","BrushComponent'BrushComponent_2'",true))
				a_pTempA.Push(ValueT3D_UTX.new("Name","\"Brush_2\"",true))
				a_pTempA.Push(ValueT3D_UTX.new("ObjectArchetype","Brush'Engine.Default__Brush'",true))
	
				@m_pFile.write(a_pTempA.toText(2))
	
			end
	
			# ---
		end

		# ---
		
	end

	#-----------------------------------------------------------------------------------

	def BeginBrush(sConfig)

		@m_pCurrentContainerB = ContainerT3D_UTX.new("Actor")

		@m_pCurrentContainerB.Push(ValueT3D_UTX.new("Class","Brush",false))
		@m_pCurrentContainerB.Push(ValueT3D_UTX.new("Name","Brush_#{"%0.f" %(@m_nCurrentIndex)}",false))
		@m_pCurrentContainerB.Push(ValueT3D_UTX.new("Archetype","Brush'Engine.Default__Brush'",false))

		a_pTempE = ContainerT3D_UTX.new("Object")
			
			a_pTempE.Push(ValueT3D_UTX.new("Class","BrushComponent",false))
			a_pTempE.Push(ValueT3D_UTX.new("Name","BrushComponent0",false))
			a_pTempE.Push(ValueT3D_UTX.new("ObjName","BrushComponent_#{"%0.f" %(@m_nCurrentIndex)}",false))
			a_pTempE.Push(ValueT3D_UTX.new("Archetype","BrushComponent'Engine.Default__Brush:BrushComponent0'",false))
	
			a_pTempE.Push(ValueT3D_UTX.new("Brush","Model'Model_#{"%0.f" %(@m_nCurrentIndex)}'",true))
			a_pTempE.Push(ValueT3D_UTX.new("LightingChannels","(bInitialized=True,Dynamic=True)",true))
			a_pTempE.Push(ValueT3D_UTX.new("Name","\"BrushComponent_#{"%0.f" %(@m_nCurrentIndex)}\"",true))
			a_pTempE.Push(ValueT3D_UTX.new("ObjectArchetype","BrushComponent'Engine.Default__Brush:BrushComponent0'",true))

		@m_pCurrentContainerB.Push(a_pTempE)

		if @m_bAdditiveBrush

			@m_pCurrentContainerB.Push(ValueT3D_UTX.new("CsgOper","CSG_Add",true))
		else

			@m_pCurrentContainerB.Push(ValueT3D_UTX.new("CsgOper","CSG_Subtract",true))
		end

		@m_pCurrentContainerC = ContainerT3D_UTX.new("Brush")
		@m_pCurrentContainerC.Push(ValueT3D_UTX.new("Name","Model_#{"%0.f" %(@m_nCurrentIndex)}",false))

		@m_pCurrentContainerD = ContainerT3D_UTX.new("PolyList")
	end

	#-----------------------------------------------------------------------------------

	def ProcessFace(pFace, pMaterial, pTransformations)

		# --- extract the mesh -----------------------------------------------------
		a_pMesh = pFace.mesh 7

		# --- transform the mesh ---------------------------------------------------
		a_pMesh.transform! pTransformations

		# --- extract material info ------------------------------------------------
		a_nMaterialU = 1.0
		a_nMaterialV = 1.0
		a_sTexture = "Engine.DefaultTexture"
		if pMaterial != nil
			a_sTexture = pMaterial.name
			if a_sTexture == nil
				a_sTexture = "Engine.DefaultTexture"
			end
			a_pTexture = pMaterial.texture
			if a_pTexture != nil
				if a_pTexture.width != 0.0
					a_nMaterialU = a_pTexture.width
				else
					a_nMaterialU = 1.0
				end
				if a_pTexture.height != 0.0
					a_nMaterialV = a_pTexture.height
				else
					a_nMaterialV = 1.0
				end
			end
		end
		a_nMaterialU *= 0.125
		a_nMaterialV *= 0.125

		# --- loop for each triangle in the face -----------------------------------
		a_nIndex = 1
		while (a_nIndex <= a_pMesh.count_polygons)

			# --- extract the triangle indexes ---------------------------------
			a_pPolygon = a_pMesh.polygon_at a_nIndex

			# --- extract the triangle vertex ----------------------------------
			a_vA = VertexT3D_UTX.FromSketchup(a_pMesh.point_at a_pPolygon[0])
			a_vB = VertexT3D_UTX.FromSketchup(a_pMesh.point_at a_pPolygon[1])
			a_vC = VertexT3D_UTX.FromSketchup(a_pMesh.point_at a_pPolygon[2])
			a_vVertexList = [a_vA, a_vB, a_vC]

			# --- extract the texture coordinate -------------------------------
			a_vA = VertexT3D_UTX.FromSketchup(a_pMesh.uv_at a_pPolygon[0],a_pPolygon[0])
			a_vB = VertexT3D_UTX.FromSketchup(a_pMesh.uv_at a_pPolygon[1],a_pPolygon[1])
			a_vC = VertexT3D_UTX.FromSketchup(a_pMesh.uv_at a_pPolygon[2],a_pPolygon[2])
			a_vA.y = 1.0 - a_vA.y
			a_vB.y = 1.0 - a_vB.y
			a_vC.y = 1.0 - a_vC.y
			a_vUvsList = [a_vA, a_vB, a_vC]

			# --- convert to value ---------------------------------------------
			a_pPolygon = PolygonT3D_UTX.new(a_vVertexList, a_vUvsList)
			a_pPolygon.SetTextureScaling(a_nMaterialU, a_nMaterialV)
			a_pContainer = a_pPolygon.toValue(a_sTexture, @m_bBlankFace)
			@m_pCurrentContainerD.Push(a_pContainer)

			# --- next triangle ------------------------------------------------
			a_nIndex += 1
		end
	end

	#-----------------------------------------------------------------------------------

	def ProcessText(pText, pTransformations)

		

	end

	#-----------------------------------------------------------------------------------

	def FinishBrush

		@m_pCurrentContainerC.Push(@m_pCurrentContainerD)
		@m_pCurrentContainerB.Push(@m_pCurrentContainerC)

		@m_pCurrentContainerB.Push(ValueT3D_UTX.new("Brush","Model'Model_#{"%0.f" %(@m_nCurrentIndex)}'",true))
		@m_pCurrentContainerB.Push(ValueT3D_UTX.new("BrushComponent","BrushComponent'BrushComponent_#{"%0.f" %(@m_nCurrentIndex)}'",true))
		@m_pCurrentContainerB.Push(ValueT3D_UTX.new("Components(0)","BrushComponent'BrushComponent_#{"%0.f" %(@m_nCurrentIndex)}'",true))
		@m_pCurrentContainerB.Push(ValueT3D_UTX.new("CreationTime","4.670985",true))
		@m_pCurrentContainerB.Push(ValueT3D_UTX.new("Tag","\"Brush\"",true))
		@m_pCurrentContainerB.Push(ValueT3D_UTX.new("CollisionComponent","BrushComponent'BrushComponent_#{"%0.f" %(@m_nCurrentIndex)}'",true))
		@m_pCurrentContainerB.Push(ValueT3D_UTX.new("Name","\"Brush_#{"%0.f" %(@m_nCurrentIndex)}\"",true))
		@m_pCurrentContainerB.Push(ValueT3D_UTX.new("ObjectArchetype","Brush'Engine.Default__Brush'",true))

		# --- write brush class ---
		@m_pFile.write(@m_pCurrentContainerB.toText(2))
		@m_pFile.flush

		# --- next brush ---
		@m_nCurrentIndex += 1

	end

	#-----------------------------------------------------------------------------------

	def FinishWorld

		# --- write footer to file ---
		@m_pFile.write("   End Level\r\n")
		@m_pFile.write("   Begin Surface\r\n")
		@m_pFile.write("   End Surface\r\n")
		@m_pFile.write("End Map\r\n")

		# --- flush and close the file ---
		@m_pFile.flush
		@m_pFile.close

	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class DataExporterASE_UTX

	#===================================================================================
	#
	#===================================================================================

	def initialize(sAsePath = "C:\ut3.t3d")

		@m_pFile = nil
		@m_sAsePath = sAsePath

		@m_vVertex = []
		@m_vVertex.fill([0.0,0.0,0.0],0,15000)
		@m_nNextVertex = 0

		@m_vVertexUV = []
		@m_vVertexUV.fill([0.0,0.0],0,15000)
		@m_nNextVertexUV = 0

		@m_vNormal = []
		@m_vNormal.fill([0.0,0.0,0.0],0,15000)
		@m_nNextNormal = 0

		@m_pTriangle = []
		@m_pTriangle.fill([0,0,0,1,1,1],0,15000)
		@m_nNextTriangle = 0

		@m_pMaterial = []
		@m_pMaterial.fill([""],0,1000)
		@m_nNextMaterial = 0
	end

	#===================================================================================
	# Private
	#===================================================================================

	def NextVertexXYZ(pPacket)

		# --- Search an existing XYZ ---------------------------------------------------
		a_nIndex = @m_nNextVertex - 1
		a_nCount = 256
		if a_nCount > a_nIndex + 1

			a_nCount = a_nIndex + 1
		end
		while a_nCount > -1

			a_pTemp = @m_vVertex[a_nIndex]
			if (a_pTemp[0] == pPacket[0]) && (a_pTemp[1] == pPacket[1]) && (a_pTemp[2] == pPacket[2])

				if (a_pTemp[3] == pPacket[3]) && (a_pTemp[4] == pPacket[4]) && (a_pTemp[5] == pPacket[5])

					return a_nIndex
				end
			end
			a_nIndex = a_nIndex - 1
			a_nCount = a_nCount - 1
		end

		# --- Create a new XYZ ---------------------------------------------------------
		@m_vVertex[@m_nNextVertex] = pPacket
		@m_nNextVertex = @m_nNextVertex + 1
		return @m_nNextVertex - 1
	end

	#-----------------------------------------------------------------------------------

	def NextVertexUV(pPacket)

		# --- Search an existing UV ----------------------------------------------------
		a_nIndex = @m_nNextVertexUV - 1
		a_nCount = 256
		if a_nCount > a_nIndex + 1

			a_nCount = a_nIndex + 1
		end
		while a_nCount > -1

			a_pTemp = @m_vVertexUV[a_nIndex]
			if (a_pTemp[0] == pPacket[0]) && (a_pTemp[1] == pPacket[1]) && (a_pTemp[2] == pPacket[2])

				return a_nIndex
			end
			a_nIndex = a_nIndex - 1
			a_nCount = a_nCount - 1
		end

		# --- Create a new UV ----------------------------------------------------------
		@m_vVertexUV[@m_nNextVertexUV] = pPacket
		@m_nNextVertexUV = @m_nNextVertexUV + 1
		return @m_nNextVertexUV - 1
	end

	#-----------------------------------------------------------------------------------

	def NextNormalXYZ(pPacket)

		@m_vNormal[@m_nNextNormal] = pPacket
		@m_nNextNormal = @m_nNextNormal + 1
		return @m_nNextNormal - 1
	end

	#-----------------------------------------------------------------------------------

	def NextFaceABC(pPacket)

		@m_pTriangle[@m_nNextTriangle] = pPacket
		@m_nNextTriangle = @m_nNextTriangle + 1
		return @m_nNextTriangle - 1
	end

	#-----------------------------------------------------------------------------------

	def NextMaterial(pPacket)

		# --- Search an existing material ----------------------------------------------
		a_nIndex = @m_nNextMaterial - 1
		while a_nIndex > -1

			a_pTempPacket = @m_pMaterial[a_nIndex]
			if a_pTempPacket[0] == pPacket[0]

				return a_nIndex
			end
			a_nIndex = a_nIndex - 1
		end

		# --- Create a new material ----------------------------------------------------
		@m_pMaterial[@m_nNextMaterial] = pPacket
		@m_nNextMaterial = @m_nNextMaterial + 1
		return @m_nNextMaterial - 1
	end

	#-----------------------------------------------------------------------------------

	def VectorFromSketchup(pVector)

		return [pVector.x, pVector.y, pVector.z]
	end

	#-----------------------------------------------------------------------------------

	def VectorUvFromSketchup(pVector)

		return [pVector.x, pVector.y]
	end

	#-----------------------------------------------------------------------------------

	def WriteHeader

		@m_pFile.write("*3DSMAX_ASCIIEXPORT\r\n")
		@m_pFile.write("*COMMENT \"HardPCM'S ASE output, Using HSKP2UNR.rb Plugin for Sketchup \"\r\n")

		@m_pFile.flush
	end

	#-----------------------------------------------------------------------------------

	def WriteMaterials

		@m_pFile.write("*MATERIAL_LIST {\r\n")
		@m_pFile.write("*MATERIAL_COUNT 1\r\n")
		@m_pFile.write("	*MATERIAL 0 {\r\n")
		@m_pFile.write("		*MATERIAL_NAME \"HSKP2UNR_Tex\"\r\n")
		@m_pFile.write("		*MATERIAL_CLASS \"Multi/Sub-Object\"\r\n")

		@m_pFile.write("		*NUMSUBMTLS #{"%i" %(@m_nNextMaterial)}\r\n")

		a_nIndex = 0
		a_nCount = @m_nNextMaterial
		while (a_nIndex < a_nCount)

			a_pPacket = @m_pMaterial[a_nIndex]
			@m_pFile.write("		*SUBMATERIAL #{"%i" %(a_nIndex)} {\r\n");
			@m_pFile.write("			*MATERIAL_NAME \"#{"%s" %(a_pPacket[0])}\"\r\n");
			@m_pFile.write("			*MATERIAL_CLASS \"Standard\"\r\n");
			@m_pFile.write("			*MAP_DIFFUSE {\r\n");
			@m_pFile.write("				*MAP_CLASS \"Bitmap\"\r\n");
			@m_pFile.write("				*BITMAP \"C:\\ut3\\toto.bmp\"\r\n");
			@m_pFile.write("				*UVW_U_OFFSET 0.0\r\n");
			@m_pFile.write("				*UVW_V_OFFSET 0.0\r\n");
			@m_pFile.write("				*UVW_U_TILING 1.0\r\n");
			@m_pFile.write("				*UVW_V_TILING 1.0\r\n");
			@m_pFile.write("			}\r\n");
			a_nIndex = a_nIndex + 1
		end

		@m_pFile.write("		}\r\n")
		@m_pFile.write("	}\r\n")
		@m_pFile.write("}\r\n")

		@m_pFile.flush
	end

	#-----------------------------------------------------------------------------------

	def WriteGeometryBegin

		@m_pFile.write("*GEOMOBJECT {\r\n")
		@m_pFile.write("	*NODE_NAME \"pCubeShape1\"\r\n")
		@m_pFile.write("	*NODE_TM {\r\n")
		@m_pFile.write("	*NODE_NAME \"pCubeShape1\"\r\n")
		@m_pFile.write("	}\r\n")
		@m_pFile.write("	*MESH {\r\n")
		@m_pFile.write("		*TIMEVALUE 0\r\n")

		@m_pFile.flush
	end

	#-----------------------------------------------------------------------------------

	def WriteVertexXYZ

		@m_pFile.write("		*MESH_NUMVERTEX #{"%i" %(@m_nNextVertex)}\r\n")
		@m_pFile.write("			*MESH_NUMFACES #{"%i" %(@m_nNextTriangle)}\r\n")
		@m_pFile.write("		*MESH_VERTEX_LIST {\r\n")

		a_nIndex = 0
		a_nCount = @m_nNextVertex
		while (a_nIndex < a_nCount)

			a_pPacket = @m_vVertex[a_nIndex]
			@m_pFile.write("			*MESH_VERTEX    #{"%i" %(a_nIndex)}	#{"%+013.6f" %(a_pPacket[0])}	#{"%+013.6f" %(a_pPacket[1])}	#{"%+013.6f" %(a_pPacket[2])}\r\n");
			a_nIndex = a_nIndex + 1
		end

		@m_pFile.write("		}\r\n")

		@m_pFile.flush
	end

	#-----------------------------------------------------------------------------------

	def WriteFaceVertexXYZ

		@m_pFile.write("		*MESH_FACE_LIST {\r\n")

		a_nIndex = 0
		a_nCount = @m_nNextTriangle
		while (a_nIndex < a_nCount)

			a_pPacket = @m_pTriangle[a_nIndex]
			@m_pFile.write("			*MESH_FACE    #{"%i" %(a_nIndex)}: A:    #{"%i" %(a_pPacket[0])} B:    #{"%i" %(a_pPacket[1])} C:    #{"%i" %(a_pPacket[2])}	 AB:	1	BC:	1	CA:	1	*MESH_SMOOTHING 0 	*MESH_MTLID #{"%i" %(a_pPacket[6])}\r\n");
			a_nIndex = a_nIndex + 1
		end

		@m_pFile.write("		}\r\n")

		@m_pFile.flush
	end

	#-----------------------------------------------------------------------------------

	def WriteVertexUV

		@m_pFile.write("		*MESH_NUMTVERTEX #{"%i" %(@m_nNextVertexUV)}\r\n")
		@m_pFile.write("		*MESH_TVERTLIST {\r\n")

		a_nIndex = 0
		a_nCount = @m_nNextVertexUV
		while (a_nIndex < a_nCount)

			a_pPacket = @m_vVertexUV[a_nIndex]
			@m_pFile.write("			*MESH_TVERT #{"%i" %(a_nIndex)}	#{"%+013.6f" %(a_pPacket[0])}	#{"%+013.6f" %(a_pPacket[1])}	0.000000\r\n");
			a_nIndex = a_nIndex + 1
		end

		@m_pFile.write("		}\r\n")

		@m_pFile.flush
	end

	#-----------------------------------------------------------------------------------

	def WriteFaceVertexUV

		@m_pFile.write("		*MESH_NUMTVFACES #{"%i" %(@m_nNextTriangle)}\r\n")
		@m_pFile.write("		*MESH_TFACELIST {\r\n")

		a_nIndex = 0
		a_nCount = @m_nNextTriangle
		while (a_nIndex < a_nCount)

			a_pPacket = @m_pTriangle[a_nIndex]
			@m_pFile.write("			*MESH_TFACE #{"%i" %(a_nIndex)}	#{"%i" %(a_pPacket[3])}	#{"%i" %(a_pPacket[4])}	#{"%i" %(a_pPacket[5])}\r\n");
			a_nIndex = a_nIndex + 1
		end

		@m_pFile.write("		}\r\n")

		@m_pFile.flush
	end

	#-----------------------------------------------------------------------------------

	def WriteFaceNormals

		@m_pFile.write("		*MESH_NORMALS {\r\n")

		a_nIndex = 0
		a_nCount = @m_nNextTriangle
		while (a_nIndex < a_nCount)

			a_pPacket = @m_pTriangle[a_nIndex]
			@m_pFile.write("			*MESH_FACENORMAL #{"%i" %(a_nIndex)}    #{"%+013.6f" %(a_pPacket[7])}    #{"%+013.6f" %(a_pPacket[8])}    #{"%+013.6f" %(a_pPacket[9])}\r\n");
			a_nID = a_pPacket[0]
			a_pTP = @m_vVertex[a_nID]
			@m_pFile.write("				*MESH_VERTEXNORMAL #{"%i" %(a_nID)}	#{"%+013.6f" %(a_pTP[3])}	#{"%+013.6f" %(a_pTP[4])}	#{"%+013.6f" %(a_pTP[5])}\r\n");
			a_nID = a_pPacket[1]
			a_pTP = @m_vVertex[a_nID]
			@m_pFile.write("				*MESH_VERTEXNORMAL #{"%i" %(a_nID)}	#{"%+013.6f" %(a_pTP[3])}	#{"%+013.6f" %(a_pTP[4])}	#{"%+013.6f" %(a_pTP[5])}\r\n");
			a_nID = a_pPacket[2]
			a_pTP = @m_vVertex[a_nID]
			@m_pFile.write("				*MESH_VERTEXNORMAL #{"%i" %(a_nID)}	#{"%+013.6f" %(a_pTP[3])}	#{"%+013.6f" %(a_pTP[4])}	#{"%+013.6f" %(a_pTP[5])}\r\n");
			a_nIndex = a_nIndex + 1
		end

		@m_pFile.write("		}\r\n")

		@m_pFile.flush
	end

	#-----------------------------------------------------------------------------------

	def WriteGeometryFinish

		@m_pFile.write("	}\r\n")
		@m_pFile.write("	*PROP_MOTIONBLUR 0\r\n")
		@m_pFile.write("	*PROP_CASTSHADOW 1\r\n")
		@m_pFile.write("	*PROP_RECVSHADOW 1\r\n")
		@m_pFile.write("	*MATERIAL_REF 0\r\n")
		@m_pFile.write("}\r\n")
	end

	#===================================================================================
	# Public
	#===================================================================================

	def BeginWorld

		@m_pFile = File.new(@m_sAsePath,"wb")

		
	end

	#-----------------------------------------------------------------------------------

	def BeginBrush(sConfig)

	end

	#-----------------------------------------------------------------------------------

	def ProcessFace(pFace, pMaterial, pTransformations)

		# --- extract the mesh -----------------------------------------------------

		a_pMesh = pFace.mesh 7

		# --- transform the mesh ---------------------------------------------------

		a_pMesh.transform! pTransformations

		# --- extract material info ------------------------------------------------

		a_nMaterialU = 1.0
		a_nMaterialV = 1.0
		a_sTexture = "Engine.DefaultTexture"
		if pMaterial != nil
			a_sTexture = pMaterial.name
			if a_sTexture == nil
				a_sTexture = "Engine.DefaultTexture"
			end
			a_pTexture = pMaterial.texture
			if a_pTexture != nil
				if a_pTexture.width != 0.0
					a_nMaterialU = a_pTexture.width
				else
					a_nMaterialU = 1.0
				end
				if a_pTexture.height != 0.0
					a_nMaterialV = a_pTexture.height
				else
					a_nMaterialV = 1.0
				end
			end
		end
		a_nMaterialU *= 0.125
		a_nMaterialV *= 0.125

		a_pMatPack = [a_sTexture, pMaterial]
		a_nMaterialID = self.NextMaterial(a_pMatPack)

		# --- extract the triangle normal ------------------------------------------

		a_vNormal = pFace.normal 
		a_nNx = a_vNormal.x
		a_nNy = a_vNormal.y
		a_nNz = a_vNormal.z

		# --- loop for each triangle in the face -----------------------------------
		a_nIndex = 1
		while (a_nIndex <= a_pMesh.count_polygons)

			# --- extract the triangle indexes ---------------------------------
			a_pPolygon = a_pMesh.polygon_at a_nIndex

			# --- extract the triangle vertex ----------------------------------

			a_pVa = self.VectorFromSketchup(a_pMesh.point_at a_pPolygon[0])
			a_pVb = self.VectorFromSketchup(a_pMesh.point_at a_pPolygon[1])
			a_pVc = self.VectorFromSketchup(a_pMesh.point_at a_pPolygon[2])

#			a_pVa[1] = 1.0 - a_pVa[1]
#			a_pVb[1] = 1.0 - a_pVb[1]
#			a_pVc[1] = 1.0 - a_pVc[1]

			a_pVa[3] = a_nNx
			a_pVa[4] = a_nNy
			a_pVa[5] = a_nNz

			a_pVb[3] = a_nNx
			a_pVb[4] = a_nNy
			a_pVb[5] = a_nNz

			a_pVc[3] = a_nNx
			a_pVc[4] = a_nNy
			a_pVc[5] = a_nNz

			a_nVa = self.NextVertexXYZ(a_pVa)
			a_nVb = self.NextVertexXYZ(a_pVb)
			a_nVc = self.NextVertexXYZ(a_pVc)

			# --- extract the texture coordinate -------------------------------
			a_pTa = self.VectorUvFromSketchup(a_pMesh.uv_at a_pPolygon[0],a_pPolygon[0])
			a_pTb = self.VectorUvFromSketchup(a_pMesh.uv_at a_pPolygon[1],a_pPolygon[1])
			a_pTc = self.VectorUvFromSketchup(a_pMesh.uv_at a_pPolygon[2],a_pPolygon[2])
#			a_pTa.y = -a_pTa.y
#			a_pTb.y = -a_pTb.y
#			a_pTc.y = -a_pTc.y
			a_nTa = self.NextVertexUV(a_pTa)
			a_nTb = self.NextVertexUV(a_pTb)
			a_nTc = self.NextVertexUV(a_pTc)			

			# --- convert to face ---------------------------------------------
			a_pFace = [a_nVa, a_nVb, a_nVc, a_nTa, a_nTb, a_nTc, a_nMaterialID, a_nNx, a_nNy, a_nNz]
			a_nFace = NextFaceABC(a_pFace)

			# --- next triangle ------------------------------------------------
			a_nIndex += 1
		end
	end

	#-----------------------------------------------------------------------------------

	def ProcessText(pText, pTransformations)

	end

	#-----------------------------------------------------------------------------------

	def FinishBrush

		
	end

	#-----------------------------------------------------------------------------------

	def FinishWorld

		self.WriteHeader
		self.WriteMaterials
		self.WriteGeometryBegin
		self.WriteVertexXYZ
		self.WriteFaceVertexXYZ
		self.WriteVertexUV
		self.WriteFaceVertexUV
		self.WriteFaceNormals
		self.WriteGeometryFinish

		@m_pFile.flush
		@m_pFile.close
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class DataExporterOBJ_UTX

	#===================================================================================
	#
	#===================================================================================

	def initialize(sObjPath = "C:\ut3.obj")

		@m_pFile = nil
		@m_sObjPath = sObjPath

		@m_vVertex = []
		@m_vVertex.fill([0.0,0.0,0.0],0,30000)
		@m_nNextVertex = 1

		@m_vVertexUV = []
		@m_vVertexUV.fill([0.0,0.0],0,30000)
		@m_nNextVertexUV = 1

		@m_vNormal = []
		@m_vNormal.fill([0.0,0.0,0.0],0,15000)
		@m_nNextNormal = 1

		@m_pTriangle = []
		@m_pTriangle.fill([0,0,0],0,15000)
		@m_nNextTriangle = 1

		@m_pMaterial = []
		@m_pMaterial.fill(["",nil],0,1000)
		@m_nNextMaterial = 1
		@m_nLastMaterial = 0

		@m_nNextModel = 1
	end

	#===================================================================================
	# Private
	#===================================================================================

	def NextMaterial(pPacket)

		# --- Search an existing material ----------------------------------------------
		a_nIndex = @m_nNextMaterial
		while a_nIndex > 0

			a_pTempPacket = @m_pMaterial[a_nIndex]
			if a_pTempPacket[0] == pPacket[0]

				return a_nIndex
			end
			a_nIndex = a_nIndex - 1
		end

		# --- Create a new material ----------------------------------------------------
		@m_pMaterial[@m_nNextMaterial] = pPacket
		@m_nNextMaterial = @m_nNextMaterial + 1
		@m_pFile.write("usemtl ");
		@m_pFile.write(pPacket[0]);
		@m_pFile.write("\r\n");
		return @m_nNextMaterial - 1
	end

	#-----------------------------------------------------------------------------------

	def NextNormalXYZ(pPacket)

		# --- Search an existing UV ----------------------------------------------------
		a_nIndex = @m_nNextNormal
		a_nCount = 2048
		if a_nCount > a_nIndex - 1

			a_nCount = a_nIndex - 1
		end
		while a_nCount > 0

			a_pTemp = @m_vNormal[a_nIndex]
			if (a_pTemp[0] == pPacket[0]) && (a_pTemp[1] == pPacket[1]) && (a_pTemp[2] == pPacket[2])

				return a_nIndex
			end
			a_nIndex = a_nIndex - 1
			a_nCount = a_nCount - 1
		end

		# --- Create a new Normal ----------------------------------------------------------
		@m_vNormal[@m_nNextNormal] = pPacket
		@m_nNextNormal = @m_nNextNormal + 1
		@m_pFile.write("vn #{"%f" %(pPacket[0])} #{"%f" %(pPacket[1])} #{"%f" %(pPacket[2])}\r\n");
		return @m_nNextNormal - 1
	end

	#-----------------------------------------------------------------------------------

	def NextVertexXYZ(pPacket)

		# --- Search an existing XYZ ---------------------------------------------------
		a_nIndex = @m_nNextVertex
		a_nCount = 2048
		if a_nCount > a_nIndex - 1

			a_nCount = a_nIndex - 1
		end
		while a_nCount > 0

			a_pTemp = @m_vVertex[a_nIndex]
			if (a_pTemp[0] == pPacket[0]) && (a_pTemp[1] == pPacket[1]) && (a_pTemp[2] == pPacket[2])

				return a_nIndex
			end
			a_nIndex = a_nIndex - 1
			a_nCount = a_nCount - 1
		end

		# --- Create a new XYZ ---------------------------------------------------------
		@m_vVertex[@m_nNextVertex] = pPacket
		@m_nNextVertex = @m_nNextVertex + 1
		@m_pFile.write("v #{"%f" %(pPacket[0])} #{"%f" %(pPacket[1])} #{"%f" %(pPacket[2])}\r\n");
		return @m_nNextVertex - 1
	end

	#-----------------------------------------------------------------------------------

	def NextVertexUV(pPacket)

		# --- Search an existing UV ----------------------------------------------------
		a_nIndex = @m_nNextVertexUV
		a_nCount = 2048
		if a_nCount > a_nIndex - 1

			a_nCount = a_nIndex - 1
		end
		while a_nCount > 0

			a_pTemp = @m_vVertexUV[a_nIndex]
			if (a_pTemp[0] == pPacket[0]) && (a_pTemp[1] == pPacket[1])

				return a_nIndex
			end
			a_nIndex = a_nIndex - 1
			a_nCount = a_nCount - 1
		end

		# --- Create a new UV ----------------------------------------------------------
		@m_vVertexUV[@m_nNextVertexUV] = pPacket
		@m_nNextVertexUV = @m_nNextVertexUV + 1
		@m_pFile.write("vt #{"%f" %(pPacket[0])} #{"%f" %(pPacket[1])}\r\n");
		return @m_nNextVertexUV - 1
	end

	#-----------------------------------------------------------------------------------

	def NextFaceABC(pPacket)

		@m_pTriangle[@m_nNextTriangle] = pPacket
		@m_nNextTriangle = @m_nNextTriangle + 1
		@m_pFile.write("f #{"%i" %(pPacket[2])}/#{"%i" %(pPacket[5])}/#{"%i" %(pPacket[1])} #{"%i" %(pPacket[3])}/#{"%i" %(pPacket[6])}/#{"%i" %(pPacket[1])} #{"%i" %(pPacket[4])}/#{"%i" %(pPacket[7])}/#{"%i" %(pPacket[1])}\r\n");
		@m_pFile.write("\r\n");
		return @m_nNextTriangle - 1
	end

	#-----------------------------------------------------------------------------------

	def NextModel(sName)

		@m_pFile.write("g Mesh#{"%i" %(@m_nNextModel)}\r\n");
		@m_pFile.write("\r\n");
		@m_nNextModel = @m_nNextModel + 1
	end

	#-----------------------------------------------------------------------------------

	def VectorFromSketchup(pVector)

		return [pVector.x, pVector.y, pVector.z]
	end

	#-----------------------------------------------------------------------------------

	def VectorUvFromSketchup(pVector)

		return [pVector.x, pVector.y]
	end

	#===================================================================================
	# Public
	#===================================================================================

	def BeginWorld

		@m_pFile = File.new(@m_sObjPath,"wb")

		@m_pFile.write("# Alias OBJ Model File\r\n")
		@m_pFile.write("# Exported With The HardPCM's Google Sketchup Exporter to T3D/PSK/ASE/OBJ Format.\r\n")
		@m_pFile.write("# File units = inches\r\n")
		@m_pFile.write("\r\n")
		@m_pFile.write("# mtllib Textures.mtl\r\n")
		@m_pFile.write("\r\n")

		@m_pFile.flush

	end

	#-----------------------------------------------------------------------------------

	def BeginBrush(sConfig)

		self.NextModel(sConfig)
	end

	#-----------------------------------------------------------------------------------

	def ProcessFace(pFace, pMaterial, pTransformations)

		# --- extract the mesh -----------------------------------------------------
		a_pMesh = pFace.mesh 7

		# --- transform the mesh ---------------------------------------------------
		a_pMesh.transform! pTransformations

		# --- extract material info ------------------------------------------------
		a_nMaterialU = 1.0
		a_nMaterialV = 1.0
		a_sTexture = "Engine.DefaultTexture"
		if pMaterial != nil
			a_sTexture = pMaterial.name
			if a_sTexture == nil
				a_sTexture = "Engine.DefaultTexture"
			end
			a_pTexture = pMaterial.texture
			if a_pTexture != nil
				if a_pTexture.width != 0.0
					a_nMaterialU = a_pTexture.width
				else
					a_nMaterialU = 1.0
				end
				if a_pTexture.height != 0.0
					a_nMaterialV = a_pTexture.height
				else
					a_nMaterialV = 1.0
				end
			end
		end
		a_nMaterialU *= 0.125
		a_nMaterialV *= 0.125

		a_pMatPack = [a_sTexture, pMaterial]
		a_nMaterialID = self.NextMaterial(a_pMatPack)

		# --- extract the triangle normal ------------------------------------------
		a_pVn = VectorFromSketchup(pFace.normal)
		a_nVn = self.NextNormalXYZ(a_pVn)

		# --- loop for each triangle in the face -----------------------------------
		a_nIndex = 1
		while (a_nIndex <= a_pMesh.count_polygons)

			# --- extract the triangle indexes ---------------------------------
			a_pPolygon = a_pMesh.polygon_at a_nIndex

			# --- extract the triangle vertex ----------------------------------

			a_pVa = self.VectorFromSketchup(a_pMesh.point_at a_pPolygon[0])
			a_pVb = self.VectorFromSketchup(a_pMesh.point_at a_pPolygon[1])
			a_pVc = self.VectorFromSketchup(a_pMesh.point_at a_pPolygon[2])

#			a_pVa[1] = 1.0 - a_pVa[1]
#			a_pVb[1] = 1.0 - a_pVb[1]
#			a_pVc[1] = 1.0 - a_pVc[1]

			a_nVa = self.NextVertexXYZ(a_pVa)
			a_nVb = self.NextVertexXYZ(a_pVb)
			a_nVc = self.NextVertexXYZ(a_pVc)

			# --- extract the texture coordinate -------------------------------
			a_pTa = self.VectorUvFromSketchup(a_pMesh.uv_at a_pPolygon[0],a_pPolygon[0])
			a_pTb = self.VectorUvFromSketchup(a_pMesh.uv_at a_pPolygon[1],a_pPolygon[1])
			a_pTc = self.VectorUvFromSketchup(a_pMesh.uv_at a_pPolygon[2],a_pPolygon[2])
			a_pTa.y = -a_pTa.y
			a_pTb.y = -a_pTb.y
			a_pTc.y = -a_pTc.y
			a_nTa = self.NextVertexUV(a_pTa)
			a_nTb = self.NextVertexUV(a_pTb)
			a_nTc = self.NextVertexUV(a_pTc)			

			# --- convert to face ---------------------------------------------
			a_pFace = [a_nMaterialID, a_nVn, a_nVa, a_nVb, a_nVc, a_nTa, a_nTb, a_nTc]
			a_nFace = NextFaceABC(a_pFace)

			# --- next triangle -----------------------------------------------
			a_nIndex += 1
		end

		# --- Remember Last Material ------------------------------------------ 
		@m_nLastMaterial = a_nMaterialID

	end

	#-----------------------------------------------------------------------------------

	def ProcessText(pText, pTransformations)

	end

	#-----------------------------------------------------------------------------------

	def FinishBrush

		
	end

	#-----------------------------------------------------------------------------------

	def FinishWorld

		@m_pFile.flush
		@m_pFile.close
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class DataIteratorerT3D_UTX

	#===================================================================================
	#
	#===================================================================================

	def initialize(pFirstNode = nil, pProcessorNode = nil, bBlankFace = true)

		@m_pProcessorNode = pProcessorNode
		@m_pFirstNode = pFirstNode
		@m_bBlankFace = bBlankFace

		@m_pMaterials = []
		@m_pMaterials.push(nil)
		@m_bMaterialsChanged = false

		@m_Transformations = []
		@m_Transformations.push(Geom::Transformation.new)

		@m_nLevel = 1

		@m_nFacesDone = 0
		@m_nGroupsDone = 0
		@m_nComponentsDone = 0

		@m_pWalker = WaitWalkerT3D_UTX.new
	end

	#===================================================================================
	# Public
	#===================================================================================

	def Start(bReport)

		@m_pProcessorNode.BeginWorld
		self.Process(@m_pFirstNode, @m_pFirstNode)
		@m_pProcessorNode.FinishWorld

		@m_pWalker.Say("Job Done !!!")

		if bReport == true

			a_sStatisticText = "Faces exported: #{"%0.f" %(@m_nFacesDone)} \n"
			a_sStatisticText += "Groups exported: #{"%0.f" %(@m_nGroupsDone)} \n"
			a_sStatisticText += "Components exported: #{"%0.f" %(@m_nComponentsDone)} \n"
			UI.messagebox a_sStatisticText

		end
	end

	#-----------------------------------------------------------------------------------

	def Process(pCurrentNode, pParent)

		# --- determine where is the entity array ----------------------------------
		a_sName = nil
		if(pCurrentNode.class == Sketchup::Group)

			a_sName = pCurrentNode.name
			pCurrentNode = pCurrentNode.entities 
		elsif(pCurrentNode.class == Sketchup::ComponentInstance)

			a_sName = pCurrentNode.name
			pCurrentNode = pCurrentNode.definition.entities
		end

		# --- scan the texts into the array ----------------------------------------

		for a_Entity in pCurrentNode

			# ----- find a text ? ----------------------------------------------
			if(a_Entity.class == Sketchup::Text and a_Entity.layer.visible?)

				@m_pProcessorNode.ProcessText(a_Entity, @m_Transformations.last)
			end

		end

		# --- scan the faces into the array ----------------------------------------
		a_bBrushCreated = false
		for a_Entity in pCurrentNode

			# ----- find a face ? ----------------------------------------------
			if(a_Entity.class == Sketchup::Face and a_Entity.layer.visible?)

				if a_bBrushCreated == false

					@m_pProcessorNode.BeginBrush(a_sName)
					a_bBrushCreated = true
				end

				a_pMaterial = nil
				if @m_bBlankFace == true
		
					a_pMaterial = a_Entity.material
				else
		
					a_pMaterial = a_Entity.back_material
				end
		
				@m_bMaterialsChanged = false
				if a_pMaterial != nil
		
					@m_pMaterials.push(a_pMaterial)
					@m_bMaterialsChanged = true
				end

				@m_pProcessorNode.ProcessFace(a_Entity, @m_pMaterials.last, @m_Transformations.last)

				if @m_bMaterialsChanged == true
		
					@m_pMaterials.pop
					@m_bMaterialsChanged = false;
				end

				@m_nFacesDone += 1
				@m_pWalker.Tick(@m_nLevel)
			end
		end
		if a_bBrushCreated == true

			@m_pProcessorNode.FinishBrush
			a_bBrushCreated = false
		end

		# --- scan the stuff into the array ----------------------------------------
		for a_Entity in pCurrentNode

			# ----- find a group ? ---------------------------------------------
			if(a_Entity.class == Sketchup::Group and a_Entity.layer.visible?)

				@m_nLevel += 1
				@m_Transformations.push(@m_Transformations.last * a_Entity.transformation)	# todo: shit ut99?
				self.Process(a_Entity, pCurrentNode)
				@m_Transformations.pop
				@m_nLevel -= 1

				@m_nGroupsDone += 1
			end

			# ----- find a component ? -----------------------------------------
			if(a_Entity.class == Sketchup::ComponentInstance and a_Entity.layer.visible?)

				@m_nLevel += 1
				@m_Transformations.push(@m_Transformations.last * a_Entity.transformation)
				self.Process(a_Entity, pCurrentNode)
				@m_Transformations.pop
				@m_nLevel -= 1

				@m_nComponentsDone += 1
			end

		end
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class FQuatPSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize(nX=0.0, nY=0.0, nZ=0.0, nW=1.0)

		@X = nX		# FLOAT
		@Y = nY		# FLOAT
		@Z = nZ		# FLOAT
		@W = nW		# FLOAT
	end

	#-----------------------------------------------------------------------------------

	def X

		return @X
	end

	#-----------------------------------------------------------------------------------

	def Y

		return @Y
	end

	#-----------------------------------------------------------------------------------

	def Z

		return @Z
	end

	#-----------------------------------------------------------------------------------

	def W

		return @W
	end

	#-----------------------------------------------------------------------------------

	def X=(value)

		@X = value
	end

	#-----------------------------------------------------------------------------------

	def Y=(value)

		@Y = value
	end

	#-----------------------------------------------------------------------------------

	def Z=(value)

		@Z = value
	end

	#-----------------------------------------------------------------------------------

	def W=(value)

		@W = value
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)

		# ----- FLOAT -----
		tmp = [@X].pack("e1")
		pFile.write tmp

		# ----- FLOAT -----
		tmp = [@Y].pack("e1")
		pFile.write tmp

		# ----- FLOAT -----
		tmp = [@Z].pack("e1")
		pFile.write tmp

		# ----- FLOAT -----
		tmp = [@W].pack("e1")
		pFile.write tmp
	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class FVectorPSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize(nX=0.0, nY=0.0, nZ=0.0)

		@X = nX		# FLOAT
		@Y = nY		# FLOAT
		@Z = nZ		# FLOAT
	end

	#-----------------------------------------------------------------------------------

	def X

		return @X
	end

	#-----------------------------------------------------------------------------------

	def Y

		return @Y
	end

	#-----------------------------------------------------------------------------------

	def Z

		return @Z
	end

	#-----------------------------------------------------------------------------------

	def X=(value)

		@X = value
	end

	#-----------------------------------------------------------------------------------

	def Y=(value)

		@Y = value
	end

	#-----------------------------------------------------------------------------------

	def Z=(value)

		@Z = value
	end

	#-----------------------------------------------------------------------------------

	def offsetfrom(loc)

		@X = @X - loc.X
		@Y = @Y - loc.Y
		@Z = @Z - loc.Z
	end

	#-----------------------------------------------------------------------------------

	def compare_search(other)

		if @X != other.X
			return 0
		end

		if @Y != other.Y
			return 0
		end

		if @Z != other.Z
			return 0
		end

		return 1
	end

	#-----------------------------------------------------------------------------------

	def dot(other)

		return ((@X * other.X) + (@Y * other.Y) + (@Z * other.Z))
	end

	#-----------------------------------------------------------------------------------

	def cross(other)

		nX = (@Y * other.Z) - (@Z * other.Y)
		nY = (@Z * other.X) - (@X * other.Z)
		nZ = (@X * other.Y) - (@Y * other.X)
		return FVector.new(nX,nY,nZ)
	end

	#-----------------------------------------------------------------------------------

	def substract(other)

		@X = @X - other.X
		@Y = @Y - other.Y
		@Z = @Z - other.Z
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)

		# ----- FLOAT -----
		tmp = [@X].pack("e1")
		pFile.write tmp

		# ----- FLOAT -----
		tmp = [@Y].pack("e1")
		pFile.write tmp

		# ----- FLOAT -----
		tmp = [@Z].pack("e1")
		pFile.write tmp
	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class VJointPosPSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize(loca)

		@Orientation = FQuatPSK_UTX.new()	# MISC
		@Position = loca		# MISC
		@Length = 0.0			# FLOAT
		@XSize = 0.0			# FLOAT
		@YSize = 0.0			# FLOAT
		@ZSize = 0.0			# FLOAT
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)

		@Orientation.export_bin(pFile)
		@Position.export_bin(pFile)
		tmp = [@Length,@XSize,@YSize,@ZSize].pack("e1e1e1e1")
		pFile.write tmp
	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class AnimInfoBinaryPSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize(aname="",bname="",bcount=0)

		@Name = aname			# STRING 32
		@MainBone = bname		# STRING 32
		@Group = "None"			# STRING 64
		@TotalBones = bcount		# INT
		@RootInclude = 0		# INT
		@KeyCompressionStyle = 0	# INT
		@KeyQuotum = 1			# INT
		@KeyPrediction = 1.0		# FLOAT
		@TrackTime = 1.0		# FLOAT
		@AnimRate = 30.0		# FLOAT
		@StartBone = 0			# INT
		@FirstRawFrame = 0		# INT
		@NumRawFrames = 1		# INT
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)

		# ----- STRING 32 -----
		tmp = [@Name].pack("a32")
		pFile.write tmp

		# ----- STRING 32 -----
		tmp = [@MainBone,0].pack("A31C1")
		pFile.write tmp

		# ----- STRING 32 -----
		tmp = [@Group].pack("a64")
		pFile.write tmp

		# ----- INT -----
		tmp = [@TotalBones.to_i].pack("i1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@RootInclude.to_i].pack("i1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@KeyCompressionStyle.to_i].pack("i1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@KeyQuotum.to_i].pack("i1")
		pFile.write tmp

		# ----- FLOAT -----
		tmp = [@KeyPrediction].pack("e1")
		pFile.write tmp

		# ----- FLOAT -----
		tmp = [@TrackTime].pack("e1")
		pFile.write tmp

		# ----- FLOAT -----
		tmp = [@AnimRate].pack("e1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@StartBone].pack("i1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@FirstRawFrame].pack("i1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@NumRawFrames].pack("i1")
		pFile.write tmp
		
	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class VChunkHeaderPSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize(name, type_size)

		@ChunkID = name 			# STRING 20
		@TypeFlag = 2003321			# INT CONST
		@DataSize = type_size			# INT
		@DataCount = 0				# INT
	end

	#-----------------------------------------------------------------------------------

	def DataCount

		return @DataCount
	end

	#-----------------------------------------------------------------------------------

	def DataCount=(value)

		@DataCount = value
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)

		# ----- STRING 20 -----
		tmp = [@ChunkID].pack("a20")
		pFile.write tmp

		# ----- INT -----
		tmp = [@TypeFlag.to_i].pack("i1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@DataSize.to_i].pack("i1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@DataCount.to_i].pack("i1")
		pFile.write tmp
	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class VMaterialPSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize(name="", idx=0)

		@MaterialName = name		# STRING 64
		@TextureIndex = idx		# INT
		@PolyFlags = 0			# DWORD
		@AuxMaterial = 0		# INT
		@AuxFlags = 0			# DWORD
		@LodBias = 0			# INT
		@LodStyle = 0			# INT
	end

	#-----------------------------------------------------------------------------------

	def compare_search(other)

		if other == @MaterialName
			return 1
		end

		return 0
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)

		# ----- STRING 64 -----
		tmp = [@MaterialName].pack("a64")
		pFile.write tmp

		# ----- INT -----
		tmp = [@TextureIndex].pack("i1")
		pFile.write tmp

		# ----- DWORD -----
		tmp = [@PolyFlags].pack("L1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@AuxMaterial.to_i].pack("i1")
		pFile.write tmp

		# ----- DWORD -----
		tmp = [@AuxFlags].pack("L1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@LodBias].pack("i1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@LodStyle].pack("i1")
		pFile.write tmp
	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class VBonePSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize(name="",parent=0,loca=nil)

		@Name = name				# STRING 64
		@Flags = 0				# DWORD
		@NumChildren = 0			# INT
		@ParentIndex = parent			# INT
		@BonePos = VJointPosPSK_UTX.new(loca)		# MISC
	end

	#-----------------------------------------------------------------------------------

	def object_name

		return @Name
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)

		tmp = [@Name,0,"",@Flags,@NumChildren,@ParentIndex].pack("A31C1a32L1i1i1")
		pFile.write tmp
		@BonePos.export_bin(pFile)
	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class FNamedBoneBinaryPSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize

		@Name = ""				# STRING 64
		@Flags = 0				# DWORD
		@NumChildren = 0			# INT
		@ParentIndex = 0			# INT
		@BonePos = VJointPosPSK_UTX.new(nil)		# MISC

		@IsRealBone = 0				# INT ?
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)

		# ----- STRING 64 -----
		tmp = [@Name].pack("A64")
		pFile.write tmp

		# ----- DWORD -----
		tmp = [@Flags].pack("L1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@NumChildren].pack("i1")
		pFile.write tmp

		# ----- INT -----
		tmp = [@ParentIndex].pack("i1")
		pFile.write tmp

		# ----- MISC -----
		BonePos.export_bin(pFile)
	end
	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class VRawBoneInfluencePSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize(point=0,bone=0)

		@Weight = 1.0		# FLOAT
		@PointIndex = point	# INT
		@BoneIndex = bone	# INT
	end

	#-----------------------------------------------------------------------------------

	def PointIndex

		return @PointIndex
	end

	#-----------------------------------------------------------------------------------

	def BoneIndex

		return @BoneIndex
	end

	#-----------------------------------------------------------------------------------

	def compare_search(other)

		if @PointIndex != other.PointIndex
			return 0
		end

		if @BoneIndex != other.BoneIndex
			return 0
		end

		return 1
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)

		tmp = [@Weight,@PointIndex,@BoneIndex].pack("e1i1i1")
		pFile.write tmp
	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class VQuatAnimKeyPSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize()

		@Position = FVectorPSK_UTX.new()		# MISC
		@Orientation = FQuatPSK_UTX.new()		# MISC
		@Time = 1.0				# FLOAT
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)

		# ----- MISC ----
		@Position.export_bin(pFile)

		# ----- MISC ----
		@Orientation.export_bin(pFile)

		# ----- FLOAT -----
		tmp = [@Time].pack("e1")
		pFile.write tmp
	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class VVertexPSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize(nMI=0,nPI=0,nU=0.0,nV=0.0)

		@PointIndex = nPI		# WORD
		@U = nU				# FLOAT
		@V = nV				# FLOAT
		@MatIndex = nMI			# BYTE
		@Reserved = 0			# BYTE
	end

	#-----------------------------------------------------------------------------------

	def PointIndex

		return @PointIndex
	end

	#-----------------------------------------------------------------------------------

	def U

		return @U
	end

	#-----------------------------------------------------------------------------------

	def V

		return @V
	end

	#-----------------------------------------------------------------------------------

	def MatIndex

		return @MatIndex
	end

	#-----------------------------------------------------------------------------------

	def compare_search(other)

		if @PointIndex != other.PointIndex
			return 0
		end

		if @U != other.U
			return 0
		end

		if @V != other.V
			return 0
		end

		if @MatIndex != other.MatIndex
			return 0
		end

		return 1
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)

		tmp = [@PointIndex,0,@U,@V,@MatIndex,@Reserved,0].pack("S1S1e1e1C1C1S1")
		pFile.write tmp

	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class VTrianglePSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize(nA=0,nB=0,nC=0)

		@WedgeIndex0 = nA		# WORD
		@WedgeIndex1 = nB		# WORD
		@WedgeIndex2 = nC		# WORD
		@MatIndex = 0			# BYTE
		@AuxMatIndex = 66		# BYTE
		@SmoothingGroups = 0		# DWORD
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)


		tmp = [@WedgeIndex0,@WedgeIndex1,@WedgeIndex2,@MatIndex,@AuxMatIndex,@SmoothingGroups].pack("S1S1S1C1C1L1")
		pFile.write tmp
	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class FileSectionPSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize(name, type_size)

		@Header = VChunkHeaderPSK_UTX.new(name, type_size)
		@Data = []

		@SearchFirst = 0;
	end

	#-----------------------------------------------------------------------------------

	def bookmark_search

		# we must only optimise for each face
		# instead of the whole thing ??? :(
		@SearchFirst = @Data.length
	end

	#-----------------------------------------------------------------------------------

	def search_name(name)

		if @Data.length > 0
			for i in (@SearchFirst..@Data.length-1)
				if @Data[i].object_name == name
					return i
				end
			end
		end

		return -1
	end

	#-----------------------------------------------------------------------------------

	def compare_search(other)

		if @Data.length > 0
			for i in (@SearchFirst..@Data.length-1)
				if @Data[i].compare_search(other) == 1
					return i
				end
			end
		end

		return -1
	end

	#-----------------------------------------------------------------------------------

	def push_that(pObject)

		@Data.push(pObject)

		return @Data.length - 1
	end

	#-----------------------------------------------------------------------------------

	def stack_length

		return @Data.length
	end

	#-----------------------------------------------------------------------------------

	def export_bin(pFile)

		# ----- UPDATE -----
		@Header.DataCount = @Data.length

		# ----- MISC -----
		@Header.export_bin(pFile)

		# ----- MISC -----
		if @Data.length > 0
			for i in (0..@Data.length-1)
				@Data[i].export_bin(pFile)
			end
		end
	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class UnrealFilePSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize

		# ----- psk shits -----
		@GeneralHeader = VChunkHeaderPSK_UTX.new("ACTRHEAD", 0)
		@Points = FileSectionPSK_UTX.new("PNTS0000", 12)
		@Wedges = FileSectionPSK_UTX.new("VTXW0000", 16)
		@Faces = FileSectionPSK_UTX.new("FACE0000", 12)
		@Materials = FileSectionPSK_UTX.new("MATT0000", 88)
		@Bones = FileSectionPSK_UTX.new("REFSKELT", 120)
		@Influences = FileSectionPSK_UTX.new("RAWWEIGHTS", 12)

		# ----- psa shits -----
		@AnimationHeader = VChunkHeaderPSK_UTX.new("ANIMHEAD", 0)
		@BonesName = FileSectionPSK_UTX.new("BONENAMES", 120)
		@Animations = FileSectionPSK_UTX.new("ANIMINFO", 168)
		@RawKeys = FileSectionPSK_UTX.new("ANIMKEYS", 32)
		@ScaleKey = VChunkHeaderPSK_UTX.new("SKALEKEYS", 0)
	end

	#-----------------------------------------------------------------------------------

	def push_point(pObject)

		@Points.push_that(pObject)
	end

	#-----------------------------------------------------------------------------------

	def push_wedge(pObject)

		@Wedges.push_that(pObject)
	end

	#-----------------------------------------------------------------------------------

	def push_face(pObject)

		@Faces.push_that(pObject)
	end

	#-----------------------------------------------------------------------------------

	def push_material(pObject)

		@Materials.push_that(pObject)
	end

	#-----------------------------------------------------------------------------------

	def push_bone(pObject)

		nI = @Bones.push_that(pObject)
		@BonesName.push_that(pObject)
		return nI
	end

	#-----------------------------------------------------------------------------------

	def search_bone(name)

		return @Bones.search_name(name)
	end

	#-----------------------------------------------------------------------------------

	def push_animation(pObject)

		@Animations.push_that(pObject)
	end

	#-----------------------------------------------------------------------------------

	def push_rawkey(pObject)

		@RawKeys.push_that(pObject)
	end

	#-----------------------------------------------------------------------------------

	def bookmark_face

		@Wedges.bookmark_search
	end

	#-----------------------------------------------------------------------------------

	def material_index(other)

		nI = @Materials.compare_search(other)
		if nI < 0
			nI = @Materials.stack_length
			nO = VMaterialPSK_UTX.new(other,nI)
			nI = @Materials.push_that(nO)
		end
		return nI
	end

	#-----------------------------------------------------------------------------------

	def point_index(other)

		nI = @Points.compare_search(other)
		if nI < 0
			nI = @Points.push_that(other)
		end
		return nI
	end

	#-----------------------------------------------------------------------------------

	def wedge_index(other)

		nI = @Wedges.compare_search(other)
		if nI < 0
			nI = @Wedges.push_that(other)
		end
		return nI
	end

	#-----------------------------------------------------------------------------------

	def influence_index(other)

		nI = @Influences.compare_search(other)
		if nI < 0
			nI = @Influences.push_that(other)
		end
		return nI
	end

	#-----------------------------------------------------------------------------------

	def export_psk(pFile)

		@GeneralHeader.export_bin(pFile)
		@Points.export_bin(pFile)
		@Wedges.export_bin(pFile)
		@Faces.export_bin(pFile)
		@Materials.export_bin(pFile)
		@Bones.export_bin(pFile)
		@Influences.export_bin(pFile)
	end

	#-----------------------------------------------------------------------------------

	def export_psa(pFile)

		nBones = @BonesName.stack_length
		pMainAnim = AnimInfoBinaryPSK_UTX.new("IdlePosition","Root",nBones);
		$pDestination.push_animation pMainAnim

		pRawKey = VQuatAnimKeyPSK_UTX.new()
		$pDestination.push_rawkey pRawKey

		@AnimationHeader.export_bin(pFile)
		@BonesName.export_bin(pFile)
		@Animations.export_bin(pFile)
		@RawKeys.export_bin(pFile)
		@ScaleKey.export_bin(pFile)
	end

	#===================================================================================
	#
	#===================================================================================
 
end

#===========================================================================================
#
#===========================================================================================

class BoneTamponPSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize

		@m_pListing = []
	end

	#-----------------------------------------------------------------------------------

	def compare_search(other)

		if @m_pListing.length > 0
			for i in (0..@m_pListing.length-1)
				pTemp = @m_pListing[i]
				if pTemp[1] == other
					return i
				end
			end
		end
		return -1
	end

	#-----------------------------------------------------------------------------------

	def push_stuff(sParentName,sName,vLocation)

		a_pTemp = [sParentName,sName,vLocation]
		@m_pListing.push(a_pTemp)
	end

	#-----------------------------------------------------------------------------------

	def process(fsys,sParentName,nParentId,vLocation)

		if @m_pListing.length > 0
			for i in (0..@m_pListing.length-1)
				a_pTemp = @m_pListing[i]
				if a_pTemp[0] == sParentName
					a_pTemp[2].offsetfrom(vLocation)
					a_pBone = VBonePSK_UTX.new(a_pTemp[1],nParentId,a_pTemp[2])
					a_nId = fsys.push_bone(a_pBone)
					if sParentName != a_pTemp[1]
						process(fsys,a_pTemp[1],a_nId,a_pTemp[2])
					end
				end
			end
		end
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class FaceTamponPSK_UTX
 
	#===================================================================================
	#
	#===================================================================================

	def initialize

		@m_pListing = []
	end

	#-----------------------------------------------------------------------------------

	def push_stuff(pFace,pTransformation,pMaterial,bGetBackNormal,pabone)

		a_pTemp = [pFace,pTransformation,pMaterial,bGetBackNormal,pabone]
		@m_pListing.push(a_pTemp)
	end

	#-----------------------------------------------------------------------------------

	def export_face(fsys,pFace,pTransformation,pMat,bGetBackNormal,nBonedID)

		# -- extract the mesh --
		pMesh = pFace.mesh 7

		# -- transform the mesh --
		pMesh.transform! pTransformation

		# -- extract material info --
		if pMat != nil
			pCurrentTexture = pMat.name
			if pCurrentTexture == nil
				pCurrentTexture = "Engine.DefaultTexture"
			end
		else
			pCurrentTexture = "Engine.DefaultTexture"
		end

		# -- get the index --
		nMatID = fsys.material_index(pCurrentTexture)

		# -- loop for each triangle in the face --
		nCurrentPolygon = 1
		while (nCurrentPolygon <= pMesh.count_polygons)

			# -- extract the triangle indexes --
			pPolygon = pMesh.polygon_at nCurrentPolygon

			# -- extract the triangle vertex --
			pTriA = pMesh.point_at pPolygon[0]
			pTriB = pMesh.point_at pPolygon[1]
			pTriC = pMesh.point_at pPolygon[2]

			# -- extract the texture coordinate --
			pTexA = pMesh.uv_at pPolygon[0],pPolygon[0]
			pTexA.y = 1.0 - pTexA.y

			pTexB = pMesh.uv_at pPolygon[1],pPolygon[1]
			pTexB.y = 1.0 - pTexB.y

			pTexC = pMesh.uv_at pPolygon[2],pPolygon[2]
			pTexC.y = 1.0 - pTexC.y

			# -- create the texture point with the correct order --
			pPtA = nil
			pPtB = nil
			pPtC = nil
			if bGetBackNormal
				pPtA = FVectorPSK_UTX.new(pTriA.x,-pTriA.y,pTriA.z)
				pPtB = FVectorPSK_UTX.new(pTriB.x,-pTriB.y,pTriB.z)
				pPtC = FVectorPSK_UTX.new(pTriC.x,-pTriC.y,pTriC.z)
			else
				pPtA = FVectorPSK_UTX.new(pTriB.x,-pTriB.y,pTriB.z)
				pPtB = FVectorPSK_UTX.new(pTriA.x,-pTriA.y,pTriA.z)
				pPtC = FVectorPSK_UTX.new(pTriC.x,-pTriC.y,pTriC.z)
			end

			# -- POINT A --
			nPtA = fsys.point_index(pPtA)					# register the point and get the index

			pWdA = VVertexPSK_UTX.new(nMatID,nPtA,pTexA.x,pTexA.y)		# make the wedge
			nWdA = fsys.wedge_index(pWdA)					# and get the index

			pBiA = VRawBoneInfluencePSK_UTX.new(nPtA,nBonedID)		# make the influence
			nBiA = fsys.influence_index(pBiA)				# and get the index

			# -- POINT B --
			nPtB = fsys.point_index(pPtB)

			pWdB = VVertexPSK_UTX.new(nMatID,nPtB,pTexB.x,pTexB.y)
			nWdB = fsys.wedge_index(pWdB)

			pBiB = VRawBoneInfluencePSK_UTX.new(nPtB,nBonedID)
			nBiB = fsys.influence_index(pBiB)

			# -- POINT C --
			nPtC = fsys.point_index(pPtC)

			pWdC = VVertexPSK_UTX.new(nMatID,nPtC,pTexC.x,pTexC.y)
			nWdC = fsys.wedge_index(pWdC)

			pBiC = VRawBoneInfluencePSK_UTX.new(nPtC,nBonedID)
			nBiC = fsys.influence_index(pBiC)

			# -- TRIANGLE --
			pTriV = VTrianglePSK_UTX.new(nWdA,nWdB,nWdC)			# make the triangle from wedge indexes
			fsys.push_face(pTriV)						# and push it
			
			# -- next triangle --
			nCurrentPolygon += 1

		end
	end

	#-----------------------------------------------------------------------------------

	def process(fsys)

		if @m_pListing.length > 0
			for i in (0..@m_pListing.length-1)
				a_pTemp = @m_pListing[i]
				a_nId = fsys.search_bone(a_pTemp[4])
				if a_nId >= 0
					self.export_face(fsys,a_pTemp[0],a_pTemp[1],a_pTemp[2],a_pTemp[3],a_nId)
				end
			end
		end
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class DataProcessorPSK_UTX

	#===================================================================================
	#
	#===================================================================================

	def initialize(sPskPath = "C:\utx.psk", sPsaPath = "C:\utx.psa", bBlankFace = true)

		@m_sPskPath = sPskPath
		@m_sPsaPath = sPsaPath
		@m_bBlankFace = bBlankFace

		@m_sBoneNames = []
		@m_sBoneNames.push("Root")
		@m_bBoneChanged = false

		@m_pTampBone = BoneTamponPSK_UTX.new()
		@m_pTampFace = FaceTamponPSK_UTX.new()

		@m_pFileOut = UnrealFilePSK_UTX.new()
	end

	#===================================================================================
	# Private
	#===================================================================================

	def SaveToFile(sPath)

		# --- create the file ------------------------------------------------------
		a_pFile = File.new(sPath,"wb")
		return if a_pFile == nil

		# --- dump to file ---------------------------------------------------------
		@m_pFileOut.export_psk(a_pFile)
		
		# --- close and flush the file ---------------------------------------------
		a_pFile.close
	end

	#===================================================================================
	# Public
	#===================================================================================

	def BeginWorld

		a_vLocation = FVectorPSK_UTX.new(0.0,0.0,0.0)
		@m_pTampBone.push_stuff("Root","Root",a_vLocation)
		
	end

	#-----------------------------------------------------------------------------------

	def BeginBrush(sConfig)

#		print "BeginBrush\r\n"

		if sConfig != nil

#			a_sConfig = sConfig.upcase
			a_sConfig = sConfig

			a_sStrSplit = a_sConfig.split(".")
			if a_sStrSplit[0] == "UTX_BONE"

				a_sNewBoneName = a_sStrSplit[1]
#				print a_sNewBoneName
#				print "\r\n"
				if a_sNewBoneName != nil

					@m_sBoneNames.push(a_sNewBoneName)
					@m_bBoneChanged = true
				end
			end
		end
	end

	#-----------------------------------------------------------------------------------

	def ProcessFace(pFace, pMaterial, pTransformations)

#		print "ProcessFace\r\n"

		@m_pTampFace.push_stuff(pFace,pTransformations,pMaterial,@m_bBlankFace,@m_sBoneNames.last)
	end

	#-----------------------------------------------------------------------------------

	def ProcessText(pText, pTransformations)

#		print "ProcessText\r\n"
#		print pText
#		print "\r\n"

		a_sConfig = pText.text
		if a_sConfig != nil

#			a_sConfig = a_sConfig.upcase
			a_sConfig = a_sConfig
			a_sStrSplit = a_sConfig.split(".")
			if a_sStrSplit[0] == "UTX_BONE"

				a_vLocation = pText.point.clone
				a_vLocation.transform! pTransformations
				a_vBoneLocation = FVectorPSK_UTX.new(a_vLocation.x, -a_vLocation.y, a_vLocation.z)
#				print a_sStrSplit[1]
#				print "\r\n"
				if a_sStrSplit[1] == nil

					return
				end
#				print a_sStrSplit[2]
#				print "\r\n"
				if a_sStrSplit[2] == nil

					return
				end
				@m_pTampBone.push_stuff(a_sStrSplit[1],a_sStrSplit[2],a_vBoneLocation)
			end
		end
	end

	#-----------------------------------------------------------------------------------

	def FinishBrush

		if @m_bBoneChanged == true
			
			@m_sBoneNames.pop
			@m_bBoneChanged = false
		end
	end

	#-----------------------------------------------------------------------------------

	def FinishWorld

		# --- compile bones --------------------------------------------------------
		a_vLocation = FVectorPSK_UTX.new(0.0,0.0,0.0)
		@m_pTampBone.process(@m_pFileOut,"Root",0,a_vLocation)

		# --- compile faces --------------------------------------------------------
		@m_pTampFace.process(@m_pFileOut)

		# --- shit to file ---------------------------------------------------------
		SaveToFile(@m_sPskPath)
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class TextureImporterOBJ_UTX

	#===================================================================================
	#
	#===================================================================================

	def initialize

		@m_pFile = nil

		@m_pMaterials = []
		@m_pMaterials.fill(["","",nil],0,100)
		@m_nNextMaterial = 0
		
	end

	#===================================================================================
	# Private
	#===================================================================================

	def NextMaterial(sMatName = nil, sMatPath = nil)

		# --- there is a name? ---------------------------------------------------------
		if sMatName == nil

			return nil
		end
#		print sMatName
#		print "\r\n"

		# --- try to find already existing material ------------------------------------
		a_nIndex = 0
		if @m_nNextMaterial > 0
			a_nIndex = @m_nNextMaterial - 1
			while a_nIndex > -1
	
				a_pPacketA = @m_pMaterials[a_nIndex]
				if a_pPacketA[0] == sMatName
	
					return a_pPacketA[2]
				end
	
				a_nIndex = a_nIndex - 1
			end
		end

		# --- create a new material ----------------------------------------------------
		a_pModel = Sketchup.active_model
		a_pMaterials = a_pModel.materials
		a_pNewMat = a_pMaterials.add(sMatName)

		# --- there is a file path? ----------------------------------------------------
		if sMatPath != nil

			a_pNewMat.texture = sMatPath
		end

		# --- return the new material --------------------------------------------------
		@m_pMaterials[@m_nNextMaterial] = [sMatName, sMatPath, a_pNewMat]
		@m_nNextMaterial = @m_nNextMaterial + 1
		return a_pNewMat
	end

	#-----------------------------------------------------------------------------------

	def UnderscoreSpaces(sFileName)

		a_nIndex = sFileName.length - 1
		while a_nIndex > 0

			if sFileName[a_nIndex].chr == ' '

				sFileName[a_nIndex] = '_'
			end
			a_nIndex = a_nIndex - 1
		end
	end

	#-----------------------------------------------------------------------------------

	def RecomposeStringToEnd(nIndex,pPacket)

		a_sString = ""
		a_nCount = pPacket.length
		while nIndex < a_nCount

			a_sString = a_sString + pPacket[nIndex]

			nIndex = nIndex + 1
			if nIndex < a_nCount

				a_sString = a_sString + " "
			end
		end

		return a_sString
	end

	#===================================================================================
	# Public
	#===================================================================================

	def Process(sMtlPath = nil)

		@m_pFile = File.new(sMtlPath,"rb")

		a_sMatName = nil
		a_sMatPath = nil
		
		while @m_pFile.eof? != true

			a_sTextLine = @m_pFile.readline
			a_pTextSplit = a_sTextLine.split(" ")

			if a_pTextSplit[0] == "newmtl"

				if a_sMatName != nil

					NextMaterial(a_sMatName, a_sMatPath)
				end
				a_sMatName = self.RecomposeStringToEnd(1,a_pTextSplit)
				a_sMatName = a_sMatName.gsub("\"","")
				self.UnderscoreSpaces(a_sMatName)
#				print a_sMatName
#				print "\r\n"
			end

			if a_pTextSplit[0] == "map_Kd"

				a_sMatPath = self.RecomposeStringToEnd(1,a_pTextSplit)
				a_sMatPath = a_sMatPath.gsub("\"","")
			end

		end

		if a_sMatName != nil

			NextMaterial(a_sMatName, a_sMatPath)
		end

		@m_pFile.close

		code = Sketchup.set_status_text "Job Done !!!"
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class FileImporterOBJ_UTX

	#===================================================================================
	#
	#===================================================================================

	def initialize(sObjPath = nil, sMtlPath = nil)

		@m_pFile = nil
		@m_sObjPath = sObjPath
		@m_sMtlPath = sMtlPath

		@m_pCurrentModel = nil

		@m_vVertex = []
		@m_vVertex.fill([0.0,0.0,0.0],0,65000)
		@m_nNextVertex = 0

		@m_vVertexUV = []
		@m_vVertexUV.fill([0.0,0.0],0,65000)
		@m_nNextVertexUV = 0

		@m_vNormal = []
		@m_vNormal.fill([0.0,0.0,0.0],0,65000)
		@m_nNextNormal = 0

		@m_nCreatedFaces = 0
		@m_pFaceMaterial = []

		@m_pCurrentTex = nil
		@m_TexImporter = TextureImporterOBJ_UTX.new
	end

	#===================================================================================
	# Private
	#===================================================================================

	def NextModel(sName = nil)

		a_pModel = Sketchup.active_model
		a_pEntities = a_pModel.active_entities
		@m_pCurrentModel = a_pEntities.add_group
		@m_pCurrentModel.description = sName
		@m_pCurrentModel.name = sName

	end

	#-----------------------------------------------------------------------------------

	def NextVertexXYZ(nX, nY, nZ)

		a_vTemp = [nX, nY, nZ]
		@m_vVertex[@m_nNextVertex] = a_vTemp
		@m_nNextVertex = @m_nNextVertex + 1
	end

	#-----------------------------------------------------------------------------------

	def GetVertexXYZ(nIndex)

		if nIndex > 0

			return @m_vVertex[nIndex - 1]
		end

		return @m_vVertex[@m_nNextVertex + nIndex]
	end

	#-----------------------------------------------------------------------------------

	def NextVertexUV(nX, nY)

		a_vTemp = [nX, nY]
		@m_vVertexUV[@m_nNextVertexUV] = a_vTemp
		@m_nNextVertexUV = @m_nNextVertexUV + 1
	end

	#-----------------------------------------------------------------------------------

	def GetVertexUV(nIndex)

		if @m_nNextVertexUV < 1

			return [0.0, 0.0]
		end
		if nIndex > 0

			return @m_vVertexUV[nIndex - 1]
		end

		return @m_vVertexUV[@m_nNextVertexUV + nIndex]
	end

	#-----------------------------------------------------------------------------------

	def NextNormalXYZ(nX, nY, nZ)

		a_vTemp = [nX, nY, nZ]
		@m_vNormal[@m_nNextNormal] = a_vTemp
		@m_nNextNormal = @m_nNextNormal + 1
	end

	#-----------------------------------------------------------------------------------

	def GetNormalXYZ(nIndex)

		if nIndex > 0

			return @m_vNormal[nIndex - 1]
		end

		return @m_vNormal[@m_nNextNormal + nIndex]
	end

	#-----------------------------------------------------------------------------------

	def NextFaceXYZ(pTableVTN)

		a_pPoints = []
		a_pPointsUV = []

		a_nIndex = 0
		a_nCount = pTableVTN.length

		while a_nIndex < a_nCount

			a_pSingleVTN = pTableVTN[a_nIndex]

			a_vVertex = self.GetVertexXYZ(a_pSingleVTN[0])
			a_pPoints[a_pPoints.length] = a_vVertex

			a_pPointsUV[a_pPointsUV.length] = a_vVertex
			a_pPointsUV[a_pPointsUV.length] = self.GetVertexUV(a_pSingleVTN[1])

			a_nIndex += 1
		end

		begin
			a_pNewFace = @m_pCurrentModel.entities.add_face a_pPoints
			if @m_pCurrentTex != nil

				a_pNewFace.position_material @m_pCurrentTex, a_pPointsUV, true 
			end
		rescue => e
#			print e
#			print "\r\n"
#			print a_pPoints.to_s
#			print "\r\n"
		end 

		@m_nCreatedFaces = @m_nCreatedFaces + 1
		code = Sketchup.set_status_text ("Current Faces = " + @m_nCreatedFaces.to_s)
	end

	#-----------------------------------------------------------------------------------

	def NextMaterial(sName = nil, sFilePath = nil)

		
	end

	#-----------------------------------------------------------------------------------

	def OverwriteBasePath(sObjPath, sMtlPath)

		# --- find the length of the lower segment path --------------------------------
		a_nIndex = sObjPath.length - 1
		while a_nIndex > 0

			if sObjPath[a_nIndex].chr == '\\'

				break
			end
			a_nIndex = a_nIndex - 1
		end
		if a_nIndex < 1

			return sMtlPath
		end

		# --- extract base path --------------------------------------------------------
		a_sBasePath = ""
		a_nIndexB = 0
		while a_nIndexB <= a_nIndex

			a_sBasePath = a_sBasePath + sObjPath[a_nIndexB].chr
			a_nIndexB = a_nIndexB + 1
		end

		# --- return new path ----------------------------------------------------------
		return (a_sBasePath + sMtlPath)
	end

	#-----------------------------------------------------------------------------------

	def UnderscoreSpaces(sFileName)

		a_nIndex = sFileName.length - 1
		while a_nIndex > 0

			if sFileName[a_nIndex].chr == ' '

				sFileName[a_nIndex] = '_'
			end
			a_nIndex = a_nIndex - 1
		end
	end

	#-----------------------------------------------------------------------------------

	def RecomposeStringToEnd(nIndex,pPacket)

		a_sString = ""
		a_nCount = pPacket.length
		while nIndex < a_nCount

			a_sString = a_sString + pPacket[nIndex]

			nIndex = nIndex + 1
			if nIndex < a_nCount

				a_sString = a_sString + " "
			end
		end

		return a_sString
	end

	#===================================================================================
	# Public
	#===================================================================================

	def Start

		@m_pFile = File.new(@m_sObjPath,"rb")
		
		while @m_pFile.eof? != true

			a_sTextLine = @m_pFile.readline
			a_pTextSplit = a_sTextLine.split(" ")

			if a_pTextSplit[0] == "g"

				NextModel(a_pTextSplit[1])
			end

			if a_pTextSplit[0] == "mtllib"

				a_sTemp = self.RecomposeStringToEnd(1,a_pTextSplit)
				a_sTemp = a_sTemp.gsub("\"","")
				a_sTemp = self.OverwriteBasePath(@m_sObjPath, a_sTemp)
				@m_TexImporter.Process(a_sTemp)
			end

			if a_pTextSplit[0] == "usemtl"

				a_sTemp = self.RecomposeStringToEnd(1,a_pTextSplit)
				a_sTemp = a_sTemp.gsub("\"","")
				self.UnderscoreSpaces(a_sTemp)
#				print a_sTemp
#				print "\r\n"
				@m_pCurrentTex = @m_TexImporter.NextMaterial(a_sTemp)
			end

			if a_pTextSplit[0] == "v"

				NextVertexXYZ(a_pTextSplit[1].to_f, a_pTextSplit[2].to_f, a_pTextSplit[3].to_f)
			end

			if a_pTextSplit[0] == "vt"

				NextVertexUV(a_pTextSplit[1].to_f, a_pTextSplit[2].to_f)
			end

			if a_pTextSplit[0] == "vn"

				NextNormalXYZ(a_pTextSplit[1].to_f, a_pTextSplit[2].to_f, a_pTextSplit[3].to_f)
			end

			if a_pTextSplit[0] == "f"

				a_pTempVTN = []
				for a_nIndexA in 1..a_pTextSplit.length-1

					a_sTemp = a_pTextSplit[a_nIndexA]
					a_pSplit = a_sTemp.split("/")
					a_pBlockVTN = []
					for a_nIndexB in 0..a_pSplit.length-1

						a_pBlockVTN[a_nIndexB] = a_pSplit[a_nIndexB].to_i
					end
					if a_pBlockVTN.length > 0
	
						a_pTempVTN[a_pTempVTN.length] = a_pBlockVTN
					end
				end	
				if a_pTempVTN.length > 0

					NextFaceXYZ(a_pTempVTN)
				end
			end

		end

		@m_pFile.close

		code = Sketchup.set_status_text "Job Done !!!"
	end

	#-----------------------------------------------------------------------------------

	def Debug

		NextVertexXYZ(0.0,0.0,0.0)
		NextVertexXYZ(1.0,0.0,0.0)
		NextVertexXYZ(1.0,1.0,0.0)
		NextVertexXYZ(0.0,1.0,0.0)

		aaa = []
		aaa.push([1,0])
		aaa.push([2,0])
		aaa.push([3,0])
		aaa.push([4,0])

		NextModel("toto")
		NextFaceXYZ(aaa)
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class FileImporterT3DSM_UTX

	#===================================================================================
	#
	#===================================================================================

	def initialize(sObjPath = nil, sMtlPath = nil)

		@m_pFile = nil
		@m_sObjPath = sObjPath

		if @m_sObjPath != nil

			@m_sMtlPath = GetBasePath(@m_sObjPath)
#			puts @m_sMtlPath
		else

			@m_sMtlPath = nil
		end

		@m_pCurrentModel = nil

		@m_vVertex = []
		@m_vVertex.fill([0.0,0.0,0.0],0,65000)
		@m_nNextVertex = 0

		@m_vVertexUV = []
		@m_vVertexUV.fill([0.0,0.0],0,65000)
		@m_nNextVertexUV = 0

		@m_nCreatedFaces = 0
		@m_pFaceMaterial = []

		@m_pCurrentTex = nil
		@m_TexImporter = TextureImporterOBJ_UTX.new
	end

	#===================================================================================
	# Private
	#===================================================================================

	def NextModel(sName = nil)

		a_pModel = Sketchup.active_model
		a_pEntities = a_pModel.active_entities
		@m_pCurrentModel = a_pEntities.add_group
		@m_pCurrentModel.description = sName
		@m_pCurrentModel.name = sName

	end

	#-----------------------------------------------------------------------------------

	def NextVertexXYZ(nX, nY, nZ)

		a_vTemp = [nX, nY, nZ]
		@m_vVertex[@m_nNextVertex] = a_vTemp
		@m_nNextVertex = @m_nNextVertex + 1

		return @m_nNextVertex
	end

	#-----------------------------------------------------------------------------------

	def GetVertexXYZ(nIndex)

		if nIndex > 0

			return @m_vVertex[nIndex - 1]
		end

		return @m_vVertex[@m_nNextVertex + nIndex]
	end

	#-----------------------------------------------------------------------------------

	def NextVertexUV(nX, nY)

		a_vTemp = [nX, nY]
		@m_vVertexUV[@m_nNextVertexUV] = a_vTemp
		@m_nNextVertexUV = @m_nNextVertexUV + 1

		return @m_nNextVertexUV
	end

	#-----------------------------------------------------------------------------------

	def GetVertexUV(nIndex)

		if @m_nNextVertexUV < 1

			return [0.0, 0.0]
		end
		if nIndex > 0

			return @m_vVertexUV[nIndex - 1]
		end

		return @m_vVertexUV[@m_nNextVertexUV + nIndex]
	end

	#-----------------------------------------------------------------------------------

	def NextMaterial(sMatName = nil, sMatPath = nil)

#		puts sMatPath + sMatName + ".bmp"

		# --- there is a name? ---------------------------------------------------------
		if sMatName == nil

			return nil
		end
#		print sMatName
#		print "\r\n"

		# --- create a new material ----------------------------------------------------
		a_pModel = Sketchup.active_model
		a_pMaterials = a_pModel.materials
		a_pNewMat = a_pMaterials.add(sMatName)

		# --- there is a file path? ----------------------------------------------------
		if sMatPath != nil

			a_pNewMat.texture = sMatPath + sMatName + ".bmp"
		end

		# --- return the new material --------------------------------------------------
		@m_pCurrentTex = a_pNewMat

		return a_pNewMat
	end

	#-----------------------------------------------------------------------------------

	def NextFaceXYZ(pTableVTN)

		a_pPoints = []
		a_pPointsUV = []

		a_nIndex = 0
		a_nCount = pTableVTN.length

		while a_nIndex < a_nCount

			a_pSingleVTN = pTableVTN[a_nIndex]

			a_vVertex = self.GetVertexXYZ(a_pSingleVTN[0])
			a_pPoints[a_pPoints.length] = a_vVertex

			a_pPointsUV[a_pPointsUV.length] = a_vVertex
			a_pPointsUV[a_pPointsUV.length] = self.GetVertexUV(a_pSingleVTN[1])

			a_nIndex += 1
		end

		begin
			a_pNewFace = @m_pCurrentModel.entities.add_face a_pPoints
			if @m_pCurrentTex != nil

				a_pNewFace.position_material @m_pCurrentTex, a_pPointsUV, true 
			end
		rescue => e
#			print e
#			print "\r\n"
#			print a_pPoints.to_s
#			print "\r\n"
		end 

		@m_nCreatedFaces = @m_nCreatedFaces + 1
		code = Sketchup.set_status_text ("Current Faces = " + @m_nCreatedFaces.to_s)
	end

	#-----------------------------------------------------------------------------------

	def GetBasePath(sObjPath)

		# --- find the length of the lower segment path --------------------------------
		a_nIndex = sObjPath.length - 1
		while a_nIndex > 0

			if sObjPath[a_nIndex].chr == '\\'

				break
			end
			a_nIndex = a_nIndex - 1
		end
		if a_nIndex < 1

			return nil
		end

		# --- extract base path --------------------------------------------------------
		a_sBasePath = ""
		a_nIndexB = 0
		while a_nIndexB <= a_nIndex

			a_sBasePath = a_sBasePath + sObjPath[a_nIndexB].chr
			a_nIndexB = a_nIndexB + 1
		end

		# --- return new path ----------------------------------------------------------
		return (a_sBasePath)
	end

	#-----------------------------------------------------------------------------------

	def OverwriteBasePath(sObjPath, sMtlPath)

		# --- find the length of the lower segment path --------------------------------
		a_nIndex = sObjPath.length - 1
		while a_nIndex > 0

			if sObjPath[a_nIndex].chr == '\\'

				break
			end
			a_nIndex = a_nIndex - 1
		end
		if a_nIndex < 1

			return sMtlPath
		end

		# --- extract base path --------------------------------------------------------
		a_sBasePath = ""
		a_nIndexB = 0
		while a_nIndexB <= a_nIndex

			a_sBasePath = a_sBasePath + sObjPath[a_nIndexB].chr
			a_nIndexB = a_nIndexB + 1
		end

		# --- return new path ----------------------------------------------------------
		return (a_sBasePath + sMtlPath)
	end

	#-----------------------------------------------------------------------------------

	def UnderscoreSpaces(sFileName)

		a_nIndex = sFileName.length - 1
		while a_nIndex > 0

			if sFileName[a_nIndex].chr == ' '

				sFileName[a_nIndex] = '_'
			end
			a_nIndex = a_nIndex - 1
		end
	end

	#-----------------------------------------------------------------------------------

	def RecomposeStringToEnd(nIndex,pPacket)

		a_sString = ""
		a_nCount = pPacket.length
		while nIndex < a_nCount

			a_sString = a_sString + pPacket[nIndex]

			nIndex = nIndex + 1
			if nIndex < a_nCount

				a_sString = a_sString + " "
			end
		end

		return a_sString
	end

	#===================================================================================
	# Public
	#===================================================================================

	def Start

		@m_pFile = File.new(@m_sObjPath,"rb")
		
		pTableVTN = []
		pTableVTN[0] = [0,0,0]
		pTableVTN[1] = [0,0,0]
		pTableVTN[2] = [0,0,0]

		self.NextModel("StaticMeshes");

		while @m_pFile.eof? != true

			a_sTextLine = @m_pFile.readline
			a_pTextSplit = a_sTextLine.split(" ")

			if a_pTextSplit[0] == "Texture"

				a_TxNameFull = a_pTextSplit[1]
				a_TxNameSplited = a_TxNameFull.split(".")
				if a_TxNameSplited.length > 1
					a_TxName = a_TxNameSplited[a_TxNameSplited.length - 1]
				else
					a_TxName = "DefaultMaterial"
				end
#				puts a_TxName
				@m_pCurrentTex = @m_TexImporter.NextMaterial(a_TxName, @m_sMtlPath + a_TxName + ".bmp")
			end

			if a_pTextSplit[0] == "Vertex"

				if a_pTextSplit[1] == "0"

					pTableVTN[2] = [self.NextVertexXYZ(a_pTextSplit[2].to_f, a_pTextSplit[3].to_f, a_pTextSplit[4].to_f),self.NextVertexUV(a_pTextSplit[5].to_f, a_pTextSplit[6].to_f),0]
				end

				if a_pTextSplit[1] == "1"

					pTableVTN[1] = [self.NextVertexXYZ(a_pTextSplit[2].to_f, a_pTextSplit[3].to_f, a_pTextSplit[4].to_f),self.NextVertexUV(a_pTextSplit[5].to_f, a_pTextSplit[6].to_f),0]
				end

				if a_pTextSplit[1] == "2"

					pTableVTN[0] = [self.NextVertexXYZ(a_pTextSplit[2].to_f, a_pTextSplit[3].to_f, a_pTextSplit[4].to_f),self.NextVertexUV(a_pTextSplit[5].to_f, a_pTextSplit[6].to_f),0]
					NextFaceXYZ(pTableVTN)
				end
			end

		end

		@m_pFile.close

		code = Sketchup.set_status_text "Job Done !!!"
	end

	#-----------------------------------------------------------------------------------

	def Debug

		NextVertexXYZ(0.0,0.0,0.0)
		NextVertexXYZ(1.0,0.0,0.0)
		NextVertexXYZ(1.0,1.0,0.0)
		NextVertexXYZ(0.0,1.0,0.0)

		aaa = []
		aaa.push([1,0])
		aaa.push([2,0])
		aaa.push([3,0])
		aaa.push([4,0])

		NextModel("toto")
		NextFaceXYZ(aaa)
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class FileImporterT3D_UTX

	#===================================================================================
	#
	#===================================================================================

	def initialize(sPath = nil)

		@m_sPath = sPath
		@m_pParser = ParserT3D_UTX.new()
		@m_pTranslator = PolygonTranslationToSkpT3D_UTX.new()
	end

	#===================================================================================
	# Private
	#===================================================================================



	#===================================================================================
	# Public
	#===================================================================================

	def Start

		code = Sketchup.set_status_text "Job Started !!!"

		@m_pParser.ImportT3D(@m_sPath)
		@m_pParser.TranslateTheKeys(@m_pTranslator)

		code = Sketchup.set_status_text "Job Done !!!"
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class DataProcessor_Align_TEX

	#===================================================================================
	#
	#===================================================================================

	def initialize(nXs = 1.0, nYs = 1.0, nXo = 0.0, nYo = 0.0, bAlignType = 0)

		@m_nXs = nXs
		@m_nYs = nYs
		@m_nXo = nXo
		@m_nYo = nYo
		@m_bAlignType = bAlignType

	end

	#===================================================================================
	# Public
	#===================================================================================

	def BeginWorld


		
	end

	#-----------------------------------------------------------------------------------

	def BeginBrush(sConfig)


	end

	#-----------------------------------------------------------------------------------

	def ProcessFace(pFace, pMaterial, pTransformations)

		# --- extract the mesh -----------------------------------------------------
		a_pMesh = pFace.mesh 7

		# --- transform the mesh ---------------------------------------------------
		a_pMesh.transform! pTransformations

		# --- extract material info ------------------------------------------------
		a_nMaterialU = 1.0
		a_nMaterialV = 1.0
		a_sTexture = "Engine.DefaultTexture"
		if pMaterial != nil
			a_sTexture = pMaterial.name
			if a_sTexture == nil
				a_sTexture = "Engine.DefaultTexture"
			end
			a_pTexture = pMaterial.texture
			if a_pTexture != nil
				if a_pTexture.width != 0.0
					a_nMaterialU = a_pTexture.width
				else
					a_nMaterialU = 1.0
				end
				if a_pTexture.height != 0.0
					a_nMaterialV = a_pTexture.height
				else
					a_nMaterialV = 1.0
				end
			end
		end
		a_nMaterialU *= 0.125
		a_nMaterialV *= 0.125

		# --- loop for each triangle in the face -----------------------------------
		a_nIndex = 1
		while (a_nIndex <= a_pMesh.count_polygons)

			# --- extract the triangle indexes ---------------------------------
			a_pPolygon = a_pMesh.polygon_at a_nIndex

			# --- extract the triangle vertex ----------------------------------
			a_vA = VertexT3D_UTX.FromSketchup(a_pMesh.point_at a_pPolygon[0])
			a_vB = VertexT3D_UTX.FromSketchup(a_pMesh.point_at a_pPolygon[1])
			a_vC = VertexT3D_UTX.FromSketchup(a_pMesh.point_at a_pPolygon[2])

			# --- extract the texture coordinate -------------------------------
			a_tA = VertexT3D_UTX.FromSketchup(a_pMesh.uv_at a_pPolygon[0],a_pPolygon[0])
			a_tB = VertexT3D_UTX.FromSketchup(a_pMesh.uv_at a_pPolygon[1],a_pPolygon[1])
			a_tC = VertexT3D_UTX.FromSketchup(a_pMesh.uv_at a_pPolygon[2],a_pPolygon[2])

			# --- transform the texture coordinate -----------------------------
			if pMaterial != nil

				a_tA.x = a_tA.x + @m_nXo
				a_tA.y = a_tA.y + @m_nYo

				a_tB.x = a_tB.x + @m_nXo
				a_tB.y = a_tB.y + @m_nYo

				a_tC.x = a_tC.x + @m_nXo
				a_tC.y = a_tC.y + @m_nYo

				a_tA.x = a_tA.x * @m_nXs
				a_tA.y = a_tA.y * @m_nYs

				a_tB.x = a_tB.x * @m_nXs
				a_tB.y = a_tB.y * @m_nYs

				a_tC.x = a_tC.x * @m_nXs
				a_tC.y = a_tC.y * @m_nYs

				pts_uv = []
				pts_uv[0] = [a_vA.x, a_vA.y, a_vA.z]
				pts_uv[1] = [a_tA.x, a_tA.y]
				pts_uv[2] = [a_vB.x, a_vB.y, a_vB.z]
				pts_uv[3] = [a_tB.x, a_tB.y]
				pts_uv[4] = [a_vC.x, a_vC.y, a_vC.z]
				pts_uv[5] = [a_tC.x, a_tC.y]
	
				pFace.position_material pMaterial, pts_uv, true
			end

			# --- next triangle ------------------------------------------------
			a_nIndex += 1
		end
	end

	#-----------------------------------------------------------------------------------

	def ProcessText(pText, pTransformations)

		

	end

	#-----------------------------------------------------------------------------------

	def FinishBrush



	end

	#-----------------------------------------------------------------------------------

	def FinishWorld



	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class HSCALER

	#===================================================================================
	#
	#===================================================================================

	def HSCALER::ScaleV_YYY_TEX(nY = 1.0)

		a_bBlankFace = true

		a_pProcessor = DataProcessor_Align_TEX.new(1.0, 1.0 / nY, 0.0, 0.0, 0)

		a_pEntities = Sketchup.active_model.active_entities

		a_pIteratorer = DataIteratorerT3D_UTX.new(a_pEntities, a_pProcessor, a_bBlankFace)
		a_pIteratorer.Start(false)
	end

	#===================================================================================
	#
	#===================================================================================

	def HSCALER::ScaleV_p2_TEX

		HSCALER.ScaleV_YYY_TEX(2.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_p4_TEX

		HSCALER.ScaleV_YYY_TEX(4.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_p8_TEX

		HSCALER.ScaleV_YYY_TEX(8.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_p16_TEX

		HSCALER.ScaleV_YYY_TEX(16.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_p32_TEX

		HSCALER.ScaleV_YYY_TEX(32.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_p64_TEX

		HSCALER.ScaleV_YYY_TEX(64.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_p128_TEX

		HSCALER.ScaleV_YYY_TEX(128.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_p256_TEX

		HSCALER.ScaleV_YYY_TEX(256.0)
	end

	#===================================================================================
	#
	#===================================================================================

	def HSCALER::ScaleV_i2_TEX

		HSCALER.ScaleV_YYY_TEX(1.0 / 2.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_i4_TEX

		HSCALER.ScaleV_YYY_TEX(1.0 / 4.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_i8_TEX

		HSCALER.ScaleV_YYY_TEX(1.0 / 8.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_i16_TEX

		HSCALER.ScaleV_YYY_TEX(1.0 / 16.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_i32_TEX

		HSCALER.ScaleV_YYY_TEX(1.0 / 32.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_i64_TEX

		HSCALER.ScaleV_YYY_TEX(1.0 / 64.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_i128_TEX

		HSCALER.ScaleV_YYY_TEX(1.0 / 128.0)
	end

	#===================================================================================

	def HSCALER::ScaleV_i256_TEX

		HSCALER.ScaleV_YYY_TEX(1.0 / 256.0)
	end

	#===================================================================================
	#
	#===================================================================================

	def HSCALER::ScaleU_XXX_TEX(nX = 1.0)

		a_bBlankFace = true

		a_pProcessor = DataProcessor_Align_TEX.new(1.0 / nX, 1.0, 0.0, 0.0, 0)

		a_pEntities = Sketchup.active_model.active_entities

		a_pIteratorer = DataIteratorerT3D_UTX.new(a_pEntities, a_pProcessor, a_bBlankFace)
		a_pIteratorer.Start(false)
	end

	#===================================================================================
	#
	#===================================================================================

	def HSCALER::ScaleU_p2_TEX

		HSCALER.ScaleU_XXX_TEX(2.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_p4_TEX

		HSCALER.ScaleU_XXX_TEX(4.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_p8_TEX

		HSCALER.ScaleU_XXX_TEX(8.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_p16_TEX

		HSCALER.ScaleU_XXX_TEX(16.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_p32_TEX

		HSCALER.ScaleU_XXX_TEX(32.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_p64_TEX

		HSCALER.ScaleU_XXX_TEX(64.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_p128_TEX

		HSCALER.ScaleU_XXX_TEX(128.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_p256_TEX

		HSCALER.ScaleU_XXX_TEX(256.0)
	end

	#===================================================================================
	#
	#===================================================================================

	def HSCALER::ScaleU_i2_TEX

		HSCALER.ScaleU_XXX_TEX(1.0 / 2.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_i4_TEX

		HSCALER.ScaleU_XXX_TEX(1.0 / 4.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_i8_TEX

		HSCALER.ScaleU_XXX_TEX(1.0 / 8.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_i16_TEX

		HSCALER.ScaleU_XXX_TEX(1.0 / 16.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_i32_TEX

		HSCALER.ScaleU_XXX_TEX(1.0 / 32.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_i64_TEX

		HSCALER.ScaleU_XXX_TEX(1.0 / 64.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_i128_TEX

		HSCALER.ScaleU_XXX_TEX(1.0 / 128.0)
	end

	#===================================================================================

	def HSCALER::ScaleU_i256_TEX

		HSCALER.ScaleU_XXX_TEX(1.0 / 256.0)
	end

	#===================================================================================
	#
	#===================================================================================

	def HSCALER::ScaleUV_XXX_TEX(nK = 1.0)

		a_bBlankFace = true

		a_pProcessor = DataProcessor_Align_TEX.new(1.0 / nK, 1.0 / nK, 0.0, 0.0, 0)

		a_pEntities = Sketchup.active_model.active_entities

		a_pIteratorer = DataIteratorerT3D_UTX.new(a_pEntities, a_pProcessor, a_bBlankFace)
		a_pIteratorer.Start(false)
	end

	#===================================================================================
	#
	#===================================================================================

	def HSCALER::ScaleUV_p2_TEX

		HSCALER.ScaleUV_XXX_TEX(2.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_p4_TEX

		HSCALER.ScaleUV_XXX_TEX(4.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_p8_TEX

		HSCALER.ScaleUV_XXX_TEX(8.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_p16_TEX

		HSCALER.ScaleUV_XXX_TEX(16.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_p32_TEX

		HSCALER.ScaleUV_XXX_TEX(32.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_p64_TEX

		HSCALER.ScaleUV_XXX_TEX(64.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_p128_TEX

		HSCALER.ScaleUV_XXX_TEX(128.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_p256_TEX

		HSCALER.ScaleUV_XXX_TEX(256.0)
	end

	#===================================================================================
	#
	#===================================================================================

	def HSCALER::ScaleUV_i2_TEX

		HSCALER.ScaleUV_XXX_TEX(1.0 / 2.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_i4_TEX

		HSCALER.ScaleUV_XXX_TEX(1.0 / 4.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_i8_TEX

		HSCALER.ScaleUV_XXX_TEX(1.0 / 8.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_i16_TEX

		HSCALER.ScaleUV_XXX_TEX(1.0 / 16.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_i32_TEX

		HSCALER.ScaleUV_XXX_TEX(1.0 / 32.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_i64_TEX

		HSCALER.ScaleUV_XXX_TEX(1.0 / 64.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_i128_TEX

		HSCALER.ScaleUV_XXX_TEX(1.0 / 128.0)
	end

	#===================================================================================

	def HSCALER::ScaleUV_i256_TEX

		HSCALER.ScaleUV_XXX_TEX(1.0 / 256.0)
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class HPANNER

	#===================================================================================
	#
	#===================================================================================

	def HPANNER::PanU_XXX_TEX(nX = 0.0)

		a_bBlankFace = true

		a_pProcessor = DataProcessor_Align_TEX.new(1.0, 1.0, nX, 0.0, 0)

		a_pEntities = Sketchup.active_model.active_entities

		a_pIteratorer = DataIteratorerT3D_UTX.new(a_pEntities, a_pProcessor, a_bBlankFace)
		a_pIteratorer.Start(false)
	end

	#===================================================================================
	#
	#===================================================================================

	def HPANNER::PanU_p2_TEX

		HPANNER.PanU_XXX_TEX(1.0 / 2.0)
	end

	#===================================================================================

	def HPANNER::PanU_p4_TEX

		HPANNER.PanU_XXX_TEX(1.0 / 4.0)
	end

	#===================================================================================

	def HPANNER::PanU_p8_TEX

		HPANNER.PanU_XXX_TEX(1.0 / 8.0)
	end

	#===================================================================================

	def HPANNER::PanU_p16_TEX

		HPANNER.PanU_XXX_TEX(1.0 / 16.0)
	end

	#===================================================================================

	def HPANNER::PanU_p32_TEX

		HPANNER.PanU_XXX_TEX(1.0 / 32.0)
	end

	#===================================================================================

	def HPANNER::PanU_p64_TEX

		HPANNER.PanU_XXX_TEX(1.0 / 64.0)
	end

	#===================================================================================

	def HPANNER::PanU_p128_TEX

		HPANNER.PanU_XXX_TEX(1.0 / 128.0)
	end

	#===================================================================================

	def HPANNER::PanU_p256_TEX

		HPANNER.PanU_XXX_TEX(1.0 / 256.0)
	end

	#===================================================================================
	#
	#===================================================================================

	def HPANNER::PanU_n2_TEX

		HPANNER.PanU_XXX_TEX(-1.0 / 2.0)
	end

	#===================================================================================

	def HPANNER::PanU_n4_TEX

		HPANNER.PanU_XXX_TEX(-1.0 / 4.0)
	end

	#===================================================================================

	def HPANNER::PanU_n8_TEX

		HPANNER.PanU_XXX_TEX(-1.0 / 8.0)
	end

	#===================================================================================

	def HPANNER::PanU_n16_TEX

		HPANNER.PanU_XXX_TEX(-1.0 / 16.0)
	end

	#===================================================================================

	def HPANNER::PanU_n32_TEX

		HPANNER.PanU_XXX_TEX(-1.0 / 32.0)
	end

	#===================================================================================

	def HPANNER::PanU_n64_TEX

		HPANNER.PanU_XXX_TEX(-1.0 / 64.0)
	end

	#===================================================================================

	def HPANNER::PanU_n128_TEX

		HPANNER.PanU_XXX_TEX(-1.0 / 128.0)
	end

	#===================================================================================

	def HPANNER::PanU_n256_TEX

		HPANNER.PanU_XXX_TEX(-1.0 / 256.0)
	end

	#===================================================================================
	#
	#===================================================================================

	def HPANNER::PanV_YYY_TEX(nY = 0.0)

		a_bBlankFace = true

		a_pProcessor = DataProcessor_Align_TEX.new(1.0, 1.0, 0.0, nY, 0)

		a_pEntities = Sketchup.active_model.active_entities

		a_pIteratorer = DataIteratorerT3D_UTX.new(a_pEntities, a_pProcessor, a_bBlankFace)
		a_pIteratorer.Start(false)
	end

	#===================================================================================
	#
	#===================================================================================

	def HPANNER::PanV_p2_TEX

		HPANNER.PanV_YYY_TEX(1.0 / 2.0)
	end

	#===================================================================================

	def HPANNER::PanV_p4_TEX

		HPANNER.PanV_YYY_TEX(1.0 / 4.0)
	end

	#===================================================================================

	def HPANNER::PanV_p8_TEX

		HPANNER.PanV_YYY_TEX(1.0 / 8.0)
	end

	#===================================================================================

	def HPANNER::PanV_p16_TEX

		HPANNER.PanV_YYY_TEX(1.0 / 16.0)
	end

	#===================================================================================

	def HPANNER::PanV_p32_TEX

		HPANNER.PanV_YYY_TEX(1.0 / 32.0)
	end

	#===================================================================================

	def HPANNER::PanV_p64_TEX

		HPANNER.PanV_YYY_TEX(1.0 / 64.0)
	end

	#===================================================================================

	def HPANNER::PanV_p128_TEX

		HPANNER.PanV_YYY_TEX(1.0 / 128.0)
	end

	#===================================================================================

	def HPANNER::PanV_p256_TEX

		HPANNER.PanV_YYY_TEX(1.0 / 256.0)
	end

	#===================================================================================
	#
	#===================================================================================

	def HPANNER::PanV_n2_TEX

		HPANNER.PanV_YYY_TEX(-1.0 / 2.0)
	end

	#===================================================================================

	def HPANNER::PanV_n4_TEX

		HPANNER.PanV_YYY_TEX(-1.0 / 4.0)
	end

	#===================================================================================

	def HPANNER::PanV_n8_TEX

		HPANNER.PanV_YYY_TEX(-1.0 / 8.0)
	end

	#===================================================================================

	def HPANNER::PanV_n16_TEX

		HPANNER.PanV_YYY_TEX(-1.0 / 16.0)
	end

	#===================================================================================

	def HPANNER::PanV_n32_TEX

		HPANNER.PanV_YYY_TEX(-1.0 / 32.0)
	end

	#===================================================================================

	def HPANNER::PanV_n64_TEX

		HPANNER.PanV_YYY_TEX(-1.0 / 64.0)
	end

	#===================================================================================

	def HPANNER::PanV_n128_TEX

		HPANNER.PanV_YYY_TEX(-1.0 / 128.0)
	end

	#===================================================================================

	def HPANNER::PanV_n256_TEX

		HPANNER.PanV_YYY_TEX(-1.0 / 256.0)
	end

	#===================================================================================
	#
	#===================================================================================

	def HPANNER::PanUV_XXX_TEX(nK = 0.0)

		a_bBlankFace = true

		a_pProcessor = DataProcessor_Align_TEX.new(1.0, 1.0, nK, nK, 0)

		a_pEntities = Sketchup.active_model.active_entities

		a_pIteratorer = DataIteratorerT3D_UTX.new(a_pEntities, a_pProcessor, a_bBlankFace)
		a_pIteratorer.Start(false)
	end

	#===================================================================================
	#
	#===================================================================================

	def HPANNER::PanUV_p2_TEX

		HPANNER.PanUV_XXX_TEX(1.0 / 2.0)
	end

	#===================================================================================

	def HPANNER::PanUV_p4_TEX

		HPANNER.PanUV_XXX_TEX(1.0 / 4.0)
	end

	#===================================================================================

	def HPANNER::PanUV_p8_TEX

		HPANNER.PanUV_XXX_TEX(1.0 / 8.0)
	end

	#===================================================================================

	def HPANNER::PanUV_p16_TEX

		HPANNER.PanUV_XXX_TEX(1.0 / 16.0)
	end

	#===================================================================================

	def HPANNER::PanUV_p32_TEX

		HPANNER.PanUV_XXX_TEX(1.0 / 32.0)
	end

	#===================================================================================

	def HPANNER::PanUV_p64_TEX

		HPANNER.PanUV_XXX_TEX(1.0 / 64.0)
	end

	#===================================================================================

	def HPANNER::PanUV_p128_TEX

		HPANNER.PanUV_XXX_TEX(1.0 / 128.0)
	end

	#===================================================================================

	def HPANNER::PanUV_p256_TEX

		HPANNER.PanUV_XXX_TEX(1.0 / 256.0)
	end

	#===================================================================================
	#
	#===================================================================================

	def HPANNER::PanUV_n2_TEX

		HPANNER.PanUV_XXX_TEX(-1.0 / 2.0)
	end

	#===================================================================================

	def HPANNER::PanUV_n4_TEX

		HPANNER.PanUV_XXX_TEX(-1.0 / 4.0)
	end

	#===================================================================================

	def HPANNER::PanUV_n8_TEX

		HPANNER.PanUV_XXX_TEX(-1.0 / 8.0)
	end

	#===================================================================================

	def HPANNER::PanUV_n16_TEX

		HPANNER.PanUV_XXX_TEX(-1.0 / 16.0)
	end

	#===================================================================================

	def HPANNER::PanUV_n32_TEX

		HPANNER.PanUV_XXX_TEX(-1.0 / 32.0)
	end

	#===================================================================================

	def HPANNER::PanUV_n64_TEX

		HPANNER.PanUV_XXX_TEX(-1.0 / 64.0)
	end

	#===================================================================================

	def HPANNER::PanUV_n128_TEX

		HPANNER.PanUV_XXX_TEX(-1.0 / 128.0)
	end

	#===================================================================================

	def HPANNER::PanUV_n256_TEX

		HPANNER.PanUV_XXX_TEX(-1.0 / 256.0)
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

class HSKP2UNR

	#===================================================================================
	#
	#===================================================================================

	def HSKP2UNR::About

UI.messagebox("HSKP2UNR
HardPCM's Google Sketchup Exporter to T3D/PSK/ASE/OBJ format
for UT3/UT2K4/UT2K3/UT99 game engines.
Version 0.9.8 May 26nd 2009
Freeware under the GNU Lesser General Public License version 2 or later.

 Special thanks to:

  - Dire \"solosnake\" Stockdale 
     www.solosnake.com

  - Darknet

  - Optimus_P-Fat

  - Active_Trash 

 How to install:

   filename is: \"HSKP2UNR.rb\"
   and put it in the directory like
   \"C:\Program Files\Google\Google SketchUp X\Plugins\"

 How this works:

   BEFORE IMPORTING INTO UNREAL EDITOR
   YOUR TEXTURE PACKAGES NEEDED
   MUST BE ALREADY FULLY LOADED OK!!!.

   Everything is in inch OK!!!
   so 512 inch unit = 512 unreal unit
   This apply to the model size
   and the material size.

   When you make a model, you must
   design it with all the textured
   face in WHITE MODE or BLUE MODE
   but you CANNOT MIX THE TWO MODE
   so you must be careful when you
   texturate you model. Double side
   faces are not supported.

   The NAME OF THE MATERIAL is the name
   of the texture that WILL BE EXPORTED
   like UN_Cave.BSP.Materials.M_UN_Cave_BSP_Rock_Floor for instance

   When you want to export as ASE,
   you material name must be only the name
   of the texture like M_UN_Cave_BSP_Rock_Floor

This is another HardPCM stuff from:

http://hardpcmtechnologies.blogspot.com/

Jump!!!
Destroy!!!
Spawn Again!!!
I love Distortions Yaaaa!!!
	

" , MB_MULTILINE , "HSKP2UNR - HardPCM's Exporter for UT3/UT2K4/UT2K3/UT99")

	end

	#===================================================================================
	#
	#===================================================================================

	def HSKP2UNR::ExportT3D

		a_pMenuA = %w[No Yes].join("|")
		a_pMenuB = %w[World Brush].join("|")
		a_pMenuC = %w[Blank Blue].join("|")
		a_pMenuD = %w[Additive Substractive].join("|")
		a_pMenuE = %w[Additive Substractive].join("|")
		a_pMenuF = %w[UT99 UT2K3 UT2K4 UT3].join("|")

		a_pMenuDropdowns = [a_pMenuA, a_pMenuB, a_pMenuC, a_pMenuD, a_pMenuE, a_pMenuF]
		a_pMenuPrompts = ["Export Selected   ", "Export As   ", "Visible Face   ", "Brush Mode   ", "System Mode   ", "Game Engine  "]
		a_pMenuValue = ["No", "Brush", "Blank", "Additive", "Substractive", "UT3"]

		a_pMenuResults = inputbox(a_pMenuPrompts, a_pMenuValue, a_pMenuDropdowns, "Export options")
		return nil if not a_pMenuResults

		a_bWorld = (a_pMenuResults[1] == "World")
		a_bBlankFace = (a_pMenuResults[2] == "Blank")
		a_bAdditiveBrush = (a_pMenuResults[3] == "Additive")
		a_bAdditiveSystem = (a_pMenuResults[4] == "Additive")

		a_sExportPath = UI.savepanel("Export Model to T3D", "" , "Untitled.t3d")
		return if a_sExportPath == nil

		a_pProcessor = nil
		if a_pMenuResults[5] == "UT99"
			a_pProcessor = DataProcessorT3D_UT2K4.new(a_sExportPath, a_bWorld, a_bBlankFace, a_bAdditiveBrush, a_bAdditiveSystem)
		end
		if a_pMenuResults[5] == "UT2K3"
			a_pProcessor = DataProcessorT3D_UT2K4.new(a_sExportPath, a_bWorld, a_bBlankFace, a_bAdditiveBrush, a_bAdditiveSystem)
		end
		if a_pMenuResults[5] == "UT2K4"
			a_pProcessor = DataProcessorT3D_UT2K4.new(a_sExportPath, a_bWorld, a_bBlankFace, a_bAdditiveBrush, a_bAdditiveSystem)
		end
		if a_pMenuResults[5] == "UT3"
			a_pProcessor = DataProcessorT3D_UT3.new(a_sExportPath, a_bWorld, a_bBlankFace, a_bAdditiveBrush, a_bAdditiveSystem)
		end
		return if a_pProcessor == nil

		a_pEntities = Sketchup.active_model.active_entities
		if a_pMenuResults[0] == "Yes"

			a_pEntities = []
			Sketchup.active_model.selection.each{|e| a_pEntities += [e]}
		end

		a_pIteratorer = DataIteratorerT3D_UTX.new(a_pEntities, a_pProcessor, a_bBlankFace)
		a_pIteratorer.Start(true)
	end

	#===================================================================================
	#
	#===================================================================================

	def HSKP2UNR::ExportPSK

		a_pMenuA = %w[Yes No].join("|")
		a_pMenuB = %w[Blue Blank].join("|")

		a_pMenuDropdowns = [a_pMenuA,a_pMenuB]
		a_pMenuPrompts = ["Export Selected   ","Visible face   "]
		a_pMenuValue = ["No","Blank"]

		a_pMenuResults = inputbox a_pMenuPrompts, a_pMenuValue, a_pMenuDropdowns, "Export options"
		return nil if not a_pMenuResults

		a_bBlankFace = (a_pMenuResults[1] == "Blank")

		a_sExportPath = UI.savepanel("Export Model to PSK", "" , "Untitled.psk")
		return if a_sExportPath == nil

		a_pProcessor = DataProcessorPSK_UTX.new(a_sExportPath, "", a_bBlankFace)

		a_pEntities = Sketchup.active_model.active_entities
		if a_pMenuResults[0] == "Yes"

			a_pEntities = []
			Sketchup.active_model.selection.each{|e| a_pEntities += [e]}
		end

		a_pIteratorer = DataIteratorerT3D_UTX.new(a_pEntities, a_pProcessor, a_bBlankFace)
		a_pIteratorer.Start(true)
	end

	#===================================================================================
	#
	#===================================================================================

	def HSKP2UNR::ExportASE

		a_pMenuA = %w[Yes No].join("|")
		a_pMenuB = %w[Blue Blank].join("|")

		a_pMenuDropdowns = [a_pMenuA,a_pMenuB]
		a_pMenuPrompts = ["Export Selected   ","Visible face   "]
		a_pMenuValue = ["No","Blank"]

		a_pMenuResults = inputbox a_pMenuPrompts, a_pMenuValue, a_pMenuDropdowns, "Export options"
		return nil if not a_pMenuResults

		a_bBlankFace = (a_pMenuResults[1] == "Blank")

		a_sExportPath = UI.savepanel("Export Model to ASE", "" , "Untitled.ase")
		return if a_sExportPath == nil

		a_pProcessor = DataExporterASE_UTX.new(a_sExportPath)

		a_pEntities = Sketchup.active_model.active_entities
		if a_pMenuResults[0] == "Yes"

			a_pEntities = []
			Sketchup.active_model.selection.each{|e| a_pEntities += [e]}
		end

		a_pIteratorer = DataIteratorerT3D_UTX.new(a_pEntities, a_pProcessor, a_bBlankFace)
		a_pIteratorer.Start(true)
	end

	#===================================================================================
	#
	#===================================================================================

	def HSKP2UNR::ExportOBJ

		a_pMenuA = %w[Yes No].join("|")
		a_pMenuB = %w[Blue Blank].join("|")

		a_pMenuDropdowns = [a_pMenuA,a_pMenuB]
		a_pMenuPrompts = ["Export Selected   ","Visible face   "]
		a_pMenuValue = ["No","Blank"]

		a_pMenuResults = inputbox a_pMenuPrompts, a_pMenuValue, a_pMenuDropdowns, "Export options"
		return nil if not a_pMenuResults

		a_bBlankFace = (a_pMenuResults[1] == "Blank")

		a_sExportPath = UI.savepanel("Export Model to OBJ", "" , "Untitled.obj")
		return if a_sExportPath == nil

		a_pProcessor = DataExporterOBJ_UTX.new(a_sExportPath)

		a_pEntities = Sketchup.active_model.active_entities
		if a_pMenuResults[0] == "Yes"

			a_pEntities = []
			Sketchup.active_model.selection.each{|e| a_pEntities += [e]}
		end

		a_pIteratorer = DataIteratorerT3D_UTX.new(a_pEntities, a_pProcessor, a_bBlankFace)
		a_pIteratorer.Start(true)
	end

	#===================================================================================
	#
	#===================================================================================

	def HSKP2UNR::ImportOBJ

		a_sImportPath = nil
		a_sImportPath = UI.openpanel("Import Models from OBJ", "" , "*.obj")
		return if a_sImportPath == nil

		a_pProcessor = FileImporterOBJ_UTX.new(a_sImportPath)
		a_pProcessor.Start()
	end

	#===================================================================================
	#
	#===================================================================================

	def HSKP2UNR::ImportT3D

		a_sImportPath = nil
		a_sImportPath = UI.openpanel("Import Models from T3D", "" , "*.t3d")
		return if a_sImportPath == nil

		a_pProcessor = FileImporterT3D_UTX.new(a_sImportPath)
		a_pProcessor.Start()
	end

	#===================================================================================
	#
	#===================================================================================

	def HSKP2UNR::ImportT3DSM

		a_sImportPath = nil
		a_sImportPath = UI.openpanel("Import StaticMeshes from t3d", "" , "*.t3d")
		return if a_sImportPath == nil

		a_pProcessor = FileImporterT3DSM_UTX.new(a_sImportPath)
		a_pProcessor.Start()
	end

	#===================================================================================
	#
	#===================================================================================

end

#===========================================================================================
#
#===========================================================================================

if(not file_loaded?("HSKP2UNR.rb"))

	pMenuA = UI.menu("Plugins").add_submenu("HardPCM's Exporter For Unreal Tournament")

		# --- 
		pMenuB = pMenuA.add_submenu("Export")

			pMenuB.add_item("T3D Format")		{ (HSKP2UNR.ExportT3D) }
			pMenuB.add_item("PSK Format")		{ (HSKP2UNR.ExportPSK) }
			pMenuB.add_item("ASE Format")		{ (HSKP2UNR.ExportASE) }
			pMenuB.add_item("OBJ Format")		{ (HSKP2UNR.ExportOBJ) }

		# --- 
		pMenuC = pMenuA.add_submenu("Import")

			pMenuC.add_item("OBJ Format")		{ (HSKP2UNR.ImportOBJ) }
			pMenuC.add_item("T3D Format")		{ (HSKP2UNR.ImportT3D) }
			pMenuC.add_item("T3D StaticMesh Format")	{ (HSKP2UNR.ImportT3DSM) }

		# --- 
		pMenuD = pMenuA.add_submenu("Texture")

			# --- 
			pMenuE = pMenuD.add_submenu("Pan")
	
				# --- 
				pMenuF = pMenuE.add_submenu("U")
			
					pMenuF.add_item("p 2")			{ (HPANNER.PanU_p2_TEX) }
					pMenuF.add_item("p 4")			{ (HPANNER.PanU_p4_TEX) }
					pMenuF.add_item("p 8")			{ (HPANNER.PanU_p8_TEX) }
					pMenuF.add_item("p 16")			{ (HPANNER.PanU_p16_TEX) }
					pMenuF.add_item("p 32")			{ (HPANNER.PanU_p32_TEX) }
					pMenuF.add_item("p 64")			{ (HPANNER.PanU_p64_TEX) }
					pMenuF.add_item("p 128")		{ (HPANNER.PanU_p128_TEX) }
					pMenuF.add_item("p 256")		{ (HPANNER.PanU_p256_TEX) }
				
					pMenuF.add_item("n 2")			{ (HPANNER.PanU_n2_TEX) }
					pMenuF.add_item("n 4")			{ (HPANNER.PanU_n4_TEX) }
					pMenuF.add_item("n 8")			{ (HPANNER.PanU_n8_TEX) }
					pMenuF.add_item("n 16")			{ (HPANNER.PanU_n16_TEX) }
					pMenuF.add_item("n 32")			{ (HPANNER.PanU_n32_TEX) }
					pMenuF.add_item("n 64")			{ (HPANNER.PanU_n64_TEX) }
					pMenuF.add_item("n 128")		{ (HPANNER.PanU_n128_TEX) }
					pMenuF.add_item("n 256")		{ (HPANNER.PanU_n256_TEX) }

				# --- 
				pMenuG = pMenuE.add_submenu("V")
			
					pMenuG.add_item("p 2")			{ (HPANNER.PanV_p2_TEX) }
					pMenuG.add_item("p 4")			{ (HPANNER.PanV_p4_TEX) }
					pMenuG.add_item("p 8")			{ (HPANNER.PanV_p8_TEX) }
					pMenuG.add_item("p 16")			{ (HPANNER.PanV_p16_TEX) }
					pMenuG.add_item("p 32")			{ (HPANNER.PanV_p32_TEX) }
					pMenuG.add_item("p 64")			{ (HPANNER.PanV_p64_TEX) }
					pMenuG.add_item("p 128")		{ (HPANNER.PanV_p128_TEX) }
					pMenuG.add_item("p 256")		{ (HPANNER.PanV_p256_TEX) }
				
					pMenuG.add_item("n 2")			{ (HPANNER.PanV_n2_TEX) }
					pMenuG.add_item("n 4")			{ (HPANNER.PanV_n4_TEX) }
					pMenuG.add_item("n 8")			{ (HPANNER.PanV_n8_TEX) }
					pMenuG.add_item("n 16")			{ (HPANNER.PanV_n16_TEX) }
					pMenuG.add_item("n 32")			{ (HPANNER.PanV_n32_TEX) }
					pMenuG.add_item("n 64")			{ (HPANNER.PanV_n64_TEX) }
					pMenuG.add_item("n 128")		{ (HPANNER.PanV_n128_TEX) }
					pMenuG.add_item("n 256")		{ (HPANNER.PanV_n256_TEX) }

				# --- 
				pMenuH = pMenuE.add_submenu("UV")
			
					pMenuH.add_item("p 2")			{ (HPANNER.PanUV_p2_TEX) }
					pMenuH.add_item("p 4")			{ (HPANNER.PanUV_p4_TEX) }
					pMenuH.add_item("p 8")			{ (HPANNER.PanUV_p8_TEX) }
					pMenuH.add_item("p 16")			{ (HPANNER.PanUV_p16_TEX) }
					pMenuH.add_item("p 32")			{ (HPANNER.PanUV_p32_TEX) }
					pMenuH.add_item("p 64")			{ (HPANNER.PanUV_p64_TEX) }
					pMenuH.add_item("p 128")		{ (HPANNER.PanUV_p128_TEX) }
					pMenuH.add_item("p 256")		{ (HPANNER.PanUV_p256_TEX) }
				
					pMenuH.add_item("n 2")			{ (HPANNER.PanUV_n2_TEX) }
					pMenuH.add_item("n 4")			{ (HPANNER.PanUV_n4_TEX) }
					pMenuH.add_item("n 8")			{ (HPANNER.PanUV_n8_TEX) }
					pMenuH.add_item("n 16")			{ (HPANNER.PanUV_n16_TEX) }
					pMenuH.add_item("n 32")			{ (HPANNER.PanUV_n32_TEX) }
					pMenuH.add_item("n 64")			{ (HPANNER.PanUV_n64_TEX) }
					pMenuH.add_item("n 128")		{ (HPANNER.PanUV_n128_TEX) }
					pMenuH.add_item("n 256")		{ (HPANNER.PanUV_n256_TEX) }

			# --- 
			pMenuE = pMenuD.add_submenu("Scale")

				# --- 
				pMenuF = pMenuE.add_submenu("U")
			
					pMenuF.add_item("p 2")			{ (HSCALER.ScaleU_p2_TEX) }
					pMenuF.add_item("p 4")			{ (HSCALER.ScaleU_p4_TEX) }
					pMenuF.add_item("p 8")			{ (HSCALER.ScaleU_p8_TEX) }
					pMenuF.add_item("p 16")			{ (HSCALER.ScaleU_p16_TEX) }
					pMenuF.add_item("p 32")			{ (HSCALER.ScaleU_p32_TEX) }
					pMenuF.add_item("p 64")			{ (HSCALER.ScaleU_p64_TEX) }
					pMenuF.add_item("p 128")		{ (HSCALER.ScaleU_p128_TEX) }
					pMenuF.add_item("p 256")		{ (HSCALER.ScaleU_p256_TEX) }
				
					pMenuF.add_item("i 2")			{ (HSCALER.ScaleU_i2_TEX) }
					pMenuF.add_item("i 4")			{ (HSCALER.ScaleU_i4_TEX) }
					pMenuF.add_item("i 8")			{ (HSCALER.ScaleU_i8_TEX) }
					pMenuF.add_item("i 16")			{ (HSCALER.ScaleU_i16_TEX) }
					pMenuF.add_item("i 32")			{ (HSCALER.ScaleU_i32_TEX) }
					pMenuF.add_item("i 64")			{ (HSCALER.ScaleU_i64_TEX) }
					pMenuF.add_item("i 128")		{ (HSCALER.ScaleU_i128_TEX) }
					pMenuF.add_item("i 256")		{ (HSCALER.ScaleU_i256_TEX) }

				# --- 
				pMenuG = pMenuE.add_submenu("V")

					pMenuG.add_item("p 2")			{ (HSCALER.ScaleV_p2_TEX) }
					pMenuG.add_item("p 4")			{ (HSCALER.ScaleV_p4_TEX) }
					pMenuG.add_item("p 8")			{ (HSCALER.ScaleV_p8_TEX) }
					pMenuG.add_item("p 16")			{ (HSCALER.ScaleV_p16_TEX) }
					pMenuG.add_item("p 32")			{ (HSCALER.ScaleV_p32_TEX) }
					pMenuG.add_item("p 64")			{ (HSCALER.ScaleV_p64_TEX) }
					pMenuG.add_item("p 128")		{ (HSCALER.ScaleV_p128_TEX) }
					pMenuG.add_item("p 256")		{ (HSCALER.ScaleV_p256_TEX) }
				
					pMenuG.add_item("i 2")			{ (HSCALER.ScaleV_i2_TEX) }
					pMenuG.add_item("i 4")			{ (HSCALER.ScaleV_i4_TEX) }
					pMenuG.add_item("i 8")			{ (HSCALER.ScaleV_i8_TEX) }
					pMenuG.add_item("i 16")			{ (HSCALER.ScaleV_i16_TEX) }
					pMenuG.add_item("i 32")			{ (HSCALER.ScaleV_i32_TEX) }
					pMenuG.add_item("i 64")			{ (HSCALER.ScaleV_i64_TEX) }
					pMenuG.add_item("i 128")		{ (HSCALER.ScaleV_i128_TEX) }
					pMenuG.add_item("i 256")		{ (HSCALER.ScaleV_i256_TEX) }

				# --- 
				pMenuH = pMenuE.add_submenu("UV")

					pMenuH.add_item("p 2")			{ (HSCALER.ScaleUV_p2_TEX) }
					pMenuH.add_item("p 4")			{ (HSCALER.ScaleUV_p4_TEX) }
					pMenuH.add_item("p 8")			{ (HSCALER.ScaleUV_p8_TEX) }
					pMenuH.add_item("p 16")			{ (HSCALER.ScaleUV_p16_TEX) }
					pMenuH.add_item("p 32")			{ (HSCALER.ScaleUV_p32_TEX) }
					pMenuH.add_item("p 64")			{ (HSCALER.ScaleUV_p64_TEX) }
					pMenuH.add_item("p 128")		{ (HSCALER.ScaleUV_p128_TEX) }
					pMenuH.add_item("p 256")		{ (HSCALER.ScaleUV_p256_TEX) }
				
					pMenuH.add_item("i 2")			{ (HSCALER.ScaleUV_i2_TEX) }
					pMenuH.add_item("i 4")			{ (HSCALER.ScaleUV_i4_TEX) }
					pMenuH.add_item("i 8")			{ (HSCALER.ScaleUV_i8_TEX) }
					pMenuH.add_item("i 16")			{ (HSCALER.ScaleUV_i16_TEX) }
					pMenuH.add_item("i 32")			{ (HSCALER.ScaleUV_i32_TEX) }
					pMenuH.add_item("i 64")			{ (HSCALER.ScaleUV_i64_TEX) }
					pMenuH.add_item("i 128")		{ (HSCALER.ScaleUV_i128_TEX) }
					pMenuH.add_item("i 256")		{ (HSCALER.ScaleUV_i256_TEX) }

	pMenuA.add_separator
	pMenuA.add_item("About HSKP2UNR")	{ (HSKP2UNR.About) }

	UI.add_context_menu_handler do |menu|

		menu.add_separator

		# --- 
		pMenuD = menu.add_submenu("Texture")
	
			# --- 
			pMenuE = pMenuD.add_submenu("Pan")
		
				# --- 
				pMenuF = pMenuE.add_submenu("U")
			
					pMenuF.add_item("p 2")			{ (HPANNER.PanU_p2_TEX) }
					pMenuF.add_item("p 4")			{ (HPANNER.PanU_p4_TEX) }
					pMenuF.add_item("p 8")			{ (HPANNER.PanU_p8_TEX) }
					pMenuF.add_item("p 16")			{ (HPANNER.PanU_p16_TEX) }
					pMenuF.add_item("p 32")			{ (HPANNER.PanU_p32_TEX) }
					pMenuF.add_item("p 64")			{ (HPANNER.PanU_p64_TEX) }
					pMenuF.add_item("p 128")		{ (HPANNER.PanU_p128_TEX) }
					pMenuF.add_item("p 256")		{ (HPANNER.PanU_p256_TEX) }
				
					pMenuF.add_item("n 2")			{ (HPANNER.PanU_n2_TEX) }
					pMenuF.add_item("n 4")			{ (HPANNER.PanU_n4_TEX) }
					pMenuF.add_item("n 8")			{ (HPANNER.PanU_n8_TEX) }
					pMenuF.add_item("n 16")			{ (HPANNER.PanU_n16_TEX) }
					pMenuF.add_item("n 32")			{ (HPANNER.PanU_n32_TEX) }
					pMenuF.add_item("n 64")			{ (HPANNER.PanU_n64_TEX) }
					pMenuF.add_item("n 128")		{ (HPANNER.PanU_n128_TEX) }
					pMenuF.add_item("n 256")		{ (HPANNER.PanU_n256_TEX) }

				# --- 
				pMenuG = pMenuE.add_submenu("V")
		
					pMenuG.add_item("p 2")			{ (HPANNER.PanV_p2_TEX) }
					pMenuG.add_item("p 4")			{ (HPANNER.PanV_p4_TEX) }
					pMenuG.add_item("p 8")			{ (HPANNER.PanV_p8_TEX) }
					pMenuG.add_item("p 16")			{ (HPANNER.PanV_p16_TEX) }
					pMenuG.add_item("p 32")			{ (HPANNER.PanV_p32_TEX) }
					pMenuG.add_item("p 64")			{ (HPANNER.PanV_p64_TEX) }
					pMenuG.add_item("p 128")		{ (HPANNER.PanV_p128_TEX) }
					pMenuG.add_item("p 256")		{ (HPANNER.PanV_p256_TEX) }
				
					pMenuG.add_item("n 2")			{ (HPANNER.PanV_n2_TEX) }
					pMenuG.add_item("n 4")			{ (HPANNER.PanV_n4_TEX) }
					pMenuG.add_item("n 8")			{ (HPANNER.PanV_n8_TEX) }
					pMenuG.add_item("n 16")			{ (HPANNER.PanV_n16_TEX) }
					pMenuG.add_item("n 32")			{ (HPANNER.PanV_n32_TEX) }
					pMenuG.add_item("n 64")			{ (HPANNER.PanV_n64_TEX) }
					pMenuG.add_item("n 128")		{ (HPANNER.PanV_n128_TEX) }
					pMenuG.add_item("n 256")		{ (HPANNER.PanV_n256_TEX) }

				# --- 
				pMenuH = pMenuE.add_submenu("UV")
			
					pMenuH.add_item("p 2")			{ (HPANNER.PanUV_p2_TEX) }
					pMenuH.add_item("p 4")			{ (HPANNER.PanUV_p4_TEX) }
					pMenuH.add_item("p 8")			{ (HPANNER.PanUV_p8_TEX) }
					pMenuH.add_item("p 16")			{ (HPANNER.PanUV_p16_TEX) }
					pMenuH.add_item("p 32")			{ (HPANNER.PanUV_p32_TEX) }
					pMenuH.add_item("p 64")			{ (HPANNER.PanUV_p64_TEX) }
					pMenuH.add_item("p 128")		{ (HPANNER.PanUV_p128_TEX) }
					pMenuH.add_item("p 256")		{ (HPANNER.PanUV_p256_TEX) }
				
					pMenuH.add_item("n 2")			{ (HPANNER.PanUV_n2_TEX) }
					pMenuH.add_item("n 4")			{ (HPANNER.PanUV_n4_TEX) }
					pMenuH.add_item("n 8")			{ (HPANNER.PanUV_n8_TEX) }
					pMenuH.add_item("n 16")			{ (HPANNER.PanUV_n16_TEX) }
					pMenuH.add_item("n 32")			{ (HPANNER.PanUV_n32_TEX) }
					pMenuH.add_item("n 64")			{ (HPANNER.PanUV_n64_TEX) }
					pMenuH.add_item("n 128")		{ (HPANNER.PanUV_n128_TEX) }
					pMenuH.add_item("n 256")		{ (HPANNER.PanUV_n256_TEX) }

			# --- 
			pMenuE = pMenuD.add_submenu("Scale")

				# --- 
				pMenuF = pMenuE.add_submenu("U")

					# --- 
					pMenuF.add_item("p 2")			{ (HSCALER.ScaleU_p2_TEX) }
					pMenuF.add_item("p 4")			{ (HSCALER.ScaleU_p4_TEX) }
					pMenuF.add_item("p 8")			{ (HSCALER.ScaleU_p8_TEX) }
					pMenuF.add_item("p 16")			{ (HSCALER.ScaleU_p16_TEX) }
					pMenuF.add_item("p 32")			{ (HSCALER.ScaleU_p32_TEX) }
					pMenuF.add_item("p 64")			{ (HSCALER.ScaleU_p64_TEX) }
					pMenuF.add_item("p 128")		{ (HSCALER.ScaleU_p128_TEX) }
					pMenuF.add_item("p 256")		{ (HSCALER.ScaleU_p256_TEX) }
				
					pMenuF.add_item("i 2")			{ (HSCALER.ScaleU_i2_TEX) }
					pMenuF.add_item("i 4")			{ (HSCALER.ScaleU_i4_TEX) }
					pMenuF.add_item("i 8")			{ (HSCALER.ScaleU_i8_TEX) }
					pMenuF.add_item("i 16")			{ (HSCALER.ScaleU_i16_TEX) }
					pMenuF.add_item("i 32")			{ (HSCALER.ScaleU_i32_TEX) }
					pMenuF.add_item("i 64")			{ (HSCALER.ScaleU_i64_TEX) }
					pMenuF.add_item("i 128")		{ (HSCALER.ScaleU_i128_TEX) }
					pMenuF.add_item("i 256")		{ (HSCALER.ScaleU_i256_TEX) }

				# --- 
				pMenuG = pMenuE.add_submenu("V")

					# --- 
					pMenuG.add_item("p 2")			{ (HSCALER.ScaleV_p2_TEX) }
					pMenuG.add_item("p 4")			{ (HSCALER.ScaleV_p4_TEX) }
					pMenuG.add_item("p 8")			{ (HSCALER.ScaleV_p8_TEX) }
					pMenuG.add_item("p 16")			{ (HSCALER.ScaleV_p16_TEX) }
					pMenuG.add_item("p 32")			{ (HSCALER.ScaleV_p32_TEX) }
					pMenuG.add_item("p 64")			{ (HSCALER.ScaleV_p64_TEX) }
					pMenuG.add_item("p 128")		{ (HSCALER.ScaleV_p128_TEX) }
					pMenuG.add_item("p 256")		{ (HSCALER.ScaleV_p256_TEX) }
				
					pMenuG.add_item("i 2")			{ (HSCALER.ScaleV_i2_TEX) }
					pMenuG.add_item("i 4")			{ (HSCALER.ScaleV_i4_TEX) }
					pMenuG.add_item("i 8")			{ (HSCALER.ScaleV_i8_TEX) }
					pMenuG.add_item("i 16")			{ (HSCALER.ScaleV_i16_TEX) }
					pMenuG.add_item("i 32")			{ (HSCALER.ScaleV_i32_TEX) }
					pMenuG.add_item("i 64")			{ (HSCALER.ScaleV_i64_TEX) }
					pMenuG.add_item("i 128")		{ (HSCALER.ScaleV_i128_TEX) }
					pMenuG.add_item("i 256")		{ (HSCALER.ScaleV_i256_TEX) }

				# --- 
				pMenuH = pMenuE.add_submenu("UV")

					pMenuH.add_item("p 2")			{ (HSCALER.ScaleUV_p2_TEX) }
					pMenuH.add_item("p 4")			{ (HSCALER.ScaleUV_p4_TEX) }
					pMenuH.add_item("p 8")			{ (HSCALER.ScaleUV_p8_TEX) }
					pMenuH.add_item("p 16")			{ (HSCALER.ScaleUV_p16_TEX) }
					pMenuH.add_item("p 32")			{ (HSCALER.ScaleUV_p32_TEX) }
					pMenuH.add_item("p 64")			{ (HSCALER.ScaleUV_p64_TEX) }
					pMenuH.add_item("p 128")		{ (HSCALER.ScaleUV_p128_TEX) }
					pMenuH.add_item("p 256")		{ (HSCALER.ScaleUV_p256_TEX) }
				
					pMenuH.add_item("i 2")			{ (HSCALER.ScaleUV_i2_TEX) }
					pMenuH.add_item("i 4")			{ (HSCALER.ScaleUV_i4_TEX) }
					pMenuH.add_item("i 8")			{ (HSCALER.ScaleUV_i8_TEX) }
					pMenuH.add_item("i 16")			{ (HSCALER.ScaleUV_i16_TEX) }
					pMenuH.add_item("i 32")			{ (HSCALER.ScaleUV_i32_TEX) }
					pMenuH.add_item("i 64")			{ (HSCALER.ScaleUV_i64_TEX) }
					pMenuH.add_item("i 128")		{ (HSCALER.ScaleUV_i128_TEX) }
					pMenuH.add_item("i 256")		{ (HSCALER.ScaleUV_i256_TEX) }
	end

end
file_loaded(".rb")

#===========================================================================================
#
#===========================================================================================


